module vibe.http.internal.http2.hpack.huffman;

import vibe.http.internal.http2.hpack.exception;

import std.range;

/** Huffman encoding for HPACK header compression
  * The huffman table specifications can be found at:
  * Appendix B of RFC 7541: https://tools.ietf.org/html/rfc7541#appendix-B
*/
void encodeHuffman(I, O)(I source, ref O dst) @safe
{
	ulong bits;
	int bitsLeft = 40;
	size_t len = 0;

	// compute length
	foreach(c; source) {
		auto e = HuffEncodeCodes[c];
		len += e.length;
	}

	dst.put(cast(ubyte)((len/8 ^ 128) + 1));

	foreach(c; source) {
		auto e = HuffEncodeCodes[c];
		bits |= (cast(ulong)e.code) << (bitsLeft - e.length);
		bitsLeft -= e.length;
		while(bitsLeft <= 32) {
			dst.put(cast(ubyte)(bits >> 32));
			bits <<= 8;
			bitsLeft += 8;
		}
	}

	if(bitsLeft != 40) {
		bits |= ((cast(ulong)1) << bitsLeft) - 1;
		dst.put(cast(ubyte)(bits >> 32));
	}
}

@nogc unittest {
	import vibe.internal.array : BatchBuffer;

	string src = "www.example.com";
	BatchBuffer!(ubyte, 13) bres;
	bres.putN(13);
	ubyte[13] expected = [0x8c, 0xf1, 0xe3, 0xc2, 0xe5, 0xf2, 0x3a, 0x6b, 0xa0, 0xab, 0x90, 0xf4, 0xff];
	encodeHuffman(src, bres);
	assert(bres.peekDst == expected);
}

/** Huffman decoding for HPACK header compression
  * The huffman table specifications can be found at:
  * Appendix B of RFC 7541: https://tools.ietf.org/html/rfc7541#appendix-B
*/
void decodeHuffman(I, O)(I source, ref O dst) @safe
{
	auto block = cast(immutable(ubyte)[])source;

	char state = 0;
	char eos = true; // termination flag

	while(!block.empty) {
		char ch = block[0];
		block.popFront();
		decodeSymbol(dst, state, ch >> 4, eos);
		decodeSymbol(dst, state, ch & 0xf, eos);
	}

	enforceHPACK(eos, "Invalid encoded source");
}

private void decodeSymbol(O)(ref O decoded, ref char state, int pos, ref char eos)
@safe
{
	enforceHPACK(state < 256 && pos < 16, "Invalid entry reference");
	auto entry = HuffDecodeCodes[state][pos];
	enforceHPACK(entry.next != state, "Invalid symbol");

	if (entry.emit) { // if the symbol is terminal
		auto sym = cast(immutable(char))entry.symbol;
		decoded.put(sym);
	}

	state = entry.next;
	eos = entry.ending;
}

unittest {
	import std.array;
	immutable ubyte[] test = [0xf1, 0xe3, 0xc2, 0xe5, 0xf2, 0x3a, 0x6b, 0xa0, 0xab, 0x90, 0xf4, 0xff];

	auto dst = appender!string;
	decodeHuffman(test, dst);
	assert(dst.data == "www.example.com");

	immutable char[] ctest = [0xf1, 0xe3, 0xc2, 0xe5, 0xf2, 0x3a, 0x6b, 0xa0, 0xab, 0x90, 0xf4, 0xff];
	auto dst2 = appender!string;
	decodeHuffman(ctest, dst2);
	assert(dst2.data == "www.example.com");
}

unittest { // could be @nogc (exceptions aren't)
	import vibe.internal.array : BatchBuffer;

	immutable ubyte[12] src = [0xf1, 0xe3, 0xc2, 0xe5, 0xf2, 0x3a, 0x6b, 0xa0, 0xab, 0x90, 0xf4, 0xff];
	BatchBuffer!(char, 15) bres;
	bres.putN(15);
	decodeHuffman(src, bres);
	assert(bres.peekDst == "www.example.com");
}

private struct HuffCode {
	char next;
	char emit;
	char symbol;
	char ending;
}

private struct HuffEncCode {
	uint code;
	size_t length;
}

private static immutable HuffEncCode[256] HuffEncodeCodes = [
    {0x00001ff8, 13}, {0x007fffd8, 23}, {0x0fffffe2, 28}, {0x0fffffe3, 28},
    {0x0fffffe4, 28}, {0x0fffffe5, 28}, {0x0fffffe6, 28}, {0x0fffffe7, 28},
    {0x0fffffe8, 28}, {0x00ffffea, 24}, {0x3ffffffc, 30}, {0x0fffffe9, 28},
    {0x0fffffea, 28}, {0x3ffffffd, 30}, {0x0fffffeb, 28}, {0x0fffffec, 28},
    {0x0fffffed, 28}, {0x0fffffee, 28}, {0x0fffffef, 28}, {0x0ffffff0, 28},
    {0x0ffffff1, 28}, {0x0ffffff2, 28}, {0x3ffffffe, 30}, {0x0ffffff3, 28},
    {0x0ffffff4, 28}, {0x0ffffff5, 28}, {0x0ffffff6, 28}, {0x0ffffff7, 28},
    {0x0ffffff8, 28}, {0x0ffffff9, 28}, {0x0ffffffa, 28}, {0x0ffffffb, 28},
    {0x00000014,  6}, {0x000003f8, 10}, {0x000003f9, 10}, {0x00000ffa, 12},
    {0x00001ff9, 13}, {0x00000015,  6}, {0x000000f8,  8}, {0x000007fa, 11},
    {0x000003fa, 10}, {0x000003fb, 10}, {0x000000f9,  8}, {0x000007fb, 11},
    {0x000000fa,  8}, {0x00000016,  6}, {0x00000017,  6}, {0x00000018,  6},
    {0x00000000,  5}, {0x00000001,  5}, {0x00000002,  5}, {0x00000019,  6},
    {0x0000001a,  6}, {0x0000001b,  6}, {0x0000001c,  6}, {0x0000001d,  6},
    {0x0000001e,  6}, {0x0000001f,  6}, {0x0000005c,  7}, {0x000000fb,  8},
    {0x00007ffc, 15}, {0x00000020,  6}, {0x00000ffb, 12}, {0x000003fc, 10},
    {0x00001ffa, 13}, {0x00000021,  6}, {0x0000005d,  7}, {0x0000005e,  7},
    {0x0000005f,  7}, {0x00000060,  7}, {0x00000061,  7}, {0x00000062,  7},
    {0x00000063,  7}, {0x00000064,  7}, {0x00000065,  7}, {0x00000066,  7},
    {0x00000067,  7}, {0x00000068,  7}, {0x00000069,  7}, {0x0000006a,  7},
    {0x0000006b,  7}, {0x0000006c,  7}, {0x0000006d,  7}, {0x0000006e,  7},
    {0x0000006f,  7}, {0x00000070,  7}, {0x00000071,  7}, {0x00000072,  7},
    {0x000000fc,  8}, {0x00000073,  7}, {0x000000fd,  8}, {0x00001ffb, 13},
    {0x0007fff0, 19}, {0x00001ffc, 13}, {0x00003ffc, 14}, {0x00000022,  6},
    {0x00007ffd, 15}, {0x00000003,  5}, {0x00000023,  6}, {0x00000004,  5},
    {0x00000024,  6}, {0x00000005,  5}, {0x00000025,  6}, {0x00000026,  6},
    {0x00000027,  6}, {0x00000006,  5}, {0x00000074,  7}, {0x00000075,  7},
    {0x00000028,  6}, {0x00000029,  6}, {0x0000002a,  6}, {0x00000007,  5},
    {0x0000002b,  6}, {0x00000076,  7}, {0x0000002c,  6}, {0x00000008,  5},
    {0x00000009,  5}, {0x0000002d,  6}, {0x00000077,  7}, {0x00000078,  7},
    {0x00000079,  7}, {0x0000007a,  7}, {0x0000007b,  7}, {0x00007ffe, 15},
    {0x000007fc, 11}, {0x00003ffd, 14}, {0x00001ffd, 13}, {0x0ffffffc, 28},
    {0x000fffe6, 20}, {0x003fffd2, 22}, {0x000fffe7, 20}, {0x000fffe8, 20},
    {0x003fffd3, 22}, {0x003fffd4, 22}, {0x003fffd5, 22}, {0x007fffd9, 23},
    {0x003fffd6, 22}, {0x007fffda, 23}, {0x007fffdb, 23}, {0x007fffdc, 23},
    {0x007fffdd, 23}, {0x007fffde, 23}, {0x00ffffeb, 24}, {0x007fffdf, 23},
    {0x00ffffec, 24}, {0x00ffffed, 24}, {0x003fffd7, 22}, {0x007fffe0, 23},
    {0x00ffffee, 24}, {0x007fffe1, 23}, {0x007fffe2, 23}, {0x007fffe3, 23},
    {0x007fffe4, 23}, {0x001fffdc, 21}, {0x003fffd8, 22}, {0x007fffe5, 23},
    {0x003fffd9, 22}, {0x007fffe6, 23}, {0x007fffe7, 23}, {0x00ffffef, 24},
    {0x003fffda, 22}, {0x001fffdd, 21}, {0x000fffe9, 20}, {0x003fffdb, 22},
    {0x003fffdc, 22}, {0x007fffe8, 23}, {0x007fffe9, 23}, {0x001fffde, 21},
    {0x007fffea, 23}, {0x003fffdd, 22}, {0x003fffde, 22}, {0x00fffff0, 24},
    {0x001fffdf, 21}, {0x003fffdf, 22}, {0x007fffeb, 23}, {0x007fffec, 23},
    {0x001fffe0, 21}, {0x001fffe1, 21}, {0x003fffe0, 22}, {0x001fffe2, 21},
    {0x007fffed, 23}, {0x003fffe1, 22}, {0x007fffee, 23}, {0x007fffef, 23},
    {0x000fffea, 20}, {0x003fffe2, 22}, {0x003fffe3, 22}, {0x003fffe4, 22},
    {0x007ffff0, 23}, {0x003fffe5, 22}, {0x003fffe6, 22}, {0x007ffff1, 23},
    {0x03ffffe0, 26}, {0x03ffffe1, 26}, {0x000fffeb, 20}, {0x0007fff1, 19},
    {0x003fffe7, 22}, {0x007ffff2, 23}, {0x003fffe8, 22}, {0x01ffffec, 25},
    {0x03ffffe2, 26}, {0x03ffffe3, 26}, {0x03ffffe4, 26}, {0x07ffffde, 27},
    {0x07ffffdf, 27}, {0x03ffffe5, 26}, {0x00fffff1, 24}, {0x01ffffed, 25},
    {0x0007fff2, 19}, {0x001fffe3, 21}, {0x03ffffe6, 26}, {0x07ffffe0, 27},
    {0x07ffffe1, 27}, {0x03ffffe7, 26}, {0x07ffffe2, 27}, {0x00fffff2, 24},
    {0x001fffe4, 21}, {0x001fffe5, 21}, {0x03ffffe8, 26}, {0x03ffffe9, 26},
    {0x0ffffffd, 28}, {0x07ffffe3, 27}, {0x07ffffe4, 27}, {0x07ffffe5, 27},
    {0x000fffec, 20}, {0x00fffff3, 24}, {0x000fffed, 20}, {0x001fffe6, 21},
    {0x003fffe9, 22}, {0x001fffe7, 21}, {0x001fffe8, 21}, {0x007ffff3, 23},
    {0x003fffea, 22}, {0x003fffeb, 22}, {0x01ffffee, 25}, {0x01ffffef, 25},
    {0x00fffff4, 24}, {0x00fffff5, 24}, {0x03ffffea, 26}, {0x007ffff4, 23},
    {0x03ffffeb, 26}, {0x07ffffe6, 27}, {0x03ffffec, 26}, {0x03ffffed, 26},
    {0x07ffffe7, 27}, {0x07ffffe8, 27}, {0x07ffffe9, 27}, {0x07ffffea, 27},
    {0x07ffffeb, 27}, {0x0ffffffe, 28}, {0x07ffffec, 27}, {0x07ffffed, 27},
    {0x07ffffee, 27}, {0x07ffffef, 27}, {0x07fffff0, 27}, {0x03ffffee, 26}
];

private static immutable HuffCode[16][256] HuffDecodeCodes = [
	/* 0 */
	[
// 		 next  emit  sym  ending   next  emit  sym  ending
		{0x04, 0x00, 0x00, 0x00}, {0x05, 0x00, 0x00, 0x00},
		{0x07, 0x00, 0x00, 0x00}, {0x08, 0x00, 0x00, 0x00},
		{0x0b, 0x00, 0x00, 0x00}, {0x0c, 0x00, 0x00, 0x00},
		{0x10, 0x00, 0x00, 0x00}, {0x13, 0x00, 0x00, 0x00},
		{0x19, 0x00, 0x00, 0x00}, {0x1c, 0x00, 0x00, 0x00},
		{0x20, 0x00, 0x00, 0x00}, {0x23, 0x00, 0x00, 0x00},
		{0x2a, 0x00, 0x00, 0x00}, {0x31, 0x00, 0x00, 0x00},
		{0x39, 0x00, 0x00, 0x00}, {0x40, 0x00, 0x00, 0x01}
	],
	[
		{0x00, 0x01, 0x30, 0x01}, {0x00, 0x01, 0x31, 0x01},
		{0x00, 0x01, 0x32, 0x01}, {0x00, 0x01, 0x61, 0x01},
		{0x00, 0x01, 0x63, 0x01}, {0x00, 0x01, 0x65, 0x01},
		{0x00, 0x01, 0x69, 0x01}, {0x00, 0x01, 0x6f, 0x01},
		{0x00, 0x01, 0x73, 0x01}, {0x00, 0x01, 0x74, 0x01},
		{0x0d, 0x00, 0x00, 0x00}, {0x0e, 0x00, 0x00, 0x00},
		{0x11, 0x00, 0x00, 0x00}, {0x12, 0x00, 0x00, 0x00},
		{0x14, 0x00, 0x00, 0x00}, {0x15, 0x00, 0x00, 0x00}
	],
	[
		{0x01, 0x01, 0x30, 0x00}, {0x16, 0x01, 0x30, 0x01},
		{0x01, 0x01, 0x31, 0x00}, {0x16, 0x01, 0x31, 0x01},
		{0x01, 0x01, 0x32, 0x00}, {0x16, 0x01, 0x32, 0x01},
		{0x01, 0x01, 0x61, 0x00}, {0x16, 0x01, 0x61, 0x01},
		{0x01, 0x01, 0x63, 0x00}, {0x16, 0x01, 0x63, 0x01},
		{0x01, 0x01, 0x65, 0x00}, {0x16, 0x01, 0x65, 0x01},
		{0x01, 0x01, 0x69, 0x00}, {0x16, 0x01, 0x69, 0x01},
		{0x01, 0x01, 0x6f, 0x00}, {0x16, 0x01, 0x6f, 0x01}
	],
	[
		{0x02, 0x01, 0x30, 0x00}, {0x09, 0x01, 0x30, 0x00},
		{0x17, 0x01, 0x30, 0x00}, {0x28, 0x01, 0x30, 0x01},
		{0x02, 0x01, 0x31, 0x00}, {0x09, 0x01, 0x31, 0x00},
		{0x17, 0x01, 0x31, 0x00}, {0x28, 0x01, 0x31, 0x01},
		{0x02, 0x01, 0x32, 0x00}, {0x09, 0x01, 0x32, 0x00},
		{0x17, 0x01, 0x32, 0x00}, {0x28, 0x01, 0x32, 0x01},
		{0x02, 0x01, 0x61, 0x00}, {0x09, 0x01, 0x61, 0x00},
		{0x17, 0x01, 0x61, 0x00}, {0x28, 0x01, 0x61, 0x01}
	],
	[
		{0x03, 0x01, 0x30, 0x00}, {0x06, 0x01, 0x30, 0x00},
		{0x0a, 0x01, 0x30, 0x00}, {0x0f, 0x01, 0x30, 0x00},
		{0x18, 0x01, 0x30, 0x00}, {0x1f, 0x01, 0x30, 0x00},
		{0x29, 0x01, 0x30, 0x00}, {0x38, 0x01, 0x30, 0x01},
		{0x03, 0x01, 0x31, 0x00}, {0x06, 0x01, 0x31, 0x00},
		{0x0a, 0x01, 0x31, 0x00}, {0x0f, 0x01, 0x31, 0x00},
		{0x18, 0x01, 0x31, 0x00}, {0x1f, 0x01, 0x31, 0x00},
		{0x29, 0x01, 0x31, 0x00}, {0x38, 0x01, 0x31, 0x01}
	],
	/* 5 */
	[
		{0x03, 0x01, 0x32, 0x00}, {0x06, 0x01, 0x32, 0x00},
		{0x0a, 0x01, 0x32, 0x00}, {0x0f, 0x01, 0x32, 0x00},
		{0x18, 0x01, 0x32, 0x00}, {0x1f, 0x01, 0x32, 0x00},
		{0x29, 0x01, 0x32, 0x00}, {0x38, 0x01, 0x32, 0x01},
		{0x03, 0x01, 0x61, 0x00}, {0x06, 0x01, 0x61, 0x00},
		{0x0a, 0x01, 0x61, 0x00}, {0x0f, 0x01, 0x61, 0x00},
		{0x18, 0x01, 0x61, 0x00}, {0x1f, 0x01, 0x61, 0x00},
		{0x29, 0x01, 0x61, 0x00}, {0x38, 0x01, 0x61, 0x01}
	],
	[
		{0x02, 0x01, 0x63, 0x00}, {0x09, 0x01, 0x63, 0x00},
		{0x17, 0x01, 0x63, 0x00}, {0x28, 0x01, 0x63, 0x01},
		{0x02, 0x01, 0x65, 0x00}, {0x09, 0x01, 0x65, 0x00},
		{0x17, 0x01, 0x65, 0x00}, {0x28, 0x01, 0x65, 0x01},
		{0x02, 0x01, 0x69, 0x00}, {0x09, 0x01, 0x69, 0x00},
		{0x17, 0x01, 0x69, 0x00}, {0x28, 0x01, 0x69, 0x01},
		{0x02, 0x01, 0x6f, 0x00}, {0x09, 0x01, 0x6f, 0x00},
		{0x17, 0x01, 0x6f, 0x00}, {0x28, 0x01, 0x6f, 0x01}
	],
	[
		{0x03, 0x01, 0x63, 0x00}, {0x06, 0x01, 0x63, 0x00},
		{0x0a, 0x01, 0x63, 0x00}, {0x0f, 0x01, 0x63, 0x00},
		{0x18, 0x01, 0x63, 0x00}, {0x1f, 0x01, 0x63, 0x00},
		{0x29, 0x01, 0x63, 0x00}, {0x38, 0x01, 0x63, 0x01},
		{0x03, 0x01, 0x65, 0x00}, {0x06, 0x01, 0x65, 0x00},
		{0x0a, 0x01, 0x65, 0x00}, {0x0f, 0x01, 0x65, 0x00},
		{0x18, 0x01, 0x65, 0x00}, {0x1f, 0x01, 0x65, 0x00},
		{0x29, 0x01, 0x65, 0x00}, {0x38, 0x01, 0x65, 0x01}
	],
	[
		{0x03, 0x01, 0x69, 0x00}, {0x06, 0x01, 0x69, 0x00},
		{0x0a, 0x01, 0x69, 0x00}, {0x0f, 0x01, 0x69, 0x00},
		{0x18, 0x01, 0x69, 0x00}, {0x1f, 0x01, 0x69, 0x00},
		{0x29, 0x01, 0x69, 0x00}, {0x38, 0x01, 0x69, 0x01},
		{0x03, 0x01, 0x6f, 0x00}, {0x06, 0x01, 0x6f, 0x00},
		{0x0a, 0x01, 0x6f, 0x00}, {0x0f, 0x01, 0x6f, 0x00},
		{0x18, 0x01, 0x6f, 0x00}, {0x1f, 0x01, 0x6f, 0x00},
		{0x29, 0x01, 0x6f, 0x00}, {0x38, 0x01, 0x6f, 0x01}
	],
	[
		{0x01, 0x01, 0x73, 0x00}, {0x16, 0x01, 0x73, 0x01},
		{0x01, 0x01, 0x74, 0x00}, {0x16, 0x01, 0x74, 0x01},
		{0x00, 0x01, 0x20, 0x01}, {0x00, 0x01, 0x25, 0x01},
		{0x00, 0x01, 0x2d, 0x01}, {0x00, 0x01, 0x2e, 0x01},
		{0x00, 0x01, 0x2f, 0x01}, {0x00, 0x01, 0x33, 0x01},
		{0x00, 0x01, 0x34, 0x01}, {0x00, 0x01, 0x35, 0x01},
		{0x00, 0x01, 0x36, 0x01}, {0x00, 0x01, 0x37, 0x01},
		{0x00, 0x01, 0x38, 0x01}, {0x00, 0x01, 0x39, 0x01}
	],
	/* 10 */
	[
		{0x02, 0x01, 0x73, 0x00}, {0x09, 0x01, 0x73, 0x00},
		{0x17, 0x01, 0x73, 0x00}, {0x28, 0x01, 0x73, 0x01},
		{0x02, 0x01, 0x74, 0x00}, {0x09, 0x01, 0x74, 0x00},
		{0x17, 0x01, 0x74, 0x00}, {0x28, 0x01, 0x74, 0x01},
		{0x01, 0x01, 0x20, 0x00}, {0x16, 0x01, 0x20, 0x01},
		{0x01, 0x01, 0x25, 0x00}, {0x16, 0x01, 0x25, 0x01},
		{0x01, 0x01, 0x2d, 0x00}, {0x16, 0x01, 0x2d, 0x01},
		{0x01, 0x01, 0x2e, 0x00}, {0x16, 0x01, 0x2e, 0x01}
	],
	[
		{0x03, 0x01, 0x73, 0x00}, {0x06, 0x01, 0x73, 0x00},
		{0x0a, 0x01, 0x73, 0x00}, {0x0f, 0x01, 0x73, 0x00},
		{0x18, 0x01, 0x73, 0x00}, {0x1f, 0x01, 0x73, 0x00},
		{0x29, 0x01, 0x73, 0x00}, {0x38, 0x01, 0x73, 0x01},
		{0x03, 0x01, 0x74, 0x00}, {0x06, 0x01, 0x74, 0x00},
		{0x0a, 0x01, 0x74, 0x00}, {0x0f, 0x01, 0x74, 0x00},
		{0x18, 0x01, 0x74, 0x00}, {0x1f, 0x01, 0x74, 0x00},
		{0x29, 0x01, 0x74, 0x00}, {0x38, 0x01, 0x74, 0x01}
	],
	[
		{0x02, 0x01, 0x20, 0x00}, {0x09, 0x01, 0x20, 0x00},
		{0x17, 0x01, 0x20, 0x00}, {0x28, 0x01, 0x20, 0x01},
		{0x02, 0x01, 0x25, 0x00}, {0x09, 0x01, 0x25, 0x00},
		{0x17, 0x01, 0x25, 0x00}, {0x28, 0x01, 0x25, 0x01},
		{0x02, 0x01, 0x2d, 0x00}, {0x09, 0x01, 0x2d, 0x00},
		{0x17, 0x01, 0x2d, 0x00}, {0x28, 0x01, 0x2d, 0x01},
		{0x02, 0x01, 0x2e, 0x00}, {0x09, 0x01, 0x2e, 0x00},
		{0x17, 0x01, 0x2e, 0x00}, {0x28, 0x01, 0x2e, 0x01}
	],
	[
		{0x03, 0x01, 0x20, 0x00}, {0x06, 0x01, 0x20, 0x00},
		{0x0a, 0x01, 0x20, 0x00}, {0x0f, 0x01, 0x20, 0x00},
		{0x18, 0x01, 0x20, 0x00}, {0x1f, 0x01, 0x20, 0x00},
		{0x29, 0x01, 0x20, 0x00}, {0x38, 0x01, 0x20, 0x01},
		{0x03, 0x01, 0x25, 0x00}, {0x06, 0x01, 0x25, 0x00},
		{0x0a, 0x01, 0x25, 0x00}, {0x0f, 0x01, 0x25, 0x00},
		{0x18, 0x01, 0x25, 0x00}, {0x1f, 0x01, 0x25, 0x00},
		{0x29, 0x01, 0x25, 0x00}, {0x38, 0x01, 0x25, 0x01}
	],
	[
		{0x03, 0x01, 0x2d, 0x00}, {0x06, 0x01, 0x2d, 0x00},
		{0x0a, 0x01, 0x2d, 0x00}, {0x0f, 0x01, 0x2d, 0x00},
		{0x18, 0x01, 0x2d, 0x00}, {0x1f, 0x01, 0x2d, 0x00},
		{0x29, 0x01, 0x2d, 0x00}, {0x38, 0x01, 0x2d, 0x01},
		{0x03, 0x01, 0x2e, 0x00}, {0x06, 0x01, 0x2e, 0x00},
		{0x0a, 0x01, 0x2e, 0x00}, {0x0f, 0x01, 0x2e, 0x00},
		{0x18, 0x01, 0x2e, 0x00}, {0x1f, 0x01, 0x2e, 0x00},
		{0x29, 0x01, 0x2e, 0x00}, {0x38, 0x01, 0x2e, 0x01}
	],
	/* 15 */
	[
		{0x01, 0x01, 0x2f, 0x00}, {0x16, 0x01, 0x2f, 0x01},
		{0x01, 0x01, 0x33, 0x00}, {0x16, 0x01, 0x33, 0x01},
		{0x01, 0x01, 0x34, 0x00}, {0x16, 0x01, 0x34, 0x01},
		{0x01, 0x01, 0x35, 0x00}, {0x16, 0x01, 0x35, 0x01},
		{0x01, 0x01, 0x36, 0x00}, {0x16, 0x01, 0x36, 0x01},
		{0x01, 0x01, 0x37, 0x00}, {0x16, 0x01, 0x37, 0x01},
		{0x01, 0x01, 0x38, 0x00}, {0x16, 0x01, 0x38, 0x01},
		{0x01, 0x01, 0x39, 0x00}, {0x16, 0x01, 0x39, 0x01}
	],
	[
		{0x02, 0x01, 0x2f, 0x00}, {0x09, 0x01, 0x2f, 0x00},
		{0x17, 0x01, 0x2f, 0x00}, {0x28, 0x01, 0x2f, 0x01},
		{0x02, 0x01, 0x33, 0x00}, {0x09, 0x01, 0x33, 0x00},
		{0x17, 0x01, 0x33, 0x00}, {0x28, 0x01, 0x33, 0x01},
		{0x02, 0x01, 0x34, 0x00}, {0x09, 0x01, 0x34, 0x00},
		{0x17, 0x01, 0x34, 0x00}, {0x28, 0x01, 0x34, 0x01},
		{0x02, 0x01, 0x35, 0x00}, {0x09, 0x01, 0x35, 0x00},
		{0x17, 0x01, 0x35, 0x00}, {0x28, 0x01, 0x35, 0x01}
	],
	[
		{0x03, 0x01, 0x2f, 0x00}, {0x06, 0x01, 0x2f, 0x00},
		{0x0a, 0x01, 0x2f, 0x00}, {0x0f, 0x01, 0x2f, 0x00},
		{0x18, 0x01, 0x2f, 0x00}, {0x1f, 0x01, 0x2f, 0x00},
		{0x29, 0x01, 0x2f, 0x00}, {0x38, 0x01, 0x2f, 0x01},
		{0x03, 0x01, 0x33, 0x00}, {0x06, 0x01, 0x33, 0x00},
		{0x0a, 0x01, 0x33, 0x00}, {0x0f, 0x01, 0x33, 0x00},
		{0x18, 0x01, 0x33, 0x00}, {0x1f, 0x01, 0x33, 0x00},
		{0x29, 0x01, 0x33, 0x00}, {0x38, 0x01, 0x33, 0x01}
	],
	[
		{0x03, 0x01, 0x34, 0x00}, {0x06, 0x01, 0x34, 0x00},
		{0x0a, 0x01, 0x34, 0x00}, {0x0f, 0x01, 0x34, 0x00},
		{0x18, 0x01, 0x34, 0x00}, {0x1f, 0x01, 0x34, 0x00},
		{0x29, 0x01, 0x34, 0x00}, {0x38, 0x01, 0x34, 0x01},
		{0x03, 0x01, 0x35, 0x00}, {0x06, 0x01, 0x35, 0x00},
		{0x0a, 0x01, 0x35, 0x00}, {0x0f, 0x01, 0x35, 0x00},
		{0x18, 0x01, 0x35, 0x00}, {0x1f, 0x01, 0x35, 0x00},
		{0x29, 0x01, 0x35, 0x00}, {0x38, 0x01, 0x35, 0x01}
	],
	[
		{0x02, 0x01, 0x36, 0x00}, {0x09, 0x01, 0x36, 0x00},
		{0x17, 0x01, 0x36, 0x00}, {0x28, 0x01, 0x36, 0x01},
		{0x02, 0x01, 0x37, 0x00}, {0x09, 0x01, 0x37, 0x00},
		{0x17, 0x01, 0x37, 0x00}, {0x28, 0x01, 0x37, 0x01},
		{0x02, 0x01, 0x38, 0x00}, {0x09, 0x01, 0x38, 0x00},
		{0x17, 0x01, 0x38, 0x00}, {0x28, 0x01, 0x38, 0x01},
		{0x02, 0x01, 0x39, 0x00}, {0x09, 0x01, 0x39, 0x00},
		{0x17, 0x01, 0x39, 0x00}, {0x28, 0x01, 0x39, 0x01}
	],
	/* 20 */
	[
		{0x03, 0x01, 0x36, 0x00}, {0x06, 0x01, 0x36, 0x00},
		{0x0a, 0x01, 0x36, 0x00}, {0x0f, 0x01, 0x36, 0x00},
		{0x18, 0x01, 0x36, 0x00}, {0x1f, 0x01, 0x36, 0x00},
		{0x29, 0x01, 0x36, 0x00}, {0x38, 0x01, 0x36, 0x01},
		{0x03, 0x01, 0x37, 0x00}, {0x06, 0x01, 0x37, 0x00},
		{0x0a, 0x01, 0x37, 0x00}, {0x0f, 0x01, 0x37, 0x00},
		{0x18, 0x01, 0x37, 0x00}, {0x1f, 0x01, 0x37, 0x00},
		{0x29, 0x01, 0x37, 0x00}, {0x38, 0x01, 0x37, 0x01}
	],
	[
		{0x03, 0x01, 0x38, 0x00}, {0x06, 0x01, 0x38, 0x00},
		{0x0a, 0x01, 0x38, 0x00}, {0x0f, 0x01, 0x38, 0x00},
		{0x18, 0x01, 0x38, 0x00}, {0x1f, 0x01, 0x38, 0x00},
		{0x29, 0x01, 0x38, 0x00}, {0x38, 0x01, 0x38, 0x01},
		{0x03, 0x01, 0x39, 0x00}, {0x06, 0x01, 0x39, 0x00},
		{0x0a, 0x01, 0x39, 0x00}, {0x0f, 0x01, 0x39, 0x00},
		{0x18, 0x01, 0x39, 0x00}, {0x1f, 0x01, 0x39, 0x00},
		{0x29, 0x01, 0x39, 0x00}, {0x38, 0x01, 0x39, 0x01}
	],
	[
		{0x1a, 0x00, 0x00, 0x00}, {0x1b, 0x00, 0x00, 0x00},
		{0x1d, 0x00, 0x00, 0x00}, {0x1e, 0x00, 0x00, 0x00},
		{0x21, 0x00, 0x00, 0x00}, {0x22, 0x00, 0x00, 0x00},
		{0x24, 0x00, 0x00, 0x00}, {0x25, 0x00, 0x00, 0x00},
		{0x2b, 0x00, 0x00, 0x00}, {0x2e, 0x00, 0x00, 0x00},
		{0x32, 0x00, 0x00, 0x00}, {0x35, 0x00, 0x00, 0x00},
		{0x3a, 0x00, 0x00, 0x00}, {0x3d, 0x00, 0x00, 0x00},
		{0x41, 0x00, 0x00, 0x00}, {0x44, 0x00, 0x00, 0x01}
	],
	[
		{0x00, 0x01, 0x3d, 0x01}, {0x00, 0x01, 0x41, 0x01},
		{0x00, 0x01, 0x5f, 0x01}, {0x00, 0x01, 0x62, 0x01},
		{0x00, 0x01, 0x64, 0x01}, {0x00, 0x01, 0x66, 0x01},
		{0x00, 0x01, 0x67, 0x01}, {0x00, 0x01, 0x68, 0x01},
		{0x00, 0x01, 0x6c, 0x01}, {0x00, 0x01, 0x6d, 0x01},
		{0x00, 0x01, 0x6e, 0x01}, {0x00, 0x01, 0x70, 0x01},
		{0x00, 0x01, 0x72, 0x01}, {0x00, 0x01, 0x75, 0x01},
		{0x26, 0x00, 0x00, 0x00}, {0x27, 0x00, 0x00, 0x00}
	],
	[
		{0x01, 0x01, 0x3d, 0x00}, {0x16, 0x01, 0x3d, 0x01},
		{0x01, 0x01, 0x41, 0x00}, {0x16, 0x01, 0x41, 0x01},
		{0x01, 0x01, 0x5f, 0x00}, {0x16, 0x01, 0x5f, 0x01},
		{0x01, 0x01, 0x62, 0x00}, {0x16, 0x01, 0x62, 0x01},
		{0x01, 0x01, 0x64, 0x00}, {0x16, 0x01, 0x64, 0x01},
		{0x01, 0x01, 0x66, 0x00}, {0x16, 0x01, 0x66, 0x01},
		{0x01, 0x01, 0x67, 0x00}, {0x16, 0x01, 0x67, 0x01},
		{0x01, 0x01, 0x68, 0x00}, {0x16, 0x01, 0x68, 0x01}
	],
	/* 25 */
	[
		{0x02, 0x01, 0x3d, 0x00}, {0x09, 0x01, 0x3d, 0x00},
		{0x17, 0x01, 0x3d, 0x00}, {0x28, 0x01, 0x3d, 0x01},
		{0x02, 0x01, 0x41, 0x00}, {0x09, 0x01, 0x41, 0x00},
		{0x17, 0x01, 0x41, 0x00}, {0x28, 0x01, 0x41, 0x01},
		{0x02, 0x01, 0x5f, 0x00}, {0x09, 0x01, 0x5f, 0x00},
		{0x17, 0x01, 0x5f, 0x00}, {0x28, 0x01, 0x5f, 0x01},
		{0x02, 0x01, 0x62, 0x00}, {0x09, 0x01, 0x62, 0x00},
		{0x17, 0x01, 0x62, 0x00}, {0x28, 0x01, 0x62, 0x01}
	],
	[
		{0x03, 0x01, 0x3d, 0x00}, {0x06, 0x01, 0x3d, 0x00},
		{0x0a, 0x01, 0x3d, 0x00}, {0x0f, 0x01, 0x3d, 0x00},
		{0x18, 0x01, 0x3d, 0x00}, {0x1f, 0x01, 0x3d, 0x00},
		{0x29, 0x01, 0x3d, 0x00}, {0x38, 0x01, 0x3d, 0x01},
		{0x03, 0x01, 0x41, 0x00}, {0x06, 0x01, 0x41, 0x00},
		{0x0a, 0x01, 0x41, 0x00}, {0x0f, 0x01, 0x41, 0x00},
		{0x18, 0x01, 0x41, 0x00}, {0x1f, 0x01, 0x41, 0x00},
		{0x29, 0x01, 0x41, 0x00}, {0x38, 0x01, 0x41, 0x01}
	],
	[
		{0x03, 0x01, 0x5f, 0x00}, {0x06, 0x01, 0x5f, 0x00},
		{0x0a, 0x01, 0x5f, 0x00}, {0x0f, 0x01, 0x5f, 0x00},
		{0x18, 0x01, 0x5f, 0x00}, {0x1f, 0x01, 0x5f, 0x00},
		{0x29, 0x01, 0x5f, 0x00}, {0x38, 0x01, 0x5f, 0x01},
		{0x03, 0x01, 0x62, 0x00}, {0x06, 0x01, 0x62, 0x00},
		{0x0a, 0x01, 0x62, 0x00}, {0x0f, 0x01, 0x62, 0x00},
		{0x18, 0x01, 0x62, 0x00}, {0x1f, 0x01, 0x62, 0x00},
		{0x29, 0x01, 0x62, 0x00}, {0x38, 0x01, 0x62, 0x01}
	],
	[
		{0x02, 0x01, 0x64, 0x00}, {0x09, 0x01, 0x64, 0x00},
		{0x17, 0x01, 0x64, 0x00}, {0x28, 0x01, 0x64, 0x01},
		{0x02, 0x01, 0x66, 0x00}, {0x09, 0x01, 0x66, 0x00},
		{0x17, 0x01, 0x66, 0x00}, {0x28, 0x01, 0x66, 0x01},
		{0x02, 0x01, 0x67, 0x00}, {0x09, 0x01, 0x67, 0x00},
		{0x17, 0x01, 0x67, 0x00}, {0x28, 0x01, 0x67, 0x01},
		{0x02, 0x01, 0x68, 0x00}, {0x09, 0x01, 0x68, 0x00},
		{0x17, 0x01, 0x68, 0x00}, {0x28, 0x01, 0x68, 0x01}
	],
	[
		{0x03, 0x01, 0x64, 0x00}, {0x06, 0x01, 0x64, 0x00},
		{0x0a, 0x01, 0x64, 0x00}, {0x0f, 0x01, 0x64, 0x00},
		{0x18, 0x01, 0x64, 0x00}, {0x1f, 0x01, 0x64, 0x00},
		{0x29, 0x01, 0x64, 0x00}, {0x38, 0x01, 0x64, 0x01},
		{0x03, 0x01, 0x66, 0x00}, {0x06, 0x01, 0x66, 0x00},
		{0x0a, 0x01, 0x66, 0x00}, {0x0f, 0x01, 0x66, 0x00},
		{0x18, 0x01, 0x66, 0x00}, {0x1f, 0x01, 0x66, 0x00},
		{0x29, 0x01, 0x66, 0x00}, {0x38, 0x01, 0x66, 0x01}
	],
	/* 30 */
	[
		{0x03, 0x01, 0x67, 0x00}, {0x06, 0x01, 0x67, 0x00},
		{0x0a, 0x01, 0x67, 0x00}, {0x0f, 0x01, 0x67, 0x00},
		{0x18, 0x01, 0x67, 0x00}, {0x1f, 0x01, 0x67, 0x00},
		{0x29, 0x01, 0x67, 0x00}, {0x38, 0x01, 0x67, 0x01},
		{0x03, 0x01, 0x68, 0x00}, {0x06, 0x01, 0x68, 0x00},
		{0x0a, 0x01, 0x68, 0x00}, {0x0f, 0x01, 0x68, 0x00},
		{0x18, 0x01, 0x68, 0x00}, {0x1f, 0x01, 0x68, 0x00},
		{0x29, 0x01, 0x68, 0x00}, {0x38, 0x01, 0x68, 0x01}
	],
	[
		{0x01, 0x01, 0x6c, 0x00}, {0x16, 0x01, 0x6c, 0x01},
		{0x01, 0x01, 0x6d, 0x00}, {0x16, 0x01, 0x6d, 0x01},
		{0x01, 0x01, 0x6e, 0x00}, {0x16, 0x01, 0x6e, 0x01},
		{0x01, 0x01, 0x70, 0x00}, {0x16, 0x01, 0x70, 0x01},
		{0x01, 0x01, 0x72, 0x00}, {0x16, 0x01, 0x72, 0x01},
		{0x01, 0x01, 0x75, 0x00}, {0x16, 0x01, 0x75, 0x01},
		{0x00, 0x01, 0x3a, 0x01}, {0x00, 0x01, 0x42, 0x01},
		{0x00, 0x01, 0x43, 0x01}, {0x00, 0x01, 0x44, 0x01}
	],
	[
		{0x02, 0x01, 0x6c, 0x00}, {0x09, 0x01, 0x6c, 0x00},
		{0x17, 0x01, 0x6c, 0x00}, {0x28, 0x01, 0x6c, 0x01},
		{0x02, 0x01, 0x6d, 0x00}, {0x09, 0x01, 0x6d, 0x00},
		{0x17, 0x01, 0x6d, 0x00}, {0x28, 0x01, 0x6d, 0x01},
		{0x02, 0x01, 0x6e, 0x00}, {0x09, 0x01, 0x6e, 0x00},
		{0x17, 0x01, 0x6e, 0x00}, {0x28, 0x01, 0x6e, 0x01},
		{0x02, 0x01, 0x70, 0x00}, {0x09, 0x01, 0x70, 0x00},
		{0x17, 0x01, 0x70, 0x00}, {0x28, 0x01, 0x70, 0x01}
	],
	[
		{0x03, 0x01, 0x6c, 0x00}, {0x06, 0x01, 0x6c, 0x00},
		{0x0a, 0x01, 0x6c, 0x00}, {0x0f, 0x01, 0x6c, 0x00},
		{0x18, 0x01, 0x6c, 0x00}, {0x1f, 0x01, 0x6c, 0x00},
		{0x29, 0x01, 0x6c, 0x00}, {0x38, 0x01, 0x6c, 0x01},
		{0x03, 0x01, 0x6d, 0x00}, {0x06, 0x01, 0x6d, 0x00},
		{0x0a, 0x01, 0x6d, 0x00}, {0x0f, 0x01, 0x6d, 0x00},
		{0x18, 0x01, 0x6d, 0x00}, {0x1f, 0x01, 0x6d, 0x00},
		{0x29, 0x01, 0x6d, 0x00}, {0x38, 0x01, 0x6d, 0x01}
	],
	[
		{0x03, 0x01, 0x6e, 0x00}, {0x06, 0x01, 0x6e, 0x00},
		{0x0a, 0x01, 0x6e, 0x00}, {0x0f, 0x01, 0x6e, 0x00},
		{0x18, 0x01, 0x6e, 0x00}, {0x1f, 0x01, 0x6e, 0x00},
		{0x29, 0x01, 0x6e, 0x00}, {0x38, 0x01, 0x6e, 0x01},
		{0x03, 0x01, 0x70, 0x00}, {0x06, 0x01, 0x70, 0x00},
		{0x0a, 0x01, 0x70, 0x00}, {0x0f, 0x01, 0x70, 0x00},
		{0x18, 0x01, 0x70, 0x00}, {0x1f, 0x01, 0x70, 0x00},
		{0x29, 0x01, 0x70, 0x00}, {0x38, 0x01, 0x70, 0x01}
	],
	/* 35 */
	[
		{0x02, 0x01, 0x72, 0x00}, {0x09, 0x01, 0x72, 0x00},
		{0x17, 0x01, 0x72, 0x00}, {0x28, 0x01, 0x72, 0x01},
		{0x02, 0x01, 0x75, 0x00}, {0x09, 0x01, 0x75, 0x00},
		{0x17, 0x01, 0x75, 0x00}, {0x28, 0x01, 0x75, 0x01},
		{0x01, 0x01, 0x3a, 0x00}, {0x16, 0x01, 0x3a, 0x01},
		{0x01, 0x01, 0x42, 0x00}, {0x16, 0x01, 0x42, 0x01},
		{0x01, 0x01, 0x43, 0x00}, {0x16, 0x01, 0x43, 0x01},
		{0x01, 0x01, 0x44, 0x00}, {0x16, 0x01, 0x44, 0x01}
	],
	[
		{0x03, 0x01, 0x72, 0x00}, {0x06, 0x01, 0x72, 0x00},
		{0x0a, 0x01, 0x72, 0x00}, {0x0f, 0x01, 0x72, 0x00},
		{0x18, 0x01, 0x72, 0x00}, {0x1f, 0x01, 0x72, 0x00},
		{0x29, 0x01, 0x72, 0x00}, {0x38, 0x01, 0x72, 0x01},
		{0x03, 0x01, 0x75, 0x00}, {0x06, 0x01, 0x75, 0x00},
		{0x0a, 0x01, 0x75, 0x00}, {0x0f, 0x01, 0x75, 0x00},
		{0x18, 0x01, 0x75, 0x00}, {0x1f, 0x01, 0x75, 0x00},
		{0x29, 0x01, 0x75, 0x00}, {0x38, 0x01, 0x75, 0x01}
	],
	[
		{0x02, 0x01, 0x3a, 0x00}, {0x09, 0x01, 0x3a, 0x00},
		{0x17, 0x01, 0x3a, 0x00}, {0x28, 0x01, 0x3a, 0x01},
		{0x02, 0x01, 0x42, 0x00}, {0x09, 0x01, 0x42, 0x00},
		{0x17, 0x01, 0x42, 0x00}, {0x28, 0x01, 0x42, 0x01},
		{0x02, 0x01, 0x43, 0x00}, {0x09, 0x01, 0x43, 0x00},
		{0x17, 0x01, 0x43, 0x00}, {0x28, 0x01, 0x43, 0x01},
		{0x02, 0x01, 0x44, 0x00}, {0x09, 0x01, 0x44, 0x00},
		{0x17, 0x01, 0x44, 0x00}, {0x28, 0x01, 0x44, 0x01}
	],
	[
		{0x03, 0x01, 0x3a, 0x00}, {0x06, 0x01, 0x3a, 0x00},
		{0x0a, 0x01, 0x3a, 0x00}, {0x0f, 0x01, 0x3a, 0x00},
		{0x18, 0x01, 0x3a, 0x00}, {0x1f, 0x01, 0x3a, 0x00},
		{0x29, 0x01, 0x3a, 0x00}, {0x38, 0x01, 0x3a, 0x01},
		{0x03, 0x01, 0x42, 0x00}, {0x06, 0x01, 0x42, 0x00},
		{0x0a, 0x01, 0x42, 0x00}, {0x0f, 0x01, 0x42, 0x00},
		{0x18, 0x01, 0x42, 0x00}, {0x1f, 0x01, 0x42, 0x00},
		{0x29, 0x01, 0x42, 0x00}, {0x38, 0x01, 0x42, 0x01}
	],
	[
		{0x03, 0x01, 0x43, 0x00}, {0x06, 0x01, 0x43, 0x00},
		{0x0a, 0x01, 0x43, 0x00}, {0x0f, 0x01, 0x43, 0x00},
		{0x18, 0x01, 0x43, 0x00}, {0x1f, 0x01, 0x43, 0x00},
		{0x29, 0x01, 0x43, 0x00}, {0x38, 0x01, 0x43, 0x01},
		{0x03, 0x01, 0x44, 0x00}, {0x06, 0x01, 0x44, 0x00},
		{0x0a, 0x01, 0x44, 0x00}, {0x0f, 0x01, 0x44, 0x00},
		{0x18, 0x01, 0x44, 0x00}, {0x1f, 0x01, 0x44, 0x00},
		{0x29, 0x01, 0x44, 0x00}, {0x38, 0x01, 0x44, 0x01}
	],
	/* 40 */
	[
		{0x2c, 0x00, 0x00, 0x00}, {0x2d, 0x00, 0x00, 0x00},
		{0x2f, 0x00, 0x00, 0x00}, {0x30, 0x00, 0x00, 0x00},
		{0x33, 0x00, 0x00, 0x00}, {0x34, 0x00, 0x00, 0x00},
		{0x36, 0x00, 0x00, 0x00}, {0x37, 0x00, 0x00, 0x00},
		{0x3b, 0x00, 0x00, 0x00}, {0x3c, 0x00, 0x00, 0x00},
		{0x3e, 0x00, 0x00, 0x00}, {0x3f, 0x00, 0x00, 0x00},
		{0x42, 0x00, 0x00, 0x00}, {0x43, 0x00, 0x00, 0x00},
		{0x45, 0x00, 0x00, 0x00}, {0x48, 0x00, 0x00, 0x01}
	],
	[
		{0x00, 0x01, 0x45, 0x01}, {0x00, 0x01, 0x46, 0x01},
		{0x00, 0x01, 0x47, 0x01}, {0x00, 0x01, 0x48, 0x01},
		{0x00, 0x01, 0x49, 0x01}, {0x00, 0x01, 0x4a, 0x01},
		{0x00, 0x01, 0x4b, 0x01}, {0x00, 0x01, 0x4c, 0x01},
		{0x00, 0x01, 0x4d, 0x01}, {0x00, 0x01, 0x4e, 0x01},
		{0x00, 0x01, 0x4f, 0x01}, {0x00, 0x01, 0x50, 0x01},
		{0x00, 0x01, 0x51, 0x01}, {0x00, 0x01, 0x52, 0x01},
		{0x00, 0x01, 0x53, 0x01}, {0x00, 0x01, 0x54, 0x01}
	],
	[
		{0x01, 0x01, 0x45, 0x00}, {0x16, 0x01, 0x45, 0x01},
		{0x01, 0x01, 0x46, 0x00}, {0x16, 0x01, 0x46, 0x01},
		{0x01, 0x01, 0x47, 0x00}, {0x16, 0x01, 0x47, 0x01},
		{0x01, 0x01, 0x48, 0x00}, {0x16, 0x01, 0x48, 0x01},
		{0x01, 0x01, 0x49, 0x00}, {0x16, 0x01, 0x49, 0x01},
		{0x01, 0x01, 0x4a, 0x00}, {0x16, 0x01, 0x4a, 0x01},
		{0x01, 0x01, 0x4b, 0x00}, {0x16, 0x01, 0x4b, 0x01},
		{0x01, 0x01, 0x4c, 0x00}, {0x16, 0x01, 0x4c, 0x01}
	],
	[
		{0x02, 0x01, 0x45, 0x00}, {0x09, 0x01, 0x45, 0x00},
		{0x17, 0x01, 0x45, 0x00}, {0x28, 0x01, 0x45, 0x01},
		{0x02, 0x01, 0x46, 0x00}, {0x09, 0x01, 0x46, 0x00},
		{0x17, 0x01, 0x46, 0x00}, {0x28, 0x01, 0x46, 0x01},
		{0x02, 0x01, 0x47, 0x00}, {0x09, 0x01, 0x47, 0x00},
		{0x17, 0x01, 0x47, 0x00}, {0x28, 0x01, 0x47, 0x01},
		{0x02, 0x01, 0x48, 0x00}, {0x09, 0x01, 0x48, 0x00},
		{0x17, 0x01, 0x48, 0x00}, {0x28, 0x01, 0x48, 0x01}
	],
	[
		{0x03, 0x01, 0x45, 0x00}, {0x06, 0x01, 0x45, 0x00},
		{0x0a, 0x01, 0x45, 0x00}, {0x0f, 0x01, 0x45, 0x00},
		{0x18, 0x01, 0x45, 0x00}, {0x1f, 0x01, 0x45, 0x00},
		{0x29, 0x01, 0x45, 0x00}, {0x38, 0x01, 0x45, 0x01},
		{0x03, 0x01, 0x46, 0x00}, {0x06, 0x01, 0x46, 0x00},
		{0x0a, 0x01, 0x46, 0x00}, {0x0f, 0x01, 0x46, 0x00},
		{0x18, 0x01, 0x46, 0x00}, {0x1f, 0x01, 0x46, 0x00},
		{0x29, 0x01, 0x46, 0x00}, {0x38, 0x01, 0x46, 0x01}
	],
	/* 45 */
	[
		{0x03, 0x01, 0x47, 0x00}, {0x06, 0x01, 0x47, 0x00},
		{0x0a, 0x01, 0x47, 0x00}, {0x0f, 0x01, 0x47, 0x00},
		{0x18, 0x01, 0x47, 0x00}, {0x1f, 0x01, 0x47, 0x00},
		{0x29, 0x01, 0x47, 0x00}, {0x38, 0x01, 0x47, 0x01},
		{0x03, 0x01, 0x48, 0x00}, {0x06, 0x01, 0x48, 0x00},
		{0x0a, 0x01, 0x48, 0x00}, {0x0f, 0x01, 0x48, 0x00},
		{0x18, 0x01, 0x48, 0x00}, {0x1f, 0x01, 0x48, 0x00},
		{0x29, 0x01, 0x48, 0x00}, {0x38, 0x01, 0x48, 0x01}
	],
	[
		{0x02, 0x01, 0x49, 0x00}, {0x09, 0x01, 0x49, 0x00},
		{0x17, 0x01, 0x49, 0x00}, {0x28, 0x01, 0x49, 0x01},
		{0x02, 0x01, 0x4a, 0x00}, {0x09, 0x01, 0x4a, 0x00},
		{0x17, 0x01, 0x4a, 0x00}, {0x28, 0x01, 0x4a, 0x01},
		{0x02, 0x01, 0x4b, 0x00}, {0x09, 0x01, 0x4b, 0x00},
		{0x17, 0x01, 0x4b, 0x00}, {0x28, 0x01, 0x4b, 0x01},
		{0x02, 0x01, 0x4c, 0x00}, {0x09, 0x01, 0x4c, 0x00},
		{0x17, 0x01, 0x4c, 0x00}, {0x28, 0x01, 0x4c, 0x01}
	],
	[
		{0x03, 0x01, 0x49, 0x00}, {0x06, 0x01, 0x49, 0x00},
		{0x0a, 0x01, 0x49, 0x00}, {0x0f, 0x01, 0x49, 0x00},
		{0x18, 0x01, 0x49, 0x00}, {0x1f, 0x01, 0x49, 0x00},
		{0x29, 0x01, 0x49, 0x00}, {0x38, 0x01, 0x49, 0x01},
		{0x03, 0x01, 0x4a, 0x00}, {0x06, 0x01, 0x4a, 0x00},
		{0x0a, 0x01, 0x4a, 0x00}, {0x0f, 0x01, 0x4a, 0x00},
		{0x18, 0x01, 0x4a, 0x00}, {0x1f, 0x01, 0x4a, 0x00},
		{0x29, 0x01, 0x4a, 0x00}, {0x38, 0x01, 0x4a, 0x01}
	],
	[
		{0x03, 0x01, 0x4b, 0x00}, {0x06, 0x01, 0x4b, 0x00},
		{0x0a, 0x01, 0x4b, 0x00}, {0x0f, 0x01, 0x4b, 0x00},
		{0x18, 0x01, 0x4b, 0x00}, {0x1f, 0x01, 0x4b, 0x00},
		{0x29, 0x01, 0x4b, 0x00}, {0x38, 0x01, 0x4b, 0x01},
		{0x03, 0x01, 0x4c, 0x00}, {0x06, 0x01, 0x4c, 0x00},
		{0x0a, 0x01, 0x4c, 0x00}, {0x0f, 0x01, 0x4c, 0x00},
		{0x18, 0x01, 0x4c, 0x00}, {0x1f, 0x01, 0x4c, 0x00},
		{0x29, 0x01, 0x4c, 0x00}, {0x38, 0x01, 0x4c, 0x01}
	],
	[
		{0x01, 0x01, 0x4d, 0x00}, {0x16, 0x01, 0x4d, 0x01},
		{0x01, 0x01, 0x4e, 0x00}, {0x16, 0x01, 0x4e, 0x01},
		{0x01, 0x01, 0x4f, 0x00}, {0x16, 0x01, 0x4f, 0x01},
		{0x01, 0x01, 0x50, 0x00}, {0x16, 0x01, 0x50, 0x01},
		{0x01, 0x01, 0x51, 0x00}, {0x16, 0x01, 0x51, 0x01},
		{0x01, 0x01, 0x52, 0x00}, {0x16, 0x01, 0x52, 0x01},
		{0x01, 0x01, 0x53, 0x00}, {0x16, 0x01, 0x53, 0x01},
		{0x01, 0x01, 0x54, 0x00}, {0x16, 0x01, 0x54, 0x01}
	],
	/* 50 */
	[
		{0x02, 0x01, 0x4d, 0x00}, {0x09, 0x01, 0x4d, 0x00},
		{0x17, 0x01, 0x4d, 0x00}, {0x28, 0x01, 0x4d, 0x01},
		{0x02, 0x01, 0x4e, 0x00}, {0x09, 0x01, 0x4e, 0x00},
		{0x17, 0x01, 0x4e, 0x00}, {0x28, 0x01, 0x4e, 0x01},
		{0x02, 0x01, 0x4f, 0x00}, {0x09, 0x01, 0x4f, 0x00},
		{0x17, 0x01, 0x4f, 0x00}, {0x28, 0x01, 0x4f, 0x01},
		{0x02, 0x01, 0x50, 0x00}, {0x09, 0x01, 0x50, 0x00},
		{0x17, 0x01, 0x50, 0x00}, {0x28, 0x01, 0x50, 0x01}
	],
	[
		{0x03, 0x01, 0x4d, 0x00}, {0x06, 0x01, 0x4d, 0x00},
		{0x0a, 0x01, 0x4d, 0x00}, {0x0f, 0x01, 0x4d, 0x00},
		{0x18, 0x01, 0x4d, 0x00}, {0x1f, 0x01, 0x4d, 0x00},
		{0x29, 0x01, 0x4d, 0x00}, {0x38, 0x01, 0x4d, 0x01},
		{0x03, 0x01, 0x4e, 0x00}, {0x06, 0x01, 0x4e, 0x00},
		{0x0a, 0x01, 0x4e, 0x00}, {0x0f, 0x01, 0x4e, 0x00},
		{0x18, 0x01, 0x4e, 0x00}, {0x1f, 0x01, 0x4e, 0x00},
		{0x29, 0x01, 0x4e, 0x00}, {0x38, 0x01, 0x4e, 0x01}
	],
	[
		{0x03, 0x01, 0x4f, 0x00}, {0x06, 0x01, 0x4f, 0x00},
		{0x0a, 0x01, 0x4f, 0x00}, {0x0f, 0x01, 0x4f, 0x00},
		{0x18, 0x01, 0x4f, 0x00}, {0x1f, 0x01, 0x4f, 0x00},
		{0x29, 0x01, 0x4f, 0x00}, {0x38, 0x01, 0x4f, 0x01},
		{0x03, 0x01, 0x50, 0x00}, {0x06, 0x01, 0x50, 0x00},
		{0x0a, 0x01, 0x50, 0x00}, {0x0f, 0x01, 0x50, 0x00},
		{0x18, 0x01, 0x50, 0x00}, {0x1f, 0x01, 0x50, 0x00},
		{0x29, 0x01, 0x50, 0x00}, {0x38, 0x01, 0x50, 0x01}
	],
	[
		{0x02, 0x01, 0x51, 0x00}, {0x09, 0x01, 0x51, 0x00},
		{0x17, 0x01, 0x51, 0x00}, {0x28, 0x01, 0x51, 0x01},
		{0x02, 0x01, 0x52, 0x00}, {0x09, 0x01, 0x52, 0x00},
		{0x17, 0x01, 0x52, 0x00}, {0x28, 0x01, 0x52, 0x01},
		{0x02, 0x01, 0x53, 0x00}, {0x09, 0x01, 0x53, 0x00},
		{0x17, 0x01, 0x53, 0x00}, {0x28, 0x01, 0x53, 0x01},
		{0x02, 0x01, 0x54, 0x00}, {0x09, 0x01, 0x54, 0x00},
		{0x17, 0x01, 0x54, 0x00}, {0x28, 0x01, 0x54, 0x01}
	],
	[
		{0x03, 0x01, 0x51, 0x00}, {0x06, 0x01, 0x51, 0x00},
		{0x0a, 0x01, 0x51, 0x00}, {0x0f, 0x01, 0x51, 0x00},
		{0x18, 0x01, 0x51, 0x00}, {0x1f, 0x01, 0x51, 0x00},
		{0x29, 0x01, 0x51, 0x00}, {0x38, 0x01, 0x51, 0x01},
		{0x03, 0x01, 0x52, 0x00}, {0x06, 0x01, 0x52, 0x00},
		{0x0a, 0x01, 0x52, 0x00}, {0x0f, 0x01, 0x52, 0x00},
		{0x18, 0x01, 0x52, 0x00}, {0x1f, 0x01, 0x52, 0x00},
		{0x29, 0x01, 0x52, 0x00}, {0x38, 0x01, 0x52, 0x01}
	],
	/* 55 */
	[
		{0x03, 0x01, 0x53, 0x00}, {0x06, 0x01, 0x53, 0x00},
		{0x0a, 0x01, 0x53, 0x00}, {0x0f, 0x01, 0x53, 0x00},
		{0x18, 0x01, 0x53, 0x00}, {0x1f, 0x01, 0x53, 0x00},
		{0x29, 0x01, 0x53, 0x00}, {0x38, 0x01, 0x53, 0x01},
		{0x03, 0x01, 0x54, 0x00}, {0x06, 0x01, 0x54, 0x00},
		{0x0a, 0x01, 0x54, 0x00}, {0x0f, 0x01, 0x54, 0x00},
		{0x18, 0x01, 0x54, 0x00}, {0x1f, 0x01, 0x54, 0x00},
		{0x29, 0x01, 0x54, 0x00}, {0x38, 0x01, 0x54, 0x01}
	],
	[
		{0x00, 0x01, 0x55, 0x01}, {0x00, 0x01, 0x56, 0x01},
		{0x00, 0x01, 0x57, 0x01}, {0x00, 0x01, 0x59, 0x01},
		{0x00, 0x01, 0x6a, 0x01}, {0x00, 0x01, 0x6b, 0x01},
		{0x00, 0x01, 0x71, 0x01}, {0x00, 0x01, 0x76, 0x01},
		{0x00, 0x01, 0x77, 0x01}, {0x00, 0x01, 0x78, 0x01},
		{0x00, 0x01, 0x79, 0x01}, {0x00, 0x01, 0x7a, 0x01},
		{0x46, 0x00, 0x00, 0x00}, {0x47, 0x00, 0x00, 0x00},
		{0x49, 0x00, 0x00, 0x00}, {0x4a, 0x00, 0x00, 0x01}
	],
	[
		{0x01, 0x01, 0x55, 0x00}, {0x16, 0x01, 0x55, 0x01},
		{0x01, 0x01, 0x56, 0x00}, {0x16, 0x01, 0x56, 0x01},
		{0x01, 0x01, 0x57, 0x00}, {0x16, 0x01, 0x57, 0x01},
		{0x01, 0x01, 0x59, 0x00}, {0x16, 0x01, 0x59, 0x01},
		{0x01, 0x01, 0x6a, 0x00}, {0x16, 0x01, 0x6a, 0x01},
		{0x01, 0x01, 0x6b, 0x00}, {0x16, 0x01, 0x6b, 0x01},
		{0x01, 0x01, 0x71, 0x00}, {0x16, 0x01, 0x71, 0x01},
		{0x01, 0x01, 0x76, 0x00}, {0x16, 0x01, 0x76, 0x01}
	],
	[
		{0x02, 0x01, 0x55, 0x00}, {0x09, 0x01, 0x55, 0x00},
		{0x17, 0x01, 0x55, 0x00}, {0x28, 0x01, 0x55, 0x01},
		{0x02, 0x01, 0x56, 0x00}, {0x09, 0x01, 0x56, 0x00},
		{0x17, 0x01, 0x56, 0x00}, {0x28, 0x01, 0x56, 0x01},
		{0x02, 0x01, 0x57, 0x00}, {0x09, 0x01, 0x57, 0x00},
		{0x17, 0x01, 0x57, 0x00}, {0x28, 0x01, 0x57, 0x01},
		{0x02, 0x01, 0x59, 0x00}, {0x09, 0x01, 0x59, 0x00},
		{0x17, 0x01, 0x59, 0x00}, {0x28, 0x01, 0x59, 0x01}
	],
	[
		{0x03, 0x01, 0x55, 0x00}, {0x06, 0x01, 0x55, 0x00},
		{0x0a, 0x01, 0x55, 0x00}, {0x0f, 0x01, 0x55, 0x00},
		{0x18, 0x01, 0x55, 0x00}, {0x1f, 0x01, 0x55, 0x00},
		{0x29, 0x01, 0x55, 0x00}, {0x38, 0x01, 0x55, 0x01},
		{0x03, 0x01, 0x56, 0x00}, {0x06, 0x01, 0x56, 0x00},
		{0x0a, 0x01, 0x56, 0x00}, {0x0f, 0x01, 0x56, 0x00},
		{0x18, 0x01, 0x56, 0x00}, {0x1f, 0x01, 0x56, 0x00},
		{0x29, 0x01, 0x56, 0x00}, {0x38, 0x01, 0x56, 0x01}
	],
	/* 60 */
	[
		{0x03, 0x01, 0x57, 0x00}, {0x06, 0x01, 0x57, 0x00},
		{0x0a, 0x01, 0x57, 0x00}, {0x0f, 0x01, 0x57, 0x00},
		{0x18, 0x01, 0x57, 0x00}, {0x1f, 0x01, 0x57, 0x00},
		{0x29, 0x01, 0x57, 0x00}, {0x38, 0x01, 0x57, 0x01},
		{0x03, 0x01, 0x59, 0x00}, {0x06, 0x01, 0x59, 0x00},
		{0x0a, 0x01, 0x59, 0x00}, {0x0f, 0x01, 0x59, 0x00},
		{0x18, 0x01, 0x59, 0x00}, {0x1f, 0x01, 0x59, 0x00},
		{0x29, 0x01, 0x59, 0x00}, {0x38, 0x01, 0x59, 0x01}
	],
	[
		{0x02, 0x01, 0x6a, 0x00}, {0x09, 0x01, 0x6a, 0x00},
		{0x17, 0x01, 0x6a, 0x00}, {0x28, 0x01, 0x6a, 0x01},
		{0x02, 0x01, 0x6b, 0x00}, {0x09, 0x01, 0x6b, 0x00},
		{0x17, 0x01, 0x6b, 0x00}, {0x28, 0x01, 0x6b, 0x01},
		{0x02, 0x01, 0x71, 0x00}, {0x09, 0x01, 0x71, 0x00},
		{0x17, 0x01, 0x71, 0x00}, {0x28, 0x01, 0x71, 0x01},
		{0x02, 0x01, 0x76, 0x00}, {0x09, 0x01, 0x76, 0x00},
		{0x17, 0x01, 0x76, 0x00}, {0x28, 0x01, 0x76, 0x01}
	],
	[
		{0x03, 0x01, 0x6a, 0x00}, {0x06, 0x01, 0x6a, 0x00},
		{0x0a, 0x01, 0x6a, 0x00}, {0x0f, 0x01, 0x6a, 0x00},
		{0x18, 0x01, 0x6a, 0x00}, {0x1f, 0x01, 0x6a, 0x00},
		{0x29, 0x01, 0x6a, 0x00}, {0x38, 0x01, 0x6a, 0x01},
		{0x03, 0x01, 0x6b, 0x00}, {0x06, 0x01, 0x6b, 0x00},
		{0x0a, 0x01, 0x6b, 0x00}, {0x0f, 0x01, 0x6b, 0x00},
		{0x18, 0x01, 0x6b, 0x00}, {0x1f, 0x01, 0x6b, 0x00},
		{0x29, 0x01, 0x6b, 0x00}, {0x38, 0x01, 0x6b, 0x01}
	],
	[
		{0x03, 0x01, 0x71, 0x00}, {0x06, 0x01, 0x71, 0x00},
		{0x0a, 0x01, 0x71, 0x00}, {0x0f, 0x01, 0x71, 0x00},
		{0x18, 0x01, 0x71, 0x00}, {0x1f, 0x01, 0x71, 0x00},
		{0x29, 0x01, 0x71, 0x00}, {0x38, 0x01, 0x71, 0x01},
		{0x03, 0x01, 0x76, 0x00}, {0x06, 0x01, 0x76, 0x00},
		{0x0a, 0x01, 0x76, 0x00}, {0x0f, 0x01, 0x76, 0x00},
		{0x18, 0x01, 0x76, 0x00}, {0x1f, 0x01, 0x76, 0x00},
		{0x29, 0x01, 0x76, 0x00}, {0x38, 0x01, 0x76, 0x01}
	],
	[
		{0x01, 0x01, 0x77, 0x00}, {0x16, 0x01, 0x77, 0x01},
		{0x01, 0x01, 0x78, 0x00}, {0x16, 0x01, 0x78, 0x01},
		{0x01, 0x01, 0x79, 0x00}, {0x16, 0x01, 0x79, 0x01},
		{0x01, 0x01, 0x7a, 0x00}, {0x16, 0x01, 0x7a, 0x01},
		{0x00, 0x01, 0x26, 0x01}, {0x00, 0x01, 0x2a, 0x01},
		{0x00, 0x01, 0x2c, 0x01}, {0x00, 0x01, 0x3b, 0x01},
		{0x00, 0x01, 0x58, 0x01}, {0x00, 0x01, 0x5a, 0x01},
		{0x4b, 0x00, 0x00, 0x00}, {0x4e, 0x00, 0x00, 0x01}
	],
	/* 65 */
	[
		{0x02, 0x01, 0x77, 0x00}, {0x09, 0x01, 0x77, 0x00},
		{0x17, 0x01, 0x77, 0x00}, {0x28, 0x01, 0x77, 0x01},
		{0x02, 0x01, 0x78, 0x00}, {0x09, 0x01, 0x78, 0x00},
		{0x17, 0x01, 0x78, 0x00}, {0x28, 0x01, 0x78, 0x01},
		{0x02, 0x01, 0x79, 0x00}, {0x09, 0x01, 0x79, 0x00},
		{0x17, 0x01, 0x79, 0x00}, {0x28, 0x01, 0x79, 0x01},
		{0x02, 0x01, 0x7a, 0x00}, {0x09, 0x01, 0x7a, 0x00},
		{0x17, 0x01, 0x7a, 0x00}, {0x28, 0x01, 0x7a, 0x01}
	],
	[
		{0x03, 0x01, 0x77, 0x00}, {0x06, 0x01, 0x77, 0x00},
		{0x0a, 0x01, 0x77, 0x00}, {0x0f, 0x01, 0x77, 0x00},
		{0x18, 0x01, 0x77, 0x00}, {0x1f, 0x01, 0x77, 0x00},
		{0x29, 0x01, 0x77, 0x00}, {0x38, 0x01, 0x77, 0x01},
		{0x03, 0x01, 0x78, 0x00}, {0x06, 0x01, 0x78, 0x00},
		{0x0a, 0x01, 0x78, 0x00}, {0x0f, 0x01, 0x78, 0x00},
		{0x18, 0x01, 0x78, 0x00}, {0x1f, 0x01, 0x78, 0x00},
		{0x29, 0x01, 0x78, 0x00}, {0x38, 0x01, 0x78, 0x01}
	],
	[
		{0x03, 0x01, 0x79, 0x00}, {0x06, 0x01, 0x79, 0x00},
		{0x0a, 0x01, 0x79, 0x00}, {0x0f, 0x01, 0x79, 0x00},
		{0x18, 0x01, 0x79, 0x00}, {0x1f, 0x01, 0x79, 0x00},
		{0x29, 0x01, 0x79, 0x00}, {0x38, 0x01, 0x79, 0x01},
		{0x03, 0x01, 0x7a, 0x00}, {0x06, 0x01, 0x7a, 0x00},
		{0x0a, 0x01, 0x7a, 0x00}, {0x0f, 0x01, 0x7a, 0x00},
		{0x18, 0x01, 0x7a, 0x00}, {0x1f, 0x01, 0x7a, 0x00},
		{0x29, 0x01, 0x7a, 0x00}, {0x38, 0x01, 0x7a, 0x01}
	],
	[
		{0x01, 0x01, 0x26, 0x00}, {0x16, 0x01, 0x26, 0x01},
		{0x01, 0x01, 0x2a, 0x00}, {0x16, 0x01, 0x2a, 0x01},
		{0x01, 0x01, 0x2c, 0x00}, {0x16, 0x01, 0x2c, 0x01},
		{0x01, 0x01, 0x3b, 0x00}, {0x16, 0x01, 0x3b, 0x01},
		{0x01, 0x01, 0x58, 0x00}, {0x16, 0x01, 0x58, 0x01},
		{0x01, 0x01, 0x5a, 0x00}, {0x16, 0x01, 0x5a, 0x01},
		{0x4c, 0x00, 0x00, 0x00}, {0x4d, 0x00, 0x00, 0x00},
		{0x4f, 0x00, 0x00, 0x00}, {0x51, 0x00, 0x00, 0x01}
	],
	[
		{0x02, 0x01, 0x26, 0x00}, {0x09, 0x01, 0x26, 0x00},
		{0x17, 0x01, 0x26, 0x00}, {0x28, 0x01, 0x26, 0x01},
		{0x02, 0x01, 0x2a, 0x00}, {0x09, 0x01, 0x2a, 0x00},
		{0x17, 0x01, 0x2a, 0x00}, {0x28, 0x01, 0x2a, 0x01},
		{0x02, 0x01, 0x2c, 0x00}, {0x09, 0x01, 0x2c, 0x00},
		{0x17, 0x01, 0x2c, 0x00}, {0x28, 0x01, 0x2c, 0x01},
		{0x02, 0x01, 0x3b, 0x00}, {0x09, 0x01, 0x3b, 0x00},
		{0x17, 0x01, 0x3b, 0x00}, {0x28, 0x01, 0x3b, 0x01}
	],
	/* 70 */
	[
		{0x03, 0x01, 0x26, 0x00}, {0x06, 0x01, 0x26, 0x00},
		{0x0a, 0x01, 0x26, 0x00}, {0x0f, 0x01, 0x26, 0x00},
		{0x18, 0x01, 0x26, 0x00}, {0x1f, 0x01, 0x26, 0x00},
		{0x29, 0x01, 0x26, 0x00}, {0x38, 0x01, 0x26, 0x01},
		{0x03, 0x01, 0x2a, 0x00}, {0x06, 0x01, 0x2a, 0x00},
		{0x0a, 0x01, 0x2a, 0x00}, {0x0f, 0x01, 0x2a, 0x00},
		{0x18, 0x01, 0x2a, 0x00}, {0x1f, 0x01, 0x2a, 0x00},
		{0x29, 0x01, 0x2a, 0x00}, {0x38, 0x01, 0x2a, 0x01}
	],
	[
		{0x03, 0x01, 0x2c, 0x00}, {0x06, 0x01, 0x2c, 0x00},
		{0x0a, 0x01, 0x2c, 0x00}, {0x0f, 0x01, 0x2c, 0x00},
		{0x18, 0x01, 0x2c, 0x00}, {0x1f, 0x01, 0x2c, 0x00},
		{0x29, 0x01, 0x2c, 0x00}, {0x38, 0x01, 0x2c, 0x01},
		{0x03, 0x01, 0x3b, 0x00}, {0x06, 0x01, 0x3b, 0x00},
		{0x0a, 0x01, 0x3b, 0x00}, {0x0f, 0x01, 0x3b, 0x00},
		{0x18, 0x01, 0x3b, 0x00}, {0x1f, 0x01, 0x3b, 0x00},
		{0x29, 0x01, 0x3b, 0x00}, {0x38, 0x01, 0x3b, 0x01}
	],
	[
		{0x02, 0x01, 0x58, 0x00}, {0x09, 0x01, 0x58, 0x00},
		{0x17, 0x01, 0x58, 0x00}, {0x28, 0x01, 0x58, 0x01},
		{0x02, 0x01, 0x5a, 0x00}, {0x09, 0x01, 0x5a, 0x00},
		{0x17, 0x01, 0x5a, 0x00}, {0x28, 0x01, 0x5a, 0x01},
		{0x00, 0x01, 0x21, 0x01}, {0x00, 0x01, 0x22, 0x01},
		{0x00, 0x01, 0x28, 0x01}, {0x00, 0x01, 0x29, 0x01},
		{0x00, 0x01, 0x3f, 0x01}, {0x50, 0x00, 0x00, 0x00},
		{0x52, 0x00, 0x00, 0x00}, {0x54, 0x00, 0x00, 0x01}
	],
	[
		{0x03, 0x01, 0x58, 0x00}, {0x06, 0x01, 0x58, 0x00},
		{0x0a, 0x01, 0x58, 0x00}, {0x0f, 0x01, 0x58, 0x00},
		{0x18, 0x01, 0x58, 0x00}, {0x1f, 0x01, 0x58, 0x00},
		{0x29, 0x01, 0x58, 0x00}, {0x38, 0x01, 0x58, 0x01},
		{0x03, 0x01, 0x5a, 0x00}, {0x06, 0x01, 0x5a, 0x00},
		{0x0a, 0x01, 0x5a, 0x00}, {0x0f, 0x01, 0x5a, 0x00},
		{0x18, 0x01, 0x5a, 0x00}, {0x1f, 0x01, 0x5a, 0x00},
		{0x29, 0x01, 0x5a, 0x00}, {0x38, 0x01, 0x5a, 0x01}
	],
	[
		{0x01, 0x01, 0x21, 0x00}, {0x16, 0x01, 0x21, 0x01},
		{0x01, 0x01, 0x22, 0x00}, {0x16, 0x01, 0x22, 0x01},
		{0x01, 0x01, 0x28, 0x00}, {0x16, 0x01, 0x28, 0x01},
		{0x01, 0x01, 0x29, 0x00}, {0x16, 0x01, 0x29, 0x01},
		{0x01, 0x01, 0x3f, 0x00}, {0x16, 0x01, 0x3f, 0x01},
		{0x00, 0x01, 0x27, 0x01}, {0x00, 0x01, 0x2b, 0x01},
		{0x00, 0x01, 0x7c, 0x01}, {0x53, 0x00, 0x00, 0x00},
		{0x55, 0x00, 0x00, 0x00}, {0x58, 0x00, 0x00, 0x01}
	],
	/* 75 */
	[
		{0x02, 0x01, 0x21, 0x00}, {0x09, 0x01, 0x21, 0x00},
		{0x17, 0x01, 0x21, 0x00}, {0x28, 0x01, 0x21, 0x01},
		{0x02, 0x01, 0x22, 0x00}, {0x09, 0x01, 0x22, 0x00},
		{0x17, 0x01, 0x22, 0x00}, {0x28, 0x01, 0x22, 0x01},
		{0x02, 0x01, 0x28, 0x00}, {0x09, 0x01, 0x28, 0x00},
		{0x17, 0x01, 0x28, 0x00}, {0x28, 0x01, 0x28, 0x01},
		{0x02, 0x01, 0x29, 0x00}, {0x09, 0x01, 0x29, 0x00},
		{0x17, 0x01, 0x29, 0x00}, {0x28, 0x01, 0x29, 0x01}
	],
	[
		{0x03, 0x01, 0x21, 0x00}, {0x06, 0x01, 0x21, 0x00},
		{0x0a, 0x01, 0x21, 0x00}, {0x0f, 0x01, 0x21, 0x00},
		{0x18, 0x01, 0x21, 0x00}, {0x1f, 0x01, 0x21, 0x00},
		{0x29, 0x01, 0x21, 0x00}, {0x38, 0x01, 0x21, 0x01},
		{0x03, 0x01, 0x22, 0x00}, {0x06, 0x01, 0x22, 0x00},
		{0x0a, 0x01, 0x22, 0x00}, {0x0f, 0x01, 0x22, 0x00},
		{0x18, 0x01, 0x22, 0x00}, {0x1f, 0x01, 0x22, 0x00},
		{0x29, 0x01, 0x22, 0x00}, {0x38, 0x01, 0x22, 0x01}
	],
	[
		{0x03, 0x01, 0x28, 0x00}, {0x06, 0x01, 0x28, 0x00},
		{0x0a, 0x01, 0x28, 0x00}, {0x0f, 0x01, 0x28, 0x00},
		{0x18, 0x01, 0x28, 0x00}, {0x1f, 0x01, 0x28, 0x00},
		{0x29, 0x01, 0x28, 0x00}, {0x38, 0x01, 0x28, 0x01},
		{0x03, 0x01, 0x29, 0x00}, {0x06, 0x01, 0x29, 0x00},
		{0x0a, 0x01, 0x29, 0x00}, {0x0f, 0x01, 0x29, 0x00},
		{0x18, 0x01, 0x29, 0x00}, {0x1f, 0x01, 0x29, 0x00},
		{0x29, 0x01, 0x29, 0x00}, {0x38, 0x01, 0x29, 0x01}
	],
	[
		{0x02, 0x01, 0x3f, 0x00}, {0x09, 0x01, 0x3f, 0x00},
		{0x17, 0x01, 0x3f, 0x00}, {0x28, 0x01, 0x3f, 0x01},
		{0x01, 0x01, 0x27, 0x00}, {0x16, 0x01, 0x27, 0x01},
		{0x01, 0x01, 0x2b, 0x00}, {0x16, 0x01, 0x2b, 0x01},
		{0x01, 0x01, 0x7c, 0x00}, {0x16, 0x01, 0x7c, 0x01},
		{0x00, 0x01, 0x23, 0x01}, {0x00, 0x01, 0x3e, 0x01},
		{0x56, 0x00, 0x00, 0x00}, {0x57, 0x00, 0x00, 0x00},
		{0x59, 0x00, 0x00, 0x00}, {0x5a, 0x00, 0x00, 0x01}
	],
	[
		{0x03, 0x01, 0x3f, 0x00}, {0x06, 0x01, 0x3f, 0x00},
		{0x0a, 0x01, 0x3f, 0x00}, {0x0f, 0x01, 0x3f, 0x00},
		{0x18, 0x01, 0x3f, 0x00}, {0x1f, 0x01, 0x3f, 0x00},
		{0x29, 0x01, 0x3f, 0x00}, {0x38, 0x01, 0x3f, 0x01},
		{0x02, 0x01, 0x27, 0x00}, {0x09, 0x01, 0x27, 0x00},
		{0x17, 0x01, 0x27, 0x00}, {0x28, 0x01, 0x27, 0x01},
		{0x02, 0x01, 0x2b, 0x00}, {0x09, 0x01, 0x2b, 0x00},
		{0x17, 0x01, 0x2b, 0x00}, {0x28, 0x01, 0x2b, 0x01}
	],
	/* 80 */
	[
		{0x03, 0x01, 0x27, 0x00}, {0x06, 0x01, 0x27, 0x00},
		{0x0a, 0x01, 0x27, 0x00}, {0x0f, 0x01, 0x27, 0x00},
		{0x18, 0x01, 0x27, 0x00}, {0x1f, 0x01, 0x27, 0x00},
		{0x29, 0x01, 0x27, 0x00}, {0x38, 0x01, 0x27, 0x01},
		{0x03, 0x01, 0x2b, 0x00}, {0x06, 0x01, 0x2b, 0x00},
		{0x0a, 0x01, 0x2b, 0x00}, {0x0f, 0x01, 0x2b, 0x00},
		{0x18, 0x01, 0x2b, 0x00}, {0x1f, 0x01, 0x2b, 0x00},
		{0x29, 0x01, 0x2b, 0x00}, {0x38, 0x01, 0x2b, 0x01}
	],
	[
		{0x02, 0x01, 0x7c, 0x00}, {0x09, 0x01, 0x7c, 0x00},
		{0x17, 0x01, 0x7c, 0x00}, {0x28, 0x01, 0x7c, 0x01},
		{0x01, 0x01, 0x23, 0x00}, {0x16, 0x01, 0x23, 0x01},
		{0x01, 0x01, 0x3e, 0x00}, {0x16, 0x01, 0x3e, 0x01},
		{0x00, 0x01, 0x00, 0x01}, {0x00, 0x01, 0x24, 0x01},
		{0x00, 0x01, 0x40, 0x01}, {0x00, 0x01, 0x5b, 0x01},
		{0x00, 0x01, 0x5d, 0x01}, {0x00, 0x01, 0x7e, 0x01},
		{0x5b, 0x00, 0x00, 0x00}, {0x5c, 0x00, 0x00, 0x01}
	],
	[
		{0x03, 0x01, 0x7c, 0x00}, {0x06, 0x01, 0x7c, 0x00},
		{0x0a, 0x01, 0x7c, 0x00}, {0x0f, 0x01, 0x7c, 0x00},
		{0x18, 0x01, 0x7c, 0x00}, {0x1f, 0x01, 0x7c, 0x00},
		{0x29, 0x01, 0x7c, 0x00}, {0x38, 0x01, 0x7c, 0x01},
		{0x02, 0x01, 0x23, 0x00}, {0x09, 0x01, 0x23, 0x00},
		{0x17, 0x01, 0x23, 0x00}, {0x28, 0x01, 0x23, 0x01},
		{0x02, 0x01, 0x3e, 0x00}, {0x09, 0x01, 0x3e, 0x00},
		{0x17, 0x01, 0x3e, 0x00}, {0x28, 0x01, 0x3e, 0x01}
	],
	[
		{0x03, 0x01, 0x23, 0x00}, {0x06, 0x01, 0x23, 0x00},
		{0x0a, 0x01, 0x23, 0x00}, {0x0f, 0x01, 0x23, 0x00},
		{0x18, 0x01, 0x23, 0x00}, {0x1f, 0x01, 0x23, 0x00},
		{0x29, 0x01, 0x23, 0x00}, {0x38, 0x01, 0x23, 0x01},
		{0x03, 0x01, 0x3e, 0x00}, {0x06, 0x01, 0x3e, 0x00},
		{0x0a, 0x01, 0x3e, 0x00}, {0x0f, 0x01, 0x3e, 0x00},
		{0x18, 0x01, 0x3e, 0x00}, {0x1f, 0x01, 0x3e, 0x00},
		{0x29, 0x01, 0x3e, 0x00}, {0x38, 0x01, 0x3e, 0x01}
	],
	[
		{0x01, 0x01, 0x00, 0x00}, {0x16, 0x01, 0x00, 0x01},
		{0x01, 0x01, 0x24, 0x00}, {0x16, 0x01, 0x24, 0x01},
		{0x01, 0x01, 0x40, 0x00}, {0x16, 0x01, 0x40, 0x01},
		{0x01, 0x01, 0x5b, 0x00}, {0x16, 0x01, 0x5b, 0x01},
		{0x01, 0x01, 0x5d, 0x00}, {0x16, 0x01, 0x5d, 0x01},
		{0x01, 0x01, 0x7e, 0x00}, {0x16, 0x01, 0x7e, 0x01},
		{0x00, 0x01, 0x5e, 0x01}, {0x00, 0x01, 0x7d, 0x01},
		{0x5d, 0x00, 0x00, 0x00}, {0x5e, 0x00, 0x00, 0x01}
	],
	/* 85 */
	[
		{0x02, 0x01, 0x00, 0x00}, {0x09, 0x01, 0x00, 0x00},
		{0x17, 0x01, 0x00, 0x00}, {0x28, 0x01, 0x00, 0x01},
		{0x02, 0x01, 0x24, 0x00}, {0x09, 0x01, 0x24, 0x00},
		{0x17, 0x01, 0x24, 0x00}, {0x28, 0x01, 0x24, 0x01},
		{0x02, 0x01, 0x40, 0x00}, {0x09, 0x01, 0x40, 0x00},
		{0x17, 0x01, 0x40, 0x00}, {0x28, 0x01, 0x40, 0x01},
		{0x02, 0x01, 0x5b, 0x00}, {0x09, 0x01, 0x5b, 0x00},
		{0x17, 0x01, 0x5b, 0x00}, {0x28, 0x01, 0x5b, 0x01}
	],
	[
		{0x03, 0x01, 0x00, 0x00}, {0x06, 0x01, 0x00, 0x00},
		{0x0a, 0x01, 0x00, 0x00}, {0x0f, 0x01, 0x00, 0x00},
		{0x18, 0x01, 0x00, 0x00}, {0x1f, 0x01, 0x00, 0x00},
		{0x29, 0x01, 0x00, 0x00}, {0x38, 0x01, 0x00, 0x01},
		{0x03, 0x01, 0x24, 0x00}, {0x06, 0x01, 0x24, 0x00},
		{0x0a, 0x01, 0x24, 0x00}, {0x0f, 0x01, 0x24, 0x00},
		{0x18, 0x01, 0x24, 0x00}, {0x1f, 0x01, 0x24, 0x00},
		{0x29, 0x01, 0x24, 0x00}, {0x38, 0x01, 0x24, 0x01}
	],
	[
		{0x03, 0x01, 0x40, 0x00}, {0x06, 0x01, 0x40, 0x00},
		{0x0a, 0x01, 0x40, 0x00}, {0x0f, 0x01, 0x40, 0x00},
		{0x18, 0x01, 0x40, 0x00}, {0x1f, 0x01, 0x40, 0x00},
		{0x29, 0x01, 0x40, 0x00}, {0x38, 0x01, 0x40, 0x01},
		{0x03, 0x01, 0x5b, 0x00}, {0x06, 0x01, 0x5b, 0x00},
		{0x0a, 0x01, 0x5b, 0x00}, {0x0f, 0x01, 0x5b, 0x00},
		{0x18, 0x01, 0x5b, 0x00}, {0x1f, 0x01, 0x5b, 0x00},
		{0x29, 0x01, 0x5b, 0x00}, {0x38, 0x01, 0x5b, 0x01}
	],
	[
		{0x02, 0x01, 0x5d, 0x00}, {0x09, 0x01, 0x5d, 0x00},
		{0x17, 0x01, 0x5d, 0x00}, {0x28, 0x01, 0x5d, 0x01},
		{0x02, 0x01, 0x7e, 0x00}, {0x09, 0x01, 0x7e, 0x00},
		{0x17, 0x01, 0x7e, 0x00}, {0x28, 0x01, 0x7e, 0x01},
		{0x01, 0x01, 0x5e, 0x00}, {0x16, 0x01, 0x5e, 0x01},
		{0x01, 0x01, 0x7d, 0x00}, {0x16, 0x01, 0x7d, 0x01},
		{0x00, 0x01, 0x3c, 0x01}, {0x00, 0x01, 0x60, 0x01},
		{0x00, 0x01, 0x7b, 0x01}, {0x5f, 0x00, 0x00, 0x01}
	],
	[
		{0x03, 0x01, 0x5d, 0x00}, {0x06, 0x01, 0x5d, 0x00},
		{0x0a, 0x01, 0x5d, 0x00}, {0x0f, 0x01, 0x5d, 0x00},
		{0x18, 0x01, 0x5d, 0x00}, {0x1f, 0x01, 0x5d, 0x00},
		{0x29, 0x01, 0x5d, 0x00}, {0x38, 0x01, 0x5d, 0x01},
		{0x03, 0x01, 0x7e, 0x00}, {0x06, 0x01, 0x7e, 0x00},
		{0x0a, 0x01, 0x7e, 0x00}, {0x0f, 0x01, 0x7e, 0x00},
		{0x18, 0x01, 0x7e, 0x00}, {0x1f, 0x01, 0x7e, 0x00},
		{0x29, 0x01, 0x7e, 0x00}, {0x38, 0x01, 0x7e, 0x01}
	],
	/* 90 */
	[
		{0x02, 0x01, 0x5e, 0x00}, {0x09, 0x01, 0x5e, 0x00},
		{0x17, 0x01, 0x5e, 0x00}, {0x28, 0x01, 0x5e, 0x01},
		{0x02, 0x01, 0x7d, 0x00}, {0x09, 0x01, 0x7d, 0x00},
		{0x17, 0x01, 0x7d, 0x00}, {0x28, 0x01, 0x7d, 0x01},
		{0x01, 0x01, 0x3c, 0x00}, {0x16, 0x01, 0x3c, 0x01},
		{0x01, 0x01, 0x60, 0x00}, {0x16, 0x01, 0x60, 0x01},
		{0x01, 0x01, 0x7b, 0x00}, {0x16, 0x01, 0x7b, 0x01},
		{0x60, 0x00, 0x00, 0x00}, {0x6e, 0x00, 0x00, 0x01}
	],
	[
		{0x03, 0x01, 0x5e, 0x00}, {0x06, 0x01, 0x5e, 0x00},
		{0x0a, 0x01, 0x5e, 0x00}, {0x0f, 0x01, 0x5e, 0x00},
		{0x18, 0x01, 0x5e, 0x00}, {0x1f, 0x01, 0x5e, 0x00},
		{0x29, 0x01, 0x5e, 0x00}, {0x38, 0x01, 0x5e, 0x01},
		{0x03, 0x01, 0x7d, 0x00}, {0x06, 0x01, 0x7d, 0x00},
		{0x0a, 0x01, 0x7d, 0x00}, {0x0f, 0x01, 0x7d, 0x00},
		{0x18, 0x01, 0x7d, 0x00}, {0x1f, 0x01, 0x7d, 0x00},
		{0x29, 0x01, 0x7d, 0x00}, {0x38, 0x01, 0x7d, 0x01}
	],
	[
		{0x02, 0x01, 0x3c, 0x00}, {0x09, 0x01, 0x3c, 0x00},
		{0x17, 0x01, 0x3c, 0x00}, {0x28, 0x01, 0x3c, 0x01},
		{0x02, 0x01, 0x60, 0x00}, {0x09, 0x01, 0x60, 0x00},
		{0x17, 0x01, 0x60, 0x00}, {0x28, 0x01, 0x60, 0x01},
		{0x02, 0x01, 0x7b, 0x00}, {0x09, 0x01, 0x7b, 0x00},
		{0x17, 0x01, 0x7b, 0x00}, {0x28, 0x01, 0x7b, 0x01},
		{0x61, 0x00, 0x00, 0x00}, {0x65, 0x00, 0x00, 0x00},
		{0x6f, 0x00, 0x00, 0x00}, {0x85, 0x00, 0x00, 0x01}
	],
	[
		{0x03, 0x01, 0x3c, 0x00}, {0x06, 0x01, 0x3c, 0x00},
		{0x0a, 0x01, 0x3c, 0x00}, {0x0f, 0x01, 0x3c, 0x00},
		{0x18, 0x01, 0x3c, 0x00}, {0x1f, 0x01, 0x3c, 0x00},
		{0x29, 0x01, 0x3c, 0x00}, {0x38, 0x01, 0x3c, 0x01},
		{0x03, 0x01, 0x60, 0x00}, {0x06, 0x01, 0x60, 0x00},
		{0x0a, 0x01, 0x60, 0x00}, {0x0f, 0x01, 0x60, 0x00},
		{0x18, 0x01, 0x60, 0x00}, {0x1f, 0x01, 0x60, 0x00},
		{0x29, 0x01, 0x60, 0x00}, {0x38, 0x01, 0x60, 0x01}
	],
	[
		{0x03, 0x01, 0x7b, 0x00}, {0x06, 0x01, 0x7b, 0x00},
		{0x0a, 0x01, 0x7b, 0x00}, {0x0f, 0x01, 0x7b, 0x00},
		{0x18, 0x01, 0x7b, 0x00}, {0x1f, 0x01, 0x7b, 0x00},
		{0x29, 0x01, 0x7b, 0x00}, {0x38, 0x01, 0x7b, 0x01},
		{0x62, 0x00, 0x00, 0x00}, {0x63, 0x00, 0x00, 0x00},
		{0x66, 0x00, 0x00, 0x00}, {0x69, 0x00, 0x00, 0x00},
		{0x70, 0x00, 0x00, 0x00}, {0x77, 0x00, 0x00, 0x00},
		{0x86, 0x00, 0x00, 0x00}, {0x99, 0x00, 0x00, 0x01}
	],
	/* 95 */
	[
		{0x00, 0x01, 0x5c, 0x01}, {0x00, 0x01, 0xc3, 0x01},
		{0x00, 0x01, 0xd0, 0x01}, {0x64, 0x00, 0x00, 0x00},
		{0x67, 0x00, 0x00, 0x00}, {0x68, 0x00, 0x00, 0x00},
		{0x6a, 0x00, 0x00, 0x00}, {0x6b, 0x00, 0x00, 0x00},
		{0x71, 0x00, 0x00, 0x00}, {0x74, 0x00, 0x00, 0x00},
		{0x78, 0x00, 0x00, 0x00}, {0x7e, 0x00, 0x00, 0x00},
		{0x87, 0x00, 0x00, 0x00}, {0x8e, 0x00, 0x00, 0x00},
		{0x9a, 0x00, 0x00, 0x00}, {0xa9, 0x00, 0x00, 0x01}
	],
	[
		{0x01, 0x01, 0x5c, 0x00}, {0x16, 0x01, 0x5c, 0x01},
		{0x01, 0x01, 0xc3, 0x00}, {0x16, 0x01, 0xc3, 0x01},
		{0x01, 0x01, 0xd0, 0x00}, {0x16, 0x01, 0xd0, 0x01},
		{0x00, 0x01, 0x80, 0x01}, {0x00, 0x01, 0x82, 0x01},
		{0x00, 0x01, 0x83, 0x01}, {0x00, 0x01, 0xa2, 0x01},
		{0x00, 0x01, 0xb8, 0x01}, {0x00, 0x01, 0xc2, 0x01},
		{0x00, 0x01, 0xe0, 0x01}, {0x00, 0x01, 0xe2, 0x01},
		{0x6c, 0x00, 0x00, 0x00}, {0x6d, 0x00, 0x00, 0x00}
	],
	[
		{0x02, 0x01, 0x5c, 0x00}, {0x09, 0x01, 0x5c, 0x00},
		{0x17, 0x01, 0x5c, 0x00}, {0x28, 0x01, 0x5c, 0x01},
		{0x02, 0x01, 0xc3, 0x00}, {0x09, 0x01, 0xc3, 0x00},
		{0x17, 0x01, 0xc3, 0x00}, {0x28, 0x01, 0xc3, 0x01},
		{0x02, 0x01, 0xd0, 0x00}, {0x09, 0x01, 0xd0, 0x00},
		{0x17, 0x01, 0xd0, 0x00}, {0x28, 0x01, 0xd0, 0x01},
		{0x01, 0x01, 0x80, 0x00}, {0x16, 0x01, 0x80, 0x01},
		{0x01, 0x01, 0x82, 0x00}, {0x16, 0x01, 0x82, 0x01}
	],
	[
		{0x03, 0x01, 0x5c, 0x00}, {0x06, 0x01, 0x5c, 0x00},
		{0x0a, 0x01, 0x5c, 0x00}, {0x0f, 0x01, 0x5c, 0x00},
		{0x18, 0x01, 0x5c, 0x00}, {0x1f, 0x01, 0x5c, 0x00},
		{0x29, 0x01, 0x5c, 0x00}, {0x38, 0x01, 0x5c, 0x01},
		{0x03, 0x01, 0xc3, 0x00}, {0x06, 0x01, 0xc3, 0x00},
		{0x0a, 0x01, 0xc3, 0x00}, {0x0f, 0x01, 0xc3, 0x00},
		{0x18, 0x01, 0xc3, 0x00}, {0x1f, 0x01, 0xc3, 0x00},
		{0x29, 0x01, 0xc3, 0x00}, {0x38, 0x01, 0xc3, 0x01}
	],
	[
		{0x03, 0x01, 0xd0, 0x00}, {0x06, 0x01, 0xd0, 0x00},
		{0x0a, 0x01, 0xd0, 0x00}, {0x0f, 0x01, 0xd0, 0x00},
		{0x18, 0x01, 0xd0, 0x00}, {0x1f, 0x01, 0xd0, 0x00},
		{0x29, 0x01, 0xd0, 0x00}, {0x38, 0x01, 0xd0, 0x01},
		{0x02, 0x01, 0x80, 0x00}, {0x09, 0x01, 0x80, 0x00},
		{0x17, 0x01, 0x80, 0x00}, {0x28, 0x01, 0x80, 0x01},
		{0x02, 0x01, 0x82, 0x00}, {0x09, 0x01, 0x82, 0x00},
		{0x17, 0x01, 0x82, 0x00}, {0x28, 0x01, 0x82, 0x01}
	],
	/* 100 */
	[
		{0x03, 0x01, 0x80, 0x00}, {0x06, 0x01, 0x80, 0x00},
		{0x0a, 0x01, 0x80, 0x00}, {0x0f, 0x01, 0x80, 0x00},
		{0x18, 0x01, 0x80, 0x00}, {0x1f, 0x01, 0x80, 0x00},
		{0x29, 0x01, 0x80, 0x00}, {0x38, 0x01, 0x80, 0x01},
		{0x03, 0x01, 0x82, 0x00}, {0x06, 0x01, 0x82, 0x00},
		{0x0a, 0x01, 0x82, 0x00}, {0x0f, 0x01, 0x82, 0x00},
		{0x18, 0x01, 0x82, 0x00}, {0x1f, 0x01, 0x82, 0x00},
		{0x29, 0x01, 0x82, 0x00}, {0x38, 0x01, 0x82, 0x01}
	],
	[
		{0x01, 0x01, 0x83, 0x00}, {0x16, 0x01, 0x83, 0x01},
		{0x01, 0x01, 0xa2, 0x00}, {0x16, 0x01, 0xa2, 0x01},
		{0x01, 0x01, 0xb8, 0x00}, {0x16, 0x01, 0xb8, 0x01},
		{0x01, 0x01, 0xc2, 0x00}, {0x16, 0x01, 0xc2, 0x01},
		{0x01, 0x01, 0xe0, 0x00}, {0x16, 0x01, 0xe0, 0x01},
		{0x01, 0x01, 0xe2, 0x00}, {0x16, 0x01, 0xe2, 0x01},
		{0x00, 0x01, 0x99, 0x01}, {0x00, 0x01, 0xa1, 0x01},
		{0x00, 0x01, 0xa7, 0x01}, {0x00, 0x01, 0xac, 0x01}
	],
	[
		{0x02, 0x01, 0x83, 0x00}, {0x09, 0x01, 0x83, 0x00},
		{0x17, 0x01, 0x83, 0x00}, {0x28, 0x01, 0x83, 0x01},
		{0x02, 0x01, 0xa2, 0x00}, {0x09, 0x01, 0xa2, 0x00},
		{0x17, 0x01, 0xa2, 0x00}, {0x28, 0x01, 0xa2, 0x01},
		{0x02, 0x01, 0xb8, 0x00}, {0x09, 0x01, 0xb8, 0x00},
		{0x17, 0x01, 0xb8, 0x00}, {0x28, 0x01, 0xb8, 0x01},
		{0x02, 0x01, 0xc2, 0x00}, {0x09, 0x01, 0xc2, 0x00},
		{0x17, 0x01, 0xc2, 0x00}, {0x28, 0x01, 0xc2, 0x01}
	],
	[
		{0x03, 0x01, 0x83, 0x00}, {0x06, 0x01, 0x83, 0x00},
		{0x0a, 0x01, 0x83, 0x00}, {0x0f, 0x01, 0x83, 0x00},
		{0x18, 0x01, 0x83, 0x00}, {0x1f, 0x01, 0x83, 0x00},
		{0x29, 0x01, 0x83, 0x00}, {0x38, 0x01, 0x83, 0x01},
		{0x03, 0x01, 0xa2, 0x00}, {0x06, 0x01, 0xa2, 0x00},
		{0x0a, 0x01, 0xa2, 0x00}, {0x0f, 0x01, 0xa2, 0x00},
		{0x18, 0x01, 0xa2, 0x00}, {0x1f, 0x01, 0xa2, 0x00},
		{0x29, 0x01, 0xa2, 0x00}, {0x38, 0x01, 0xa2, 0x01}
	],
	[
		{0x03, 0x01, 0xb8, 0x00}, {0x06, 0x01, 0xb8, 0x00},
		{0x0a, 0x01, 0xb8, 0x00}, {0x0f, 0x01, 0xb8, 0x00},
		{0x18, 0x01, 0xb8, 0x00}, {0x1f, 0x01, 0xb8, 0x00},
		{0x29, 0x01, 0xb8, 0x00}, {0x38, 0x01, 0xb8, 0x01},
		{0x03, 0x01, 0xc2, 0x00}, {0x06, 0x01, 0xc2, 0x00},
		{0x0a, 0x01, 0xc2, 0x00}, {0x0f, 0x01, 0xc2, 0x00},
		{0x18, 0x01, 0xc2, 0x00}, {0x1f, 0x01, 0xc2, 0x00},
		{0x29, 0x01, 0xc2, 0x00}, {0x38, 0x01, 0xc2, 0x01}
	],
	/* 105 */
	[
		{0x02, 0x01, 0xe0, 0x00}, {0x09, 0x01, 0xe0, 0x00},
		{0x17, 0x01, 0xe0, 0x00}, {0x28, 0x01, 0xe0, 0x01},
		{0x02, 0x01, 0xe2, 0x00}, {0x09, 0x01, 0xe2, 0x00},
		{0x17, 0x01, 0xe2, 0x00}, {0x28, 0x01, 0xe2, 0x01},
		{0x01, 0x01, 0x99, 0x00}, {0x16, 0x01, 0x99, 0x01},
		{0x01, 0x01, 0xa1, 0x00}, {0x16, 0x01, 0xa1, 0x01},
		{0x01, 0x01, 0xa7, 0x00}, {0x16, 0x01, 0xa7, 0x01},
		{0x01, 0x01, 0xac, 0x00}, {0x16, 0x01, 0xac, 0x01}
	],
	[
		{0x03, 0x01, 0xe0, 0x00}, {0x06, 0x01, 0xe0, 0x00},
		{0x0a, 0x01, 0xe0, 0x00}, {0x0f, 0x01, 0xe0, 0x00},
		{0x18, 0x01, 0xe0, 0x00}, {0x1f, 0x01, 0xe0, 0x00},
		{0x29, 0x01, 0xe0, 0x00}, {0x38, 0x01, 0xe0, 0x01},
		{0x03, 0x01, 0xe2, 0x00}, {0x06, 0x01, 0xe2, 0x00},
		{0x0a, 0x01, 0xe2, 0x00}, {0x0f, 0x01, 0xe2, 0x00},
		{0x18, 0x01, 0xe2, 0x00}, {0x1f, 0x01, 0xe2, 0x00},
		{0x29, 0x01, 0xe2, 0x00}, {0x38, 0x01, 0xe2, 0x01}
	],
	[
		{0x02, 0x01, 0x99, 0x00}, {0x09, 0x01, 0x99, 0x00},
		{0x17, 0x01, 0x99, 0x00}, {0x28, 0x01, 0x99, 0x01},
		{0x02, 0x01, 0xa1, 0x00}, {0x09, 0x01, 0xa1, 0x00},
		{0x17, 0x01, 0xa1, 0x00}, {0x28, 0x01, 0xa1, 0x01},
		{0x02, 0x01, 0xa7, 0x00}, {0x09, 0x01, 0xa7, 0x00},
		{0x17, 0x01, 0xa7, 0x00}, {0x28, 0x01, 0xa7, 0x01},
		{0x02, 0x01, 0xac, 0x00}, {0x09, 0x01, 0xac, 0x00},
		{0x17, 0x01, 0xac, 0x00}, {0x28, 0x01, 0xac, 0x01}
	],
	[
		{0x03, 0x01, 0x99, 0x00}, {0x06, 0x01, 0x99, 0x00},
		{0x0a, 0x01, 0x99, 0x00}, {0x0f, 0x01, 0x99, 0x00},
		{0x18, 0x01, 0x99, 0x00}, {0x1f, 0x01, 0x99, 0x00},
		{0x29, 0x01, 0x99, 0x00}, {0x38, 0x01, 0x99, 0x01},
		{0x03, 0x01, 0xa1, 0x00}, {0x06, 0x01, 0xa1, 0x00},
		{0x0a, 0x01, 0xa1, 0x00}, {0x0f, 0x01, 0xa1, 0x00},
		{0x18, 0x01, 0xa1, 0x00}, {0x1f, 0x01, 0xa1, 0x00},
		{0x29, 0x01, 0xa1, 0x00}, {0x38, 0x01, 0xa1, 0x01}
	],
	[
		{0x03, 0x01, 0xa7, 0x00}, {0x06, 0x01, 0xa7, 0x00},
		{0x0a, 0x01, 0xa7, 0x00}, {0x0f, 0x01, 0xa7, 0x00},
		{0x18, 0x01, 0xa7, 0x00}, {0x1f, 0x01, 0xa7, 0x00},
		{0x29, 0x01, 0xa7, 0x00}, {0x38, 0x01, 0xa7, 0x01},
		{0x03, 0x01, 0xac, 0x00}, {0x06, 0x01, 0xac, 0x00},
		{0x0a, 0x01, 0xac, 0x00}, {0x0f, 0x01, 0xac, 0x00},
		{0x18, 0x01, 0xac, 0x00}, {0x1f, 0x01, 0xac, 0x00},
		{0x29, 0x01, 0xac, 0x00}, {0x38, 0x01, 0xac, 0x01}
	],
	/* 110 */
	[
		{0x72, 0x00, 0x00, 0x00}, {0x73, 0x00, 0x00, 0x00},
		{0x75, 0x00, 0x00, 0x00}, {0x76, 0x00, 0x00, 0x00},
		{0x79, 0x00, 0x00, 0x00}, {0x7b, 0x00, 0x00, 0x00},
		{0x7f, 0x00, 0x00, 0x00}, {0x82, 0x00, 0x00, 0x00},
		{0x88, 0x00, 0x00, 0x00}, {0x8b, 0x00, 0x00, 0x00},
		{0x8f, 0x00, 0x00, 0x00}, {0x92, 0x00, 0x00, 0x00},
		{0x9b, 0x00, 0x00, 0x00}, {0xa2, 0x00, 0x00, 0x00},
		{0xaa, 0x00, 0x00, 0x00}, {0xb4, 0x00, 0x00, 0x01}
	],
	[
		{0x00, 0x01, 0xb0, 0x01}, {0x00, 0x01, 0xb1, 0x01},
		{0x00, 0x01, 0xb3, 0x01}, {0x00, 0x01, 0xd1, 0x01},
		{0x00, 0x01, 0xd8, 0x01}, {0x00, 0x01, 0xd9, 0x01},
		{0x00, 0x01, 0xe3, 0x01}, {0x00, 0x01, 0xe5, 0x01},
		{0x00, 0x01, 0xe6, 0x01}, {0x7a, 0x00, 0x00, 0x00},
		{0x7c, 0x00, 0x00, 0x00}, {0x7d, 0x00, 0x00, 0x00},
		{0x80, 0x00, 0x00, 0x00}, {0x81, 0x00, 0x00, 0x00},
		{0x83, 0x00, 0x00, 0x00}, {0x84, 0x00, 0x00, 0x00}
	],
	[
		{0x01, 0x01, 0xb0, 0x00}, {0x16, 0x01, 0xb0, 0x01},
		{0x01, 0x01, 0xb1, 0x00}, {0x16, 0x01, 0xb1, 0x01},
		{0x01, 0x01, 0xb3, 0x00}, {0x16, 0x01, 0xb3, 0x01},
		{0x01, 0x01, 0xd1, 0x00}, {0x16, 0x01, 0xd1, 0x01},
		{0x01, 0x01, 0xd8, 0x00}, {0x16, 0x01, 0xd8, 0x01},
		{0x01, 0x01, 0xd9, 0x00}, {0x16, 0x01, 0xd9, 0x01},
		{0x01, 0x01, 0xe3, 0x00}, {0x16, 0x01, 0xe3, 0x01},
		{0x01, 0x01, 0xe5, 0x00}, {0x16, 0x01, 0xe5, 0x01}
	],
	[
		{0x02, 0x01, 0xb0, 0x00}, {0x09, 0x01, 0xb0, 0x00},
		{0x17, 0x01, 0xb0, 0x00}, {0x28, 0x01, 0xb0, 0x01},
		{0x02, 0x01, 0xb1, 0x00}, {0x09, 0x01, 0xb1, 0x00},
		{0x17, 0x01, 0xb1, 0x00}, {0x28, 0x01, 0xb1, 0x01},
		{0x02, 0x01, 0xb3, 0x00}, {0x09, 0x01, 0xb3, 0x00},
		{0x17, 0x01, 0xb3, 0x00}, {0x28, 0x01, 0xb3, 0x01},
		{0x02, 0x01, 0xd1, 0x00}, {0x09, 0x01, 0xd1, 0x00},
		{0x17, 0x01, 0xd1, 0x00}, {0x28, 0x01, 0xd1, 0x01}
	],
	[
		{0x03, 0x01, 0xb0, 0x00}, {0x06, 0x01, 0xb0, 0x00},
		{0x0a, 0x01, 0xb0, 0x00}, {0x0f, 0x01, 0xb0, 0x00},
		{0x18, 0x01, 0xb0, 0x00}, {0x1f, 0x01, 0xb0, 0x00},
		{0x29, 0x01, 0xb0, 0x00}, {0x38, 0x01, 0xb0, 0x01},
		{0x03, 0x01, 0xb1, 0x00}, {0x06, 0x01, 0xb1, 0x00},
		{0x0a, 0x01, 0xb1, 0x00}, {0x0f, 0x01, 0xb1, 0x00},
		{0x18, 0x01, 0xb1, 0x00}, {0x1f, 0x01, 0xb1, 0x00},
		{0x29, 0x01, 0xb1, 0x00}, {0x38, 0x01, 0xb1, 0x01}
	],
	/* 115 */
	[
		{0x03, 0x01, 0xb3, 0x00}, {0x06, 0x01, 0xb3, 0x00},
		{0x0a, 0x01, 0xb3, 0x00}, {0x0f, 0x01, 0xb3, 0x00},
		{0x18, 0x01, 0xb3, 0x00}, {0x1f, 0x01, 0xb3, 0x00},
		{0x29, 0x01, 0xb3, 0x00}, {0x38, 0x01, 0xb3, 0x01},
		{0x03, 0x01, 0xd1, 0x00}, {0x06, 0x01, 0xd1, 0x00},
		{0x0a, 0x01, 0xd1, 0x00}, {0x0f, 0x01, 0xd1, 0x00},
		{0x18, 0x01, 0xd1, 0x00}, {0x1f, 0x01, 0xd1, 0x00},
		{0x29, 0x01, 0xd1, 0x00}, {0x38, 0x01, 0xd1, 0x01}
	],
	[
		{0x02, 0x01, 0xd8, 0x00}, {0x09, 0x01, 0xd8, 0x00},
		{0x17, 0x01, 0xd8, 0x00}, {0x28, 0x01, 0xd8, 0x01},
		{0x02, 0x01, 0xd9, 0x00}, {0x09, 0x01, 0xd9, 0x00},
		{0x17, 0x01, 0xd9, 0x00}, {0x28, 0x01, 0xd9, 0x01},
		{0x02, 0x01, 0xe3, 0x00}, {0x09, 0x01, 0xe3, 0x00},
		{0x17, 0x01, 0xe3, 0x00}, {0x28, 0x01, 0xe3, 0x01},
		{0x02, 0x01, 0xe5, 0x00}, {0x09, 0x01, 0xe5, 0x00},
		{0x17, 0x01, 0xe5, 0x00}, {0x28, 0x01, 0xe5, 0x01}
	],
	[
		{0x03, 0x01, 0xd8, 0x00}, {0x06, 0x01, 0xd8, 0x00},
		{0x0a, 0x01, 0xd8, 0x00}, {0x0f, 0x01, 0xd8, 0x00},
		{0x18, 0x01, 0xd8, 0x00}, {0x1f, 0x01, 0xd8, 0x00},
		{0x29, 0x01, 0xd8, 0x00}, {0x38, 0x01, 0xd8, 0x01},
		{0x03, 0x01, 0xd9, 0x00}, {0x06, 0x01, 0xd9, 0x00},
		{0x0a, 0x01, 0xd9, 0x00}, {0x0f, 0x01, 0xd9, 0x00},
		{0x18, 0x01, 0xd9, 0x00}, {0x1f, 0x01, 0xd9, 0x00},
		{0x29, 0x01, 0xd9, 0x00}, {0x38, 0x01, 0xd9, 0x01}
	],
	[
		{0x03, 0x01, 0xe3, 0x00}, {0x06, 0x01, 0xe3, 0x00},
		{0x0a, 0x01, 0xe3, 0x00}, {0x0f, 0x01, 0xe3, 0x00},
		{0x18, 0x01, 0xe3, 0x00}, {0x1f, 0x01, 0xe3, 0x00},
		{0x29, 0x01, 0xe3, 0x00}, {0x38, 0x01, 0xe3, 0x01},
		{0x03, 0x01, 0xe5, 0x00}, {0x06, 0x01, 0xe5, 0x00},
		{0x0a, 0x01, 0xe5, 0x00}, {0x0f, 0x01, 0xe5, 0x00},
		{0x18, 0x01, 0xe5, 0x00}, {0x1f, 0x01, 0xe5, 0x00},
		{0x29, 0x01, 0xe5, 0x00}, {0x38, 0x01, 0xe5, 0x01}
	],
	[
		{0x01, 0x01, 0xe6, 0x00}, {0x16, 0x01, 0xe6, 0x01},
		{0x00, 0x01, 0x81, 0x01}, {0x00, 0x01, 0x84, 0x01},
		{0x00, 0x01, 0x85, 0x01}, {0x00, 0x01, 0x86, 0x01},
		{0x00, 0x01, 0x88, 0x01}, {0x00, 0x01, 0x92, 0x01},
		{0x00, 0x01, 0x9a, 0x01}, {0x00, 0x01, 0x9c, 0x01},
		{0x00, 0x01, 0xa0, 0x01}, {0x00, 0x01, 0xa3, 0x01},
		{0x00, 0x01, 0xa4, 0x01}, {0x00, 0x01, 0xa9, 0x01},
		{0x00, 0x01, 0xaa, 0x01}, {0x00, 0x01, 0xad, 0x01}
	],
	/* 120 */
	[
		{0x02, 0x01, 0xe6, 0x00}, {0x09, 0x01, 0xe6, 0x00},
		{0x17, 0x01, 0xe6, 0x00}, {0x28, 0x01, 0xe6, 0x01},
		{0x01, 0x01, 0x81, 0x00}, {0x16, 0x01, 0x81, 0x01},
		{0x01, 0x01, 0x84, 0x00}, {0x16, 0x01, 0x84, 0x01},
		{0x01, 0x01, 0x85, 0x00}, {0x16, 0x01, 0x85, 0x01},
		{0x01, 0x01, 0x86, 0x00}, {0x16, 0x01, 0x86, 0x01},
		{0x01, 0x01, 0x88, 0x00}, {0x16, 0x01, 0x88, 0x01},
		{0x01, 0x01, 0x92, 0x00}, {0x16, 0x01, 0x92, 0x01}
	],
	[
		{0x03, 0x01, 0xe6, 0x00}, {0x06, 0x01, 0xe6, 0x00},
		{0x0a, 0x01, 0xe6, 0x00}, {0x0f, 0x01, 0xe6, 0x00},
		{0x18, 0x01, 0xe6, 0x00}, {0x1f, 0x01, 0xe6, 0x00},
		{0x29, 0x01, 0xe6, 0x00}, {0x38, 0x01, 0xe6, 0x01},
		{0x02, 0x01, 0x81, 0x00}, {0x09, 0x01, 0x81, 0x00},
		{0x17, 0x01, 0x81, 0x00}, {0x28, 0x01, 0x81, 0x01},
		{0x02, 0x01, 0x84, 0x00}, {0x09, 0x01, 0x84, 0x00},
		{0x17, 0x01, 0x84, 0x00}, {0x28, 0x01, 0x84, 0x01}
	],
	[
		{0x03, 0x01, 0x81, 0x00}, {0x06, 0x01, 0x81, 0x00},
		{0x0a, 0x01, 0x81, 0x00}, {0x0f, 0x01, 0x81, 0x00},
		{0x18, 0x01, 0x81, 0x00}, {0x1f, 0x01, 0x81, 0x00},
		{0x29, 0x01, 0x81, 0x00}, {0x38, 0x01, 0x81, 0x01},
		{0x03, 0x01, 0x84, 0x00}, {0x06, 0x01, 0x84, 0x00},
		{0x0a, 0x01, 0x84, 0x00}, {0x0f, 0x01, 0x84, 0x00},
		{0x18, 0x01, 0x84, 0x00}, {0x1f, 0x01, 0x84, 0x00},
		{0x29, 0x01, 0x84, 0x00}, {0x38, 0x01, 0x84, 0x01}
	],
	[
		{0x02, 0x01, 0x85, 0x00}, {0x09, 0x01, 0x85, 0x00},
		{0x17, 0x01, 0x85, 0x00}, {0x28, 0x01, 0x85, 0x01},
		{0x02, 0x01, 0x86, 0x00}, {0x09, 0x01, 0x86, 0x00},
		{0x17, 0x01, 0x86, 0x00}, {0x28, 0x01, 0x86, 0x01},
		{0x02, 0x01, 0x88, 0x00}, {0x09, 0x01, 0x88, 0x00},
		{0x17, 0x01, 0x88, 0x00}, {0x28, 0x01, 0x88, 0x01},
		{0x02, 0x01, 0x92, 0x00}, {0x09, 0x01, 0x92, 0x00},
		{0x17, 0x01, 0x92, 0x00}, {0x28, 0x01, 0x92, 0x01}
	],
	[
		{0x03, 0x01, 0x85, 0x00}, {0x06, 0x01, 0x85, 0x00},
		{0x0a, 0x01, 0x85, 0x00}, {0x0f, 0x01, 0x85, 0x00},
		{0x18, 0x01, 0x85, 0x00}, {0x1f, 0x01, 0x85, 0x00},
		{0x29, 0x01, 0x85, 0x00}, {0x38, 0x01, 0x85, 0x01},
		{0x03, 0x01, 0x86, 0x00}, {0x06, 0x01, 0x86, 0x00},
		{0x0a, 0x01, 0x86, 0x00}, {0x0f, 0x01, 0x86, 0x00},
		{0x18, 0x01, 0x86, 0x00}, {0x1f, 0x01, 0x86, 0x00},
		{0x29, 0x01, 0x86, 0x00}, {0x38, 0x01, 0x86, 0x01}
	],
	/* 125 */
	[
		{0x03, 0x01, 0x88, 0x00}, {0x06, 0x01, 0x88, 0x00},
		{0x0a, 0x01, 0x88, 0x00}, {0x0f, 0x01, 0x88, 0x00},
		{0x18, 0x01, 0x88, 0x00}, {0x1f, 0x01, 0x88, 0x00},
		{0x29, 0x01, 0x88, 0x00}, {0x38, 0x01, 0x88, 0x01},
		{0x03, 0x01, 0x92, 0x00}, {0x06, 0x01, 0x92, 0x00},
		{0x0a, 0x01, 0x92, 0x00}, {0x0f, 0x01, 0x92, 0x00},
		{0x18, 0x01, 0x92, 0x00}, {0x1f, 0x01, 0x92, 0x00},
		{0x29, 0x01, 0x92, 0x00}, {0x38, 0x01, 0x92, 0x01}
	],
	[
		{0x01, 0x01, 0x9a, 0x00}, {0x16, 0x01, 0x9a, 0x01},
		{0x01, 0x01, 0x9c, 0x00}, {0x16, 0x01, 0x9c, 0x01},
		{0x01, 0x01, 0xa0, 0x00}, {0x16, 0x01, 0xa0, 0x01},
		{0x01, 0x01, 0xa3, 0x00}, {0x16, 0x01, 0xa3, 0x01},
		{0x01, 0x01, 0xa4, 0x00}, {0x16, 0x01, 0xa4, 0x01},
		{0x01, 0x01, 0xa9, 0x00}, {0x16, 0x01, 0xa9, 0x01},
		{0x01, 0x01, 0xaa, 0x00}, {0x16, 0x01, 0xaa, 0x01},
		{0x01, 0x01, 0xad, 0x00}, {0x16, 0x01, 0xad, 0x01}
	],
	[
		{0x02, 0x01, 0x9a, 0x00}, {0x09, 0x01, 0x9a, 0x00},
		{0x17, 0x01, 0x9a, 0x00}, {0x28, 0x01, 0x9a, 0x01},
		{0x02, 0x01, 0x9c, 0x00}, {0x09, 0x01, 0x9c, 0x00},
		{0x17, 0x01, 0x9c, 0x00}, {0x28, 0x01, 0x9c, 0x01},
		{0x02, 0x01, 0xa0, 0x00}, {0x09, 0x01, 0xa0, 0x00},
		{0x17, 0x01, 0xa0, 0x00}, {0x28, 0x01, 0xa0, 0x01},
		{0x02, 0x01, 0xa3, 0x00}, {0x09, 0x01, 0xa3, 0x00},
		{0x17, 0x01, 0xa3, 0x00}, {0x28, 0x01, 0xa3, 0x01}
	],
	[
		{0x03, 0x01, 0x9a, 0x00}, {0x06, 0x01, 0x9a, 0x00},
		{0x0a, 0x01, 0x9a, 0x00}, {0x0f, 0x01, 0x9a, 0x00},
		{0x18, 0x01, 0x9a, 0x00}, {0x1f, 0x01, 0x9a, 0x00},
		{0x29, 0x01, 0x9a, 0x00}, {0x38, 0x01, 0x9a, 0x01},
		{0x03, 0x01, 0x9c, 0x00}, {0x06, 0x01, 0x9c, 0x00},
		{0x0a, 0x01, 0x9c, 0x00}, {0x0f, 0x01, 0x9c, 0x00},
		{0x18, 0x01, 0x9c, 0x00}, {0x1f, 0x01, 0x9c, 0x00},
		{0x29, 0x01, 0x9c, 0x00}, {0x38, 0x01, 0x9c, 0x01}
	],
	[
		{0x03, 0x01, 0xa0, 0x00}, {0x06, 0x01, 0xa0, 0x00},
		{0x0a, 0x01, 0xa0, 0x00}, {0x0f, 0x01, 0xa0, 0x00},
		{0x18, 0x01, 0xa0, 0x00}, {0x1f, 0x01, 0xa0, 0x00},
		{0x29, 0x01, 0xa0, 0x00}, {0x38, 0x01, 0xa0, 0x01},
		{0x03, 0x01, 0xa3, 0x00}, {0x06, 0x01, 0xa3, 0x00},
		{0x0a, 0x01, 0xa3, 0x00}, {0x0f, 0x01, 0xa3, 0x00},
		{0x18, 0x01, 0xa3, 0x00}, {0x1f, 0x01, 0xa3, 0x00},
		{0x29, 0x01, 0xa3, 0x00}, {0x38, 0x01, 0xa3, 0x01}
	],
	/* 130 */
	[
		{0x02, 0x01, 0xa4, 0x00}, {0x09, 0x01, 0xa4, 0x00},
		{0x17, 0x01, 0xa4, 0x00}, {0x28, 0x01, 0xa4, 0x01},
		{0x02, 0x01, 0xa9, 0x00}, {0x09, 0x01, 0xa9, 0x00},
		{0x17, 0x01, 0xa9, 0x00}, {0x28, 0x01, 0xa9, 0x01},
		{0x02, 0x01, 0xaa, 0x00}, {0x09, 0x01, 0xaa, 0x00},
		{0x17, 0x01, 0xaa, 0x00}, {0x28, 0x01, 0xaa, 0x01},
		{0x02, 0x01, 0xad, 0x00}, {0x09, 0x01, 0xad, 0x00},
		{0x17, 0x01, 0xad, 0x00}, {0x28, 0x01, 0xad, 0x01}
	],
	[
		{0x03, 0x01, 0xa4, 0x00}, {0x06, 0x01, 0xa4, 0x00},
		{0x0a, 0x01, 0xa4, 0x00}, {0x0f, 0x01, 0xa4, 0x00},
		{0x18, 0x01, 0xa4, 0x00}, {0x1f, 0x01, 0xa4, 0x00},
		{0x29, 0x01, 0xa4, 0x00}, {0x38, 0x01, 0xa4, 0x01},
		{0x03, 0x01, 0xa9, 0x00}, {0x06, 0x01, 0xa9, 0x00},
		{0x0a, 0x01, 0xa9, 0x00}, {0x0f, 0x01, 0xa9, 0x00},
		{0x18, 0x01, 0xa9, 0x00}, {0x1f, 0x01, 0xa9, 0x00},
		{0x29, 0x01, 0xa9, 0x00}, {0x38, 0x01, 0xa9, 0x01}
	],
	[
		{0x03, 0x01, 0xaa, 0x00}, {0x06, 0x01, 0xaa, 0x00},
		{0x0a, 0x01, 0xaa, 0x00}, {0x0f, 0x01, 0xaa, 0x00},
		{0x18, 0x01, 0xaa, 0x00}, {0x1f, 0x01, 0xaa, 0x00},
		{0x29, 0x01, 0xaa, 0x00}, {0x38, 0x01, 0xaa, 0x01},
		{0x03, 0x01, 0xad, 0x00}, {0x06, 0x01, 0xad, 0x00},
		{0x0a, 0x01, 0xad, 0x00}, {0x0f, 0x01, 0xad, 0x00},
		{0x18, 0x01, 0xad, 0x00}, {0x1f, 0x01, 0xad, 0x00},
		{0x29, 0x01, 0xad, 0x00}, {0x38, 0x01, 0xad, 0x01}
	],
	[
		{0x89, 0x00, 0x00, 0x00}, {0x8a, 0x00, 0x00, 0x00},
		{0x8c, 0x00, 0x00, 0x00}, {0x8d, 0x00, 0x00, 0x00},
		{0x90, 0x00, 0x00, 0x00}, {0x91, 0x00, 0x00, 0x00},
		{0x93, 0x00, 0x00, 0x00}, {0x96, 0x00, 0x00, 0x00},
		{0x9c, 0x00, 0x00, 0x00}, {0x9f, 0x00, 0x00, 0x00},
		{0xa3, 0x00, 0x00, 0x00}, {0xa6, 0x00, 0x00, 0x00},
		{0xab, 0x00, 0x00, 0x00}, {0xae, 0x00, 0x00, 0x00},
		{0xb5, 0x00, 0x00, 0x00}, {0xbe, 0x00, 0x00, 0x01}
	],
	[
		{0x00, 0x01, 0xb2, 0x01}, {0x00, 0x01, 0xb5, 0x01},
		{0x00, 0x01, 0xb9, 0x01}, {0x00, 0x01, 0xba, 0x01},
		{0x00, 0x01, 0xbb, 0x01}, {0x00, 0x01, 0xbd, 0x01},
		{0x00, 0x01, 0xbe, 0x01}, {0x00, 0x01, 0xc4, 0x01},
		{0x00, 0x01, 0xc6, 0x01}, {0x00, 0x01, 0xe4, 0x01},
		{0x00, 0x01, 0xe8, 0x01}, {0x00, 0x01, 0xe9, 0x01},
		{0x94, 0x00, 0x00, 0x00}, {0x95, 0x00, 0x00, 0x00},
		{0x97, 0x00, 0x00, 0x00}, {0x98, 0x00, 0x00, 0x00}
	],
	/* 135 */
	[
		{0x01, 0x01, 0xb2, 0x00}, {0x16, 0x01, 0xb2, 0x01},
		{0x01, 0x01, 0xb5, 0x00}, {0x16, 0x01, 0xb5, 0x01},
		{0x01, 0x01, 0xb9, 0x00}, {0x16, 0x01, 0xb9, 0x01},
		{0x01, 0x01, 0xba, 0x00}, {0x16, 0x01, 0xba, 0x01},
		{0x01, 0x01, 0xbb, 0x00}, {0x16, 0x01, 0xbb, 0x01},
		{0x01, 0x01, 0xbd, 0x00}, {0x16, 0x01, 0xbd, 0x01},
		{0x01, 0x01, 0xbe, 0x00}, {0x16, 0x01, 0xbe, 0x01},
		{0x01, 0x01, 0xc4, 0x00}, {0x16, 0x01, 0xc4, 0x01}
	],
	[
		{0x02, 0x01, 0xb2, 0x00}, {0x09, 0x01, 0xb2, 0x00},
		{0x17, 0x01, 0xb2, 0x00}, {0x28, 0x01, 0xb2, 0x01},
		{0x02, 0x01, 0xb5, 0x00}, {0x09, 0x01, 0xb5, 0x00},
		{0x17, 0x01, 0xb5, 0x00}, {0x28, 0x01, 0xb5, 0x01},
		{0x02, 0x01, 0xb9, 0x00}, {0x09, 0x01, 0xb9, 0x00},
		{0x17, 0x01, 0xb9, 0x00}, {0x28, 0x01, 0xb9, 0x01},
		{0x02, 0x01, 0xba, 0x00}, {0x09, 0x01, 0xba, 0x00},
		{0x17, 0x01, 0xba, 0x00}, {0x28, 0x01, 0xba, 0x01}
	],
	[
		{0x03, 0x01, 0xb2, 0x00}, {0x06, 0x01, 0xb2, 0x00},
		{0x0a, 0x01, 0xb2, 0x00}, {0x0f, 0x01, 0xb2, 0x00},
		{0x18, 0x01, 0xb2, 0x00}, {0x1f, 0x01, 0xb2, 0x00},
		{0x29, 0x01, 0xb2, 0x00}, {0x38, 0x01, 0xb2, 0x01},
		{0x03, 0x01, 0xb5, 0x00}, {0x06, 0x01, 0xb5, 0x00},
		{0x0a, 0x01, 0xb5, 0x00}, {0x0f, 0x01, 0xb5, 0x00},
		{0x18, 0x01, 0xb5, 0x00}, {0x1f, 0x01, 0xb5, 0x00},
		{0x29, 0x01, 0xb5, 0x00}, {0x38, 0x01, 0xb5, 0x01}
	],
	[
		{0x03, 0x01, 0xb9, 0x00}, {0x06, 0x01, 0xb9, 0x00},
		{0x0a, 0x01, 0xb9, 0x00}, {0x0f, 0x01, 0xb9, 0x00},
		{0x18, 0x01, 0xb9, 0x00}, {0x1f, 0x01, 0xb9, 0x00},
		{0x29, 0x01, 0xb9, 0x00}, {0x38, 0x01, 0xb9, 0x01},
		{0x03, 0x01, 0xba, 0x00}, {0x06, 0x01, 0xba, 0x00},
		{0x0a, 0x01, 0xba, 0x00}, {0x0f, 0x01, 0xba, 0x00},
		{0x18, 0x01, 0xba, 0x00}, {0x1f, 0x01, 0xba, 0x00},
		{0x29, 0x01, 0xba, 0x00}, {0x38, 0x01, 0xba, 0x01}
	],
	[
		{0x02, 0x01, 0xbb, 0x00}, {0x09, 0x01, 0xbb, 0x00},
		{0x17, 0x01, 0xbb, 0x00}, {0x28, 0x01, 0xbb, 0x01},
		{0x02, 0x01, 0xbd, 0x00}, {0x09, 0x01, 0xbd, 0x00},
		{0x17, 0x01, 0xbd, 0x00}, {0x28, 0x01, 0xbd, 0x01},
		{0x02, 0x01, 0xbe, 0x00}, {0x09, 0x01, 0xbe, 0x00},
		{0x17, 0x01, 0xbe, 0x00}, {0x28, 0x01, 0xbe, 0x01},
		{0x02, 0x01, 0xc4, 0x00}, {0x09, 0x01, 0xc4, 0x00},
		{0x17, 0x01, 0xc4, 0x00}, {0x28, 0x01, 0xc4, 0x01}
	],
	/* 140 */
	[
		{0x03, 0x01, 0xbb, 0x00}, {0x06, 0x01, 0xbb, 0x00},
		{0x0a, 0x01, 0xbb, 0x00}, {0x0f, 0x01, 0xbb, 0x00},
		{0x18, 0x01, 0xbb, 0x00}, {0x1f, 0x01, 0xbb, 0x00},
		{0x29, 0x01, 0xbb, 0x00}, {0x38, 0x01, 0xbb, 0x01},
		{0x03, 0x01, 0xbd, 0x00}, {0x06, 0x01, 0xbd, 0x00},
		{0x0a, 0x01, 0xbd, 0x00}, {0x0f, 0x01, 0xbd, 0x00},
		{0x18, 0x01, 0xbd, 0x00}, {0x1f, 0x01, 0xbd, 0x00},
		{0x29, 0x01, 0xbd, 0x00}, {0x38, 0x01, 0xbd, 0x01}
	],
	[
		{0x03, 0x01, 0xbe, 0x00}, {0x06, 0x01, 0xbe, 0x00},
		{0x0a, 0x01, 0xbe, 0x00}, {0x0f, 0x01, 0xbe, 0x00},
		{0x18, 0x01, 0xbe, 0x00}, {0x1f, 0x01, 0xbe, 0x00},
		{0x29, 0x01, 0xbe, 0x00}, {0x38, 0x01, 0xbe, 0x01},
		{0x03, 0x01, 0xc4, 0x00}, {0x06, 0x01, 0xc4, 0x00},
		{0x0a, 0x01, 0xc4, 0x00}, {0x0f, 0x01, 0xc4, 0x00},
		{0x18, 0x01, 0xc4, 0x00}, {0x1f, 0x01, 0xc4, 0x00},
		{0x29, 0x01, 0xc4, 0x00}, {0x38, 0x01, 0xc4, 0x01}
	],
	[
		{0x01, 0x01, 0xc6, 0x00}, {0x16, 0x01, 0xc6, 0x01},
		{0x01, 0x01, 0xe4, 0x00}, {0x16, 0x01, 0xe4, 0x01},
		{0x01, 0x01, 0xe8, 0x00}, {0x16, 0x01, 0xe8, 0x01},
		{0x01, 0x01, 0xe9, 0x00}, {0x16, 0x01, 0xe9, 0x01},
		{0x00, 0x01, 0x01, 0x01}, {0x00, 0x01, 0x87, 0x01},
		{0x00, 0x01, 0x89, 0x01}, {0x00, 0x01, 0x8a, 0x01},
		{0x00, 0x01, 0x8b, 0x01}, {0x00, 0x01, 0x8c, 0x01},
		{0x00, 0x01, 0x8d, 0x01}, {0x00, 0x01, 0x8f, 0x01}
	],
	[
		{0x02, 0x01, 0xc6, 0x00}, {0x09, 0x01, 0xc6, 0x00},
		{0x17, 0x01, 0xc6, 0x00}, {0x28, 0x01, 0xc6, 0x01},
		{0x02, 0x01, 0xe4, 0x00}, {0x09, 0x01, 0xe4, 0x00},
		{0x17, 0x01, 0xe4, 0x00}, {0x28, 0x01, 0xe4, 0x01},
		{0x02, 0x01, 0xe8, 0x00}, {0x09, 0x01, 0xe8, 0x00},
		{0x17, 0x01, 0xe8, 0x00}, {0x28, 0x01, 0xe8, 0x01},
		{0x02, 0x01, 0xe9, 0x00}, {0x09, 0x01, 0xe9, 0x00},
		{0x17, 0x01, 0xe9, 0x00}, {0x28, 0x01, 0xe9, 0x01}
	],
	[
		{0x03, 0x01, 0xc6, 0x00}, {0x06, 0x01, 0xc6, 0x00},
		{0x0a, 0x01, 0xc6, 0x00}, {0x0f, 0x01, 0xc6, 0x00},
		{0x18, 0x01, 0xc6, 0x00}, {0x1f, 0x01, 0xc6, 0x00},
		{0x29, 0x01, 0xc6, 0x00}, {0x38, 0x01, 0xc6, 0x01},
		{0x03, 0x01, 0xe4, 0x00}, {0x06, 0x01, 0xe4, 0x00},
		{0x0a, 0x01, 0xe4, 0x00}, {0x0f, 0x01, 0xe4, 0x00},
		{0x18, 0x01, 0xe4, 0x00}, {0x1f, 0x01, 0xe4, 0x00},
		{0x29, 0x01, 0xe4, 0x00}, {0x38, 0x01, 0xe4, 0x01}
	],
	/* 145 */
	[
		{0x03, 0x01, 0xe8, 0x00}, {0x06, 0x01, 0xe8, 0x00},
		{0x0a, 0x01, 0xe8, 0x00}, {0x0f, 0x01, 0xe8, 0x00},
		{0x18, 0x01, 0xe8, 0x00}, {0x1f, 0x01, 0xe8, 0x00},
		{0x29, 0x01, 0xe8, 0x00}, {0x38, 0x01, 0xe8, 0x01},
		{0x03, 0x01, 0xe9, 0x00}, {0x06, 0x01, 0xe9, 0x00},
		{0x0a, 0x01, 0xe9, 0x00}, {0x0f, 0x01, 0xe9, 0x00},
		{0x18, 0x01, 0xe9, 0x00}, {0x1f, 0x01, 0xe9, 0x00},
		{0x29, 0x01, 0xe9, 0x00}, {0x38, 0x01, 0xe9, 0x01}
	],
	[
		{0x01, 0x01, 0x01, 0x00}, {0x16, 0x01, 0x01, 0x01},
		{0x01, 0x01, 0x87, 0x00}, {0x16, 0x01, 0x87, 0x01},
		{0x01, 0x01, 0x89, 0x00}, {0x16, 0x01, 0x89, 0x01},
		{0x01, 0x01, 0x8a, 0x00}, {0x16, 0x01, 0x8a, 0x01},
		{0x01, 0x01, 0x8b, 0x00}, {0x16, 0x01, 0x8b, 0x01},
		{0x01, 0x01, 0x8c, 0x00}, {0x16, 0x01, 0x8c, 0x01},
		{0x01, 0x01, 0x8d, 0x00}, {0x16, 0x01, 0x8d, 0x01},
		{0x01, 0x01, 0x8f, 0x00}, {0x16, 0x01, 0x8f, 0x01}
	],
	[
		{0x02, 0x01, 0x01, 0x00}, {0x09, 0x01, 0x01, 0x00},
		{0x17, 0x01, 0x01, 0x00}, {0x28, 0x01, 0x01, 0x01},
		{0x02, 0x01, 0x87, 0x00}, {0x09, 0x01, 0x87, 0x00},
		{0x17, 0x01, 0x87, 0x00}, {0x28, 0x01, 0x87, 0x01},
		{0x02, 0x01, 0x89, 0x00}, {0x09, 0x01, 0x89, 0x00},
		{0x17, 0x01, 0x89, 0x00}, {0x28, 0x01, 0x89, 0x01},
		{0x02, 0x01, 0x8a, 0x00}, {0x09, 0x01, 0x8a, 0x00},
		{0x17, 0x01, 0x8a, 0x00}, {0x28, 0x01, 0x8a, 0x01}
	],
	[
		{0x03, 0x01, 0x01, 0x00}, {0x06, 0x01, 0x01, 0x00},
		{0x0a, 0x01, 0x01, 0x00}, {0x0f, 0x01, 0x01, 0x00},
		{0x18, 0x01, 0x01, 0x00}, {0x1f, 0x01, 0x01, 0x00},
		{0x29, 0x01, 0x01, 0x00}, {0x38, 0x01, 0x01, 0x01},
		{0x03, 0x01, 0x87, 0x00}, {0x06, 0x01, 0x87, 0x00},
		{0x0a, 0x01, 0x87, 0x00}, {0x0f, 0x01, 0x87, 0x00},
		{0x18, 0x01, 0x87, 0x00}, {0x1f, 0x01, 0x87, 0x00},
		{0x29, 0x01, 0x87, 0x00}, {0x38, 0x01, 0x87, 0x01}
	],
	[
		{0x03, 0x01, 0x89, 0x00}, {0x06, 0x01, 0x89, 0x00},
		{0x0a, 0x01, 0x89, 0x00}, {0x0f, 0x01, 0x89, 0x00},
		{0x18, 0x01, 0x89, 0x00}, {0x1f, 0x01, 0x89, 0x00},
		{0x29, 0x01, 0x89, 0x00}, {0x38, 0x01, 0x89, 0x01},
		{0x03, 0x01, 0x8a, 0x00}, {0x06, 0x01, 0x8a, 0x00},
		{0x0a, 0x01, 0x8a, 0x00}, {0x0f, 0x01, 0x8a, 0x00},
		{0x18, 0x01, 0x8a, 0x00}, {0x1f, 0x01, 0x8a, 0x00},
		{0x29, 0x01, 0x8a, 0x00}, {0x38, 0x01, 0x8a, 0x01}
	],
	/* 150 */
	[
		{0x02, 0x01, 0x8b, 0x00}, {0x09, 0x01, 0x8b, 0x00},
		{0x17, 0x01, 0x8b, 0x00}, {0x28, 0x01, 0x8b, 0x01},
		{0x02, 0x01, 0x8c, 0x00}, {0x09, 0x01, 0x8c, 0x00},
		{0x17, 0x01, 0x8c, 0x00}, {0x28, 0x01, 0x8c, 0x01},
		{0x02, 0x01, 0x8d, 0x00}, {0x09, 0x01, 0x8d, 0x00},
		{0x17, 0x01, 0x8d, 0x00}, {0x28, 0x01, 0x8d, 0x01},
		{0x02, 0x01, 0x8f, 0x00}, {0x09, 0x01, 0x8f, 0x00},
		{0x17, 0x01, 0x8f, 0x00}, {0x28, 0x01, 0x8f, 0x01}
	],
	[
		{0x03, 0x01, 0x8b, 0x00}, {0x06, 0x01, 0x8b, 0x00},
		{0x0a, 0x01, 0x8b, 0x00}, {0x0f, 0x01, 0x8b, 0x00},
		{0x18, 0x01, 0x8b, 0x00}, {0x1f, 0x01, 0x8b, 0x00},
		{0x29, 0x01, 0x8b, 0x00}, {0x38, 0x01, 0x8b, 0x01},
		{0x03, 0x01, 0x8c, 0x00}, {0x06, 0x01, 0x8c, 0x00},
		{0x0a, 0x01, 0x8c, 0x00}, {0x0f, 0x01, 0x8c, 0x00},
		{0x18, 0x01, 0x8c, 0x00}, {0x1f, 0x01, 0x8c, 0x00},
		{0x29, 0x01, 0x8c, 0x00}, {0x38, 0x01, 0x8c, 0x01}
	],
	[
		{0x03, 0x01, 0x8d, 0x00}, {0x06, 0x01, 0x8d, 0x00},
		{0x0a, 0x01, 0x8d, 0x00}, {0x0f, 0x01, 0x8d, 0x00},
		{0x18, 0x01, 0x8d, 0x00}, {0x1f, 0x01, 0x8d, 0x00},
		{0x29, 0x01, 0x8d, 0x00}, {0x38, 0x01, 0x8d, 0x01},
		{0x03, 0x01, 0x8f, 0x00}, {0x06, 0x01, 0x8f, 0x00},
		{0x0a, 0x01, 0x8f, 0x00}, {0x0f, 0x01, 0x8f, 0x00},
		{0x18, 0x01, 0x8f, 0x00}, {0x1f, 0x01, 0x8f, 0x00},
		{0x29, 0x01, 0x8f, 0x00}, {0x38, 0x01, 0x8f, 0x01}
	],
	[
		{0x9d, 0x00, 0x00, 0x00}, {0x9e, 0x00, 0x00, 0x00},
		{0xa0, 0x00, 0x00, 0x00}, {0xa1, 0x00, 0x00, 0x00},
		{0xa4, 0x00, 0x00, 0x00}, {0xa5, 0x00, 0x00, 0x00},
		{0xa7, 0x00, 0x00, 0x00}, {0xa8, 0x00, 0x00, 0x00},
		{0xac, 0x00, 0x00, 0x00}, {0xad, 0x00, 0x00, 0x00},
		{0xaf, 0x00, 0x00, 0x00}, {0xb1, 0x00, 0x00, 0x00},
		{0xb6, 0x00, 0x00, 0x00}, {0xb9, 0x00, 0x00, 0x00},
		{0xbf, 0x00, 0x00, 0x00}, {0xcf, 0x00, 0x00, 0x01}
	],
	[
		{0x00, 0x01, 0x93, 0x01}, {0x00, 0x01, 0x95, 0x01},
		{0x00, 0x01, 0x96, 0x01}, {0x00, 0x01, 0x97, 0x01},
		{0x00, 0x01, 0x98, 0x01}, {0x00, 0x01, 0x9b, 0x01},
		{0x00, 0x01, 0x9d, 0x01}, {0x00, 0x01, 0x9e, 0x01},
		{0x00, 0x01, 0xa5, 0x01}, {0x00, 0x01, 0xa6, 0x01},
		{0x00, 0x01, 0xa8, 0x01}, {0x00, 0x01, 0xae, 0x01},
		{0x00, 0x01, 0xaf, 0x01}, {0x00, 0x01, 0xb4, 0x01},
		{0x00, 0x01, 0xb6, 0x01}, {0x00, 0x01, 0xb7, 0x01}
	],
	/* 155 */
	[
		{0x01, 0x01, 0x93, 0x00}, {0x16, 0x01, 0x93, 0x01},
		{0x01, 0x01, 0x95, 0x00}, {0x16, 0x01, 0x95, 0x01},
		{0x01, 0x01, 0x96, 0x00}, {0x16, 0x01, 0x96, 0x01},
		{0x01, 0x01, 0x97, 0x00}, {0x16, 0x01, 0x97, 0x01},
		{0x01, 0x01, 0x98, 0x00}, {0x16, 0x01, 0x98, 0x01},
		{0x01, 0x01, 0x9b, 0x00}, {0x16, 0x01, 0x9b, 0x01},
		{0x01, 0x01, 0x9d, 0x00}, {0x16, 0x01, 0x9d, 0x01},
		{0x01, 0x01, 0x9e, 0x00}, {0x16, 0x01, 0x9e, 0x01}
	],
	[
		{0x02, 0x01, 0x93, 0x00}, {0x09, 0x01, 0x93, 0x00},
		{0x17, 0x01, 0x93, 0x00}, {0x28, 0x01, 0x93, 0x01},
		{0x02, 0x01, 0x95, 0x00}, {0x09, 0x01, 0x95, 0x00},
		{0x17, 0x01, 0x95, 0x00}, {0x28, 0x01, 0x95, 0x01},
		{0x02, 0x01, 0x96, 0x00}, {0x09, 0x01, 0x96, 0x00},
		{0x17, 0x01, 0x96, 0x00}, {0x28, 0x01, 0x96, 0x01},
		{0x02, 0x01, 0x97, 0x00}, {0x09, 0x01, 0x97, 0x00},
		{0x17, 0x01, 0x97, 0x00}, {0x28, 0x01, 0x97, 0x01}
	],
	[
		{0x03, 0x01, 0x93, 0x00}, {0x06, 0x01, 0x93, 0x00},
		{0x0a, 0x01, 0x93, 0x00}, {0x0f, 0x01, 0x93, 0x00},
		{0x18, 0x01, 0x93, 0x00}, {0x1f, 0x01, 0x93, 0x00},
		{0x29, 0x01, 0x93, 0x00}, {0x38, 0x01, 0x93, 0x01},
		{0x03, 0x01, 0x95, 0x00}, {0x06, 0x01, 0x95, 0x00},
		{0x0a, 0x01, 0x95, 0x00}, {0x0f, 0x01, 0x95, 0x00},
		{0x18, 0x01, 0x95, 0x00}, {0x1f, 0x01, 0x95, 0x00},
		{0x29, 0x01, 0x95, 0x00}, {0x38, 0x01, 0x95, 0x01}
	],
	[
		{0x03, 0x01, 0x96, 0x00}, {0x06, 0x01, 0x96, 0x00},
		{0x0a, 0x01, 0x96, 0x00}, {0x0f, 0x01, 0x96, 0x00},
		{0x18, 0x01, 0x96, 0x00}, {0x1f, 0x01, 0x96, 0x00},
		{0x29, 0x01, 0x96, 0x00}, {0x38, 0x01, 0x96, 0x01},
		{0x03, 0x01, 0x97, 0x00}, {0x06, 0x01, 0x97, 0x00},
		{0x0a, 0x01, 0x97, 0x00}, {0x0f, 0x01, 0x97, 0x00},
		{0x18, 0x01, 0x97, 0x00}, {0x1f, 0x01, 0x97, 0x00},
		{0x29, 0x01, 0x97, 0x00}, {0x38, 0x01, 0x97, 0x01}
	],
	[
		{0x02, 0x01, 0x98, 0x00}, {0x09, 0x01, 0x98, 0x00},
		{0x17, 0x01, 0x98, 0x00}, {0x28, 0x01, 0x98, 0x01},
		{0x02, 0x01, 0x9b, 0x00}, {0x09, 0x01, 0x9b, 0x00},
		{0x17, 0x01, 0x9b, 0x00}, {0x28, 0x01, 0x9b, 0x01},
		{0x02, 0x01, 0x9d, 0x00}, {0x09, 0x01, 0x9d, 0x00},
		{0x17, 0x01, 0x9d, 0x00}, {0x28, 0x01, 0x9d, 0x01},
		{0x02, 0x01, 0x9e, 0x00}, {0x09, 0x01, 0x9e, 0x00},
		{0x17, 0x01, 0x9e, 0x00}, {0x28, 0x01, 0x9e, 0x01}
	],
	/* 160 */
	[
		{0x03, 0x01, 0x98, 0x00}, {0x06, 0x01, 0x98, 0x00},
		{0x0a, 0x01, 0x98, 0x00}, {0x0f, 0x01, 0x98, 0x00},
		{0x18, 0x01, 0x98, 0x00}, {0x1f, 0x01, 0x98, 0x00},
		{0x29, 0x01, 0x98, 0x00}, {0x38, 0x01, 0x98, 0x01},
		{0x03, 0x01, 0x9b, 0x00}, {0x06, 0x01, 0x9b, 0x00},
		{0x0a, 0x01, 0x9b, 0x00}, {0x0f, 0x01, 0x9b, 0x00},
		{0x18, 0x01, 0x9b, 0x00}, {0x1f, 0x01, 0x9b, 0x00},
		{0x29, 0x01, 0x9b, 0x00}, {0x38, 0x01, 0x9b, 0x01}
	],
	[
		{0x03, 0x01, 0x9d, 0x00}, {0x06, 0x01, 0x9d, 0x00},
		{0x0a, 0x01, 0x9d, 0x00}, {0x0f, 0x01, 0x9d, 0x00},
		{0x18, 0x01, 0x9d, 0x00}, {0x1f, 0x01, 0x9d, 0x00},
		{0x29, 0x01, 0x9d, 0x00}, {0x38, 0x01, 0x9d, 0x01},
		{0x03, 0x01, 0x9e, 0x00}, {0x06, 0x01, 0x9e, 0x00},
		{0x0a, 0x01, 0x9e, 0x00}, {0x0f, 0x01, 0x9e, 0x00},
		{0x18, 0x01, 0x9e, 0x00}, {0x1f, 0x01, 0x9e, 0x00},
		{0x29, 0x01, 0x9e, 0x00}, {0x38, 0x01, 0x9e, 0x01}
	],
	[
		{0x01, 0x01, 0xa5, 0x00}, {0x16, 0x01, 0xa5, 0x01},
		{0x01, 0x01, 0xa6, 0x00}, {0x16, 0x01, 0xa6, 0x01},
		{0x01, 0x01, 0xa8, 0x00}, {0x16, 0x01, 0xa8, 0x01},
		{0x01, 0x01, 0xae, 0x00}, {0x16, 0x01, 0xae, 0x01},
		{0x01, 0x01, 0xaf, 0x00}, {0x16, 0x01, 0xaf, 0x01},
		{0x01, 0x01, 0xb4, 0x00}, {0x16, 0x01, 0xb4, 0x01},
		{0x01, 0x01, 0xb6, 0x00}, {0x16, 0x01, 0xb6, 0x01},
		{0x01, 0x01, 0xb7, 0x00}, {0x16, 0x01, 0xb7, 0x01}
	],
	[
		{0x02, 0x01, 0xa5, 0x00}, {0x09, 0x01, 0xa5, 0x00},
		{0x17, 0x01, 0xa5, 0x00}, {0x28, 0x01, 0xa5, 0x01},
		{0x02, 0x01, 0xa6, 0x00}, {0x09, 0x01, 0xa6, 0x00},
		{0x17, 0x01, 0xa6, 0x00}, {0x28, 0x01, 0xa6, 0x01},
		{0x02, 0x01, 0xa8, 0x00}, {0x09, 0x01, 0xa8, 0x00},
		{0x17, 0x01, 0xa8, 0x00}, {0x28, 0x01, 0xa8, 0x01},
		{0x02, 0x01, 0xae, 0x00}, {0x09, 0x01, 0xae, 0x00},
		{0x17, 0x01, 0xae, 0x00}, {0x28, 0x01, 0xae, 0x01}
	],
	[
		{0x03, 0x01, 0xa5, 0x00}, {0x06, 0x01, 0xa5, 0x00},
		{0x0a, 0x01, 0xa5, 0x00}, {0x0f, 0x01, 0xa5, 0x00},
		{0x18, 0x01, 0xa5, 0x00}, {0x1f, 0x01, 0xa5, 0x00},
		{0x29, 0x01, 0xa5, 0x00}, {0x38, 0x01, 0xa5, 0x01},
		{0x03, 0x01, 0xa6, 0x00}, {0x06, 0x01, 0xa6, 0x00},
		{0x0a, 0x01, 0xa6, 0x00}, {0x0f, 0x01, 0xa6, 0x00},
		{0x18, 0x01, 0xa6, 0x00}, {0x1f, 0x01, 0xa6, 0x00},
		{0x29, 0x01, 0xa6, 0x00}, {0x38, 0x01, 0xa6, 0x01}
	],
	/* 165 */
	[
		{0x03, 0x01, 0xa8, 0x00}, {0x06, 0x01, 0xa8, 0x00},
		{0x0a, 0x01, 0xa8, 0x00}, {0x0f, 0x01, 0xa8, 0x00},
		{0x18, 0x01, 0xa8, 0x00}, {0x1f, 0x01, 0xa8, 0x00},
		{0x29, 0x01, 0xa8, 0x00}, {0x38, 0x01, 0xa8, 0x01},
		{0x03, 0x01, 0xae, 0x00}, {0x06, 0x01, 0xae, 0x00},
		{0x0a, 0x01, 0xae, 0x00}, {0x0f, 0x01, 0xae, 0x00},
		{0x18, 0x01, 0xae, 0x00}, {0x1f, 0x01, 0xae, 0x00},
		{0x29, 0x01, 0xae, 0x00}, {0x38, 0x01, 0xae, 0x01}
	],
	[
		{0x02, 0x01, 0xaf, 0x00}, {0x09, 0x01, 0xaf, 0x00},
		{0x17, 0x01, 0xaf, 0x00}, {0x28, 0x01, 0xaf, 0x01},
		{0x02, 0x01, 0xb4, 0x00}, {0x09, 0x01, 0xb4, 0x00},
		{0x17, 0x01, 0xb4, 0x00}, {0x28, 0x01, 0xb4, 0x01},
		{0x02, 0x01, 0xb6, 0x00}, {0x09, 0x01, 0xb6, 0x00},
		{0x17, 0x01, 0xb6, 0x00}, {0x28, 0x01, 0xb6, 0x01},
		{0x02, 0x01, 0xb7, 0x00}, {0x09, 0x01, 0xb7, 0x00},
		{0x17, 0x01, 0xb7, 0x00}, {0x28, 0x01, 0xb7, 0x01}
	],
	[
		{0x03, 0x01, 0xaf, 0x00}, {0x06, 0x01, 0xaf, 0x00},
		{0x0a, 0x01, 0xaf, 0x00}, {0x0f, 0x01, 0xaf, 0x00},
		{0x18, 0x01, 0xaf, 0x00}, {0x1f, 0x01, 0xaf, 0x00},
		{0x29, 0x01, 0xaf, 0x00}, {0x38, 0x01, 0xaf, 0x01},
		{0x03, 0x01, 0xb4, 0x00}, {0x06, 0x01, 0xb4, 0x00},
		{0x0a, 0x01, 0xb4, 0x00}, {0x0f, 0x01, 0xb4, 0x00},
		{0x18, 0x01, 0xb4, 0x00}, {0x1f, 0x01, 0xb4, 0x00},
		{0x29, 0x01, 0xb4, 0x00}, {0x38, 0x01, 0xb4, 0x01}
	],
	[
		{0x03, 0x01, 0xb6, 0x00}, {0x06, 0x01, 0xb6, 0x00},
		{0x0a, 0x01, 0xb6, 0x00}, {0x0f, 0x01, 0xb6, 0x00},
		{0x18, 0x01, 0xb6, 0x00}, {0x1f, 0x01, 0xb6, 0x00},
		{0x29, 0x01, 0xb6, 0x00}, {0x38, 0x01, 0xb6, 0x01},
		{0x03, 0x01, 0xb7, 0x00}, {0x06, 0x01, 0xb7, 0x00},
		{0x0a, 0x01, 0xb7, 0x00}, {0x0f, 0x01, 0xb7, 0x00},
		{0x18, 0x01, 0xb7, 0x00}, {0x1f, 0x01, 0xb7, 0x00},
		{0x29, 0x01, 0xb7, 0x00}, {0x38, 0x01, 0xb7, 0x01}
	],
	[
		{0x00, 0x01, 0xbc, 0x01}, {0x00, 0x01, 0xbf, 0x01},
		{0x00, 0x01, 0xc5, 0x01}, {0x00, 0x01, 0xe7, 0x01},
		{0x00, 0x01, 0xef, 0x01}, {0xb0, 0x00, 0x00, 0x00},
		{0xb2, 0x00, 0x00, 0x00}, {0xb3, 0x00, 0x00, 0x00},
		{0xb7, 0x00, 0x00, 0x00}, {0xb8, 0x00, 0x00, 0x00},
		{0xba, 0x00, 0x00, 0x00}, {0xbb, 0x00, 0x00, 0x00},
		{0xc0, 0x00, 0x00, 0x00}, {0xc7, 0x00, 0x00, 0x00},
		{0xd0, 0x00, 0x00, 0x00}, {0xdf, 0x00, 0x00, 0x01}
	],
	/* 170 */
	[
		{0x01, 0x01, 0xbc, 0x00}, {0x16, 0x01, 0xbc, 0x01},
		{0x01, 0x01, 0xbf, 0x00}, {0x16, 0x01, 0xbf, 0x01},
		{0x01, 0x01, 0xc5, 0x00}, {0x16, 0x01, 0xc5, 0x01},
		{0x01, 0x01, 0xe7, 0x00}, {0x16, 0x01, 0xe7, 0x01},
		{0x01, 0x01, 0xef, 0x00}, {0x16, 0x01, 0xef, 0x01},
		{0x00, 0x01, 0x09, 0x01}, {0x00, 0x01, 0x8e, 0x01},
		{0x00, 0x01, 0x90, 0x01}, {0x00, 0x01, 0x91, 0x01},
		{0x00, 0x01, 0x94, 0x01}, {0x00, 0x01, 0x9f, 0x01}
	],
	[
		{0x02, 0x01, 0xbc, 0x00}, {0x09, 0x01, 0xbc, 0x00},
		{0x17, 0x01, 0xbc, 0x00}, {0x28, 0x01, 0xbc, 0x01},
		{0x02, 0x01, 0xbf, 0x00}, {0x09, 0x01, 0xbf, 0x00},
		{0x17, 0x01, 0xbf, 0x00}, {0x28, 0x01, 0xbf, 0x01},
		{0x02, 0x01, 0xc5, 0x00}, {0x09, 0x01, 0xc5, 0x00},
		{0x17, 0x01, 0xc5, 0x00}, {0x28, 0x01, 0xc5, 0x01},
		{0x02, 0x01, 0xe7, 0x00}, {0x09, 0x01, 0xe7, 0x00},
		{0x17, 0x01, 0xe7, 0x00}, {0x28, 0x01, 0xe7, 0x01}
	],
	[
		{0x03, 0x01, 0xbc, 0x00}, {0x06, 0x01, 0xbc, 0x00},
		{0x0a, 0x01, 0xbc, 0x00}, {0x0f, 0x01, 0xbc, 0x00},
		{0x18, 0x01, 0xbc, 0x00}, {0x1f, 0x01, 0xbc, 0x00},
		{0x29, 0x01, 0xbc, 0x00}, {0x38, 0x01, 0xbc, 0x01},
		{0x03, 0x01, 0xbf, 0x00}, {0x06, 0x01, 0xbf, 0x00},
		{0x0a, 0x01, 0xbf, 0x00}, {0x0f, 0x01, 0xbf, 0x00},
		{0x18, 0x01, 0xbf, 0x00}, {0x1f, 0x01, 0xbf, 0x00},
		{0x29, 0x01, 0xbf, 0x00}, {0x38, 0x01, 0xbf, 0x01}
	],
	[
		{0x03, 0x01, 0xc5, 0x00}, {0x06, 0x01, 0xc5, 0x00},
		{0x0a, 0x01, 0xc5, 0x00}, {0x0f, 0x01, 0xc5, 0x00},
		{0x18, 0x01, 0xc5, 0x00}, {0x1f, 0x01, 0xc5, 0x00},
		{0x29, 0x01, 0xc5, 0x00}, {0x38, 0x01, 0xc5, 0x01},
		{0x03, 0x01, 0xe7, 0x00}, {0x06, 0x01, 0xe7, 0x00},
		{0x0a, 0x01, 0xe7, 0x00}, {0x0f, 0x01, 0xe7, 0x00},
		{0x18, 0x01, 0xe7, 0x00}, {0x1f, 0x01, 0xe7, 0x00},
		{0x29, 0x01, 0xe7, 0x00}, {0x38, 0x01, 0xe7, 0x01}
	],
	[
		{0x02, 0x01, 0xef, 0x00}, {0x09, 0x01, 0xef, 0x00},
		{0x17, 0x01, 0xef, 0x00}, {0x28, 0x01, 0xef, 0x01},
		{0x01, 0x01, 0x09, 0x00}, {0x16, 0x01, 0x09, 0x01},
		{0x01, 0x01, 0x8e, 0x00}, {0x16, 0x01, 0x8e, 0x01},
		{0x01, 0x01, 0x90, 0x00}, {0x16, 0x01, 0x90, 0x01},
		{0x01, 0x01, 0x91, 0x00}, {0x16, 0x01, 0x91, 0x01},
		{0x01, 0x01, 0x94, 0x00}, {0x16, 0x01, 0x94, 0x01},
		{0x01, 0x01, 0x9f, 0x00}, {0x16, 0x01, 0x9f, 0x01}
	],
	/* 175 */
	[
		{0x03, 0x01, 0xef, 0x00}, {0x06, 0x01, 0xef, 0x00},
		{0x0a, 0x01, 0xef, 0x00}, {0x0f, 0x01, 0xef, 0x00},
		{0x18, 0x01, 0xef, 0x00}, {0x1f, 0x01, 0xef, 0x00},
		{0x29, 0x01, 0xef, 0x00}, {0x38, 0x01, 0xef, 0x01},
		{0x02, 0x01, 0x09, 0x00}, {0x09, 0x01, 0x09, 0x00},
		{0x17, 0x01, 0x09, 0x00}, {0x28, 0x01, 0x09, 0x01},
		{0x02, 0x01, 0x8e, 0x00}, {0x09, 0x01, 0x8e, 0x00},
		{0x17, 0x01, 0x8e, 0x00}, {0x28, 0x01, 0x8e, 0x01}
	],
	[
		{0x03, 0x01, 0x09, 0x00}, {0x06, 0x01, 0x09, 0x00},
		{0x0a, 0x01, 0x09, 0x00}, {0x0f, 0x01, 0x09, 0x00},
		{0x18, 0x01, 0x09, 0x00}, {0x1f, 0x01, 0x09, 0x00},
		{0x29, 0x01, 0x09, 0x00}, {0x38, 0x01, 0x09, 0x01},
		{0x03, 0x01, 0x8e, 0x00}, {0x06, 0x01, 0x8e, 0x00},
		{0x0a, 0x01, 0x8e, 0x00}, {0x0f, 0x01, 0x8e, 0x00},
		{0x18, 0x01, 0x8e, 0x00}, {0x1f, 0x01, 0x8e, 0x00},
		{0x29, 0x01, 0x8e, 0x00}, {0x38, 0x01, 0x8e, 0x01}
	],
	[
		{0x02, 0x01, 0x90, 0x00}, {0x09, 0x01, 0x90, 0x00},
		{0x17, 0x01, 0x90, 0x00}, {0x28, 0x01, 0x90, 0x01},
		{0x02, 0x01, 0x91, 0x00}, {0x09, 0x01, 0x91, 0x00},
		{0x17, 0x01, 0x91, 0x00}, {0x28, 0x01, 0x91, 0x01},
		{0x02, 0x01, 0x94, 0x00}, {0x09, 0x01, 0x94, 0x00},
		{0x17, 0x01, 0x94, 0x00}, {0x28, 0x01, 0x94, 0x01},
		{0x02, 0x01, 0x9f, 0x00}, {0x09, 0x01, 0x9f, 0x00},
		{0x17, 0x01, 0x9f, 0x00}, {0x28, 0x01, 0x9f, 0x01}
	],
	[
		{0x03, 0x01, 0x90, 0x00}, {0x06, 0x01, 0x90, 0x00},
		{0x0a, 0x01, 0x90, 0x00}, {0x0f, 0x01, 0x90, 0x00},
		{0x18, 0x01, 0x90, 0x00}, {0x1f, 0x01, 0x90, 0x00},
		{0x29, 0x01, 0x90, 0x00}, {0x38, 0x01, 0x90, 0x01},
		{0x03, 0x01, 0x91, 0x00}, {0x06, 0x01, 0x91, 0x00},
		{0x0a, 0x01, 0x91, 0x00}, {0x0f, 0x01, 0x91, 0x00},
		{0x18, 0x01, 0x91, 0x00}, {0x1f, 0x01, 0x91, 0x00},
		{0x29, 0x01, 0x91, 0x00}, {0x38, 0x01, 0x91, 0x01}
	],
	[
		{0x03, 0x01, 0x94, 0x00}, {0x06, 0x01, 0x94, 0x00},
		{0x0a, 0x01, 0x94, 0x00}, {0x0f, 0x01, 0x94, 0x00},
		{0x18, 0x01, 0x94, 0x00}, {0x1f, 0x01, 0x94, 0x00},
		{0x29, 0x01, 0x94, 0x00}, {0x38, 0x01, 0x94, 0x01},
		{0x03, 0x01, 0x9f, 0x00}, {0x06, 0x01, 0x9f, 0x00},
		{0x0a, 0x01, 0x9f, 0x00}, {0x0f, 0x01, 0x9f, 0x00},
		{0x18, 0x01, 0x9f, 0x00}, {0x1f, 0x01, 0x9f, 0x00},
		{0x29, 0x01, 0x9f, 0x00}, {0x38, 0x01, 0x9f, 0x01}
	],
	/* 180 */
	[
		{0x00, 0x01, 0xab, 0x01}, {0x00, 0x01, 0xce, 0x01},
		{0x00, 0x01, 0xd7, 0x01}, {0x00, 0x01, 0xe1, 0x01},
		{0x00, 0x01, 0xec, 0x01}, {0x00, 0x01, 0xed, 0x01},
		{0xbc, 0x00, 0x00, 0x00}, {0xbd, 0x00, 0x00, 0x00},
		{0xc1, 0x00, 0x00, 0x00}, {0xc4, 0x00, 0x00, 0x00},
		{0xc8, 0x00, 0x00, 0x00}, {0xcb, 0x00, 0x00, 0x00},
		{0xd1, 0x00, 0x00, 0x00}, {0xd8, 0x00, 0x00, 0x00},
		{0xe0, 0x00, 0x00, 0x00}, {0xee, 0x00, 0x00, 0x01}
	],
	[
		{0x01, 0x01, 0xab, 0x00}, {0x16, 0x01, 0xab, 0x01},
		{0x01, 0x01, 0xce, 0x00}, {0x16, 0x01, 0xce, 0x01},
		{0x01, 0x01, 0xd7, 0x00}, {0x16, 0x01, 0xd7, 0x01},
		{0x01, 0x01, 0xe1, 0x00}, {0x16, 0x01, 0xe1, 0x01},
		{0x01, 0x01, 0xec, 0x00}, {0x16, 0x01, 0xec, 0x01},
		{0x01, 0x01, 0xed, 0x00}, {0x16, 0x01, 0xed, 0x01},
		{0x00, 0x01, 0xc7, 0x01}, {0x00, 0x01, 0xcf, 0x01},
		{0x00, 0x01, 0xea, 0x01}, {0x00, 0x01, 0xeb, 0x01}
	],
	[
		{0x02, 0x01, 0xab, 0x00}, {0x09, 0x01, 0xab, 0x00},
		{0x17, 0x01, 0xab, 0x00}, {0x28, 0x01, 0xab, 0x01},
		{0x02, 0x01, 0xce, 0x00}, {0x09, 0x01, 0xce, 0x00},
		{0x17, 0x01, 0xce, 0x00}, {0x28, 0x01, 0xce, 0x01},
		{0x02, 0x01, 0xd7, 0x00}, {0x09, 0x01, 0xd7, 0x00},
		{0x17, 0x01, 0xd7, 0x00}, {0x28, 0x01, 0xd7, 0x01},
		{0x02, 0x01, 0xe1, 0x00}, {0x09, 0x01, 0xe1, 0x00},
		{0x17, 0x01, 0xe1, 0x00}, {0x28, 0x01, 0xe1, 0x01}
	],
	[
		{0x03, 0x01, 0xab, 0x00}, {0x06, 0x01, 0xab, 0x00},
		{0x0a, 0x01, 0xab, 0x00}, {0x0f, 0x01, 0xab, 0x00},
		{0x18, 0x01, 0xab, 0x00}, {0x1f, 0x01, 0xab, 0x00},
		{0x29, 0x01, 0xab, 0x00}, {0x38, 0x01, 0xab, 0x01},
		{0x03, 0x01, 0xce, 0x00}, {0x06, 0x01, 0xce, 0x00},
		{0x0a, 0x01, 0xce, 0x00}, {0x0f, 0x01, 0xce, 0x00},
		{0x18, 0x01, 0xce, 0x00}, {0x1f, 0x01, 0xce, 0x00},
		{0x29, 0x01, 0xce, 0x00}, {0x38, 0x01, 0xce, 0x01}
	],
	[
		{0x03, 0x01, 0xd7, 0x00}, {0x06, 0x01, 0xd7, 0x00},
		{0x0a, 0x01, 0xd7, 0x00}, {0x0f, 0x01, 0xd7, 0x00},
		{0x18, 0x01, 0xd7, 0x00}, {0x1f, 0x01, 0xd7, 0x00},
		{0x29, 0x01, 0xd7, 0x00}, {0x38, 0x01, 0xd7, 0x01},
		{0x03, 0x01, 0xe1, 0x00}, {0x06, 0x01, 0xe1, 0x00},
		{0x0a, 0x01, 0xe1, 0x00}, {0x0f, 0x01, 0xe1, 0x00},
		{0x18, 0x01, 0xe1, 0x00}, {0x1f, 0x01, 0xe1, 0x00},
		{0x29, 0x01, 0xe1, 0x00}, {0x38, 0x01, 0xe1, 0x01}
	],
	/* 185 */
	[
		{0x02, 0x01, 0xec, 0x00}, {0x09, 0x01, 0xec, 0x00},
		{0x17, 0x01, 0xec, 0x00}, {0x28, 0x01, 0xec, 0x01},
		{0x02, 0x01, 0xed, 0x00}, {0x09, 0x01, 0xed, 0x00},
		{0x17, 0x01, 0xed, 0x00}, {0x28, 0x01, 0xed, 0x01},
		{0x01, 0x01, 0xc7, 0x00}, {0x16, 0x01, 0xc7, 0x01},
		{0x01, 0x01, 0xcf, 0x00}, {0x16, 0x01, 0xcf, 0x01},
		{0x01, 0x01, 0xea, 0x00}, {0x16, 0x01, 0xea, 0x01},
		{0x01, 0x01, 0xeb, 0x00}, {0x16, 0x01, 0xeb, 0x01}
	],
	[
		{0x03, 0x01, 0xec, 0x00}, {0x06, 0x01, 0xec, 0x00},
		{0x0a, 0x01, 0xec, 0x00}, {0x0f, 0x01, 0xec, 0x00},
		{0x18, 0x01, 0xec, 0x00}, {0x1f, 0x01, 0xec, 0x00},
		{0x29, 0x01, 0xec, 0x00}, {0x38, 0x01, 0xec, 0x01},
		{0x03, 0x01, 0xed, 0x00}, {0x06, 0x01, 0xed, 0x00},
		{0x0a, 0x01, 0xed, 0x00}, {0x0f, 0x01, 0xed, 0x00},
		{0x18, 0x01, 0xed, 0x00}, {0x1f, 0x01, 0xed, 0x00},
		{0x29, 0x01, 0xed, 0x00}, {0x38, 0x01, 0xed, 0x01}
	],
	[
		{0x02, 0x01, 0xc7, 0x00}, {0x09, 0x01, 0xc7, 0x00},
		{0x17, 0x01, 0xc7, 0x00}, {0x28, 0x01, 0xc7, 0x01},
		{0x02, 0x01, 0xcf, 0x00}, {0x09, 0x01, 0xcf, 0x00},
		{0x17, 0x01, 0xcf, 0x00}, {0x28, 0x01, 0xcf, 0x01},
		{0x02, 0x01, 0xea, 0x00}, {0x09, 0x01, 0xea, 0x00},
		{0x17, 0x01, 0xea, 0x00}, {0x28, 0x01, 0xea, 0x01},
		{0x02, 0x01, 0xeb, 0x00}, {0x09, 0x01, 0xeb, 0x00},
		{0x17, 0x01, 0xeb, 0x00}, {0x28, 0x01, 0xeb, 0x01}
	],
	[
		{0x03, 0x01, 0xc7, 0x00}, {0x06, 0x01, 0xc7, 0x00},
		{0x0a, 0x01, 0xc7, 0x00}, {0x0f, 0x01, 0xc7, 0x00},
		{0x18, 0x01, 0xc7, 0x00}, {0x1f, 0x01, 0xc7, 0x00},
		{0x29, 0x01, 0xc7, 0x00}, {0x38, 0x01, 0xc7, 0x01},
		{0x03, 0x01, 0xcf, 0x00}, {0x06, 0x01, 0xcf, 0x00},
		{0x0a, 0x01, 0xcf, 0x00}, {0x0f, 0x01, 0xcf, 0x00},
		{0x18, 0x01, 0xcf, 0x00}, {0x1f, 0x01, 0xcf, 0x00},
		{0x29, 0x01, 0xcf, 0x00}, {0x38, 0x01, 0xcf, 0x01}
	],
	[
		{0x03, 0x01, 0xea, 0x00}, {0x06, 0x01, 0xea, 0x00},
		{0x0a, 0x01, 0xea, 0x00}, {0x0f, 0x01, 0xea, 0x00},
		{0x18, 0x01, 0xea, 0x00}, {0x1f, 0x01, 0xea, 0x00},
		{0x29, 0x01, 0xea, 0x00}, {0x38, 0x01, 0xea, 0x01},
		{0x03, 0x01, 0xeb, 0x00}, {0x06, 0x01, 0xeb, 0x00},
		{0x0a, 0x01, 0xeb, 0x00}, {0x0f, 0x01, 0xeb, 0x00},
		{0x18, 0x01, 0xeb, 0x00}, {0x1f, 0x01, 0xeb, 0x00},
		{0x29, 0x01, 0xeb, 0x00}, {0x38, 0x01, 0xeb, 0x01}
	],
	/* 190 */
	[
		{0xc2, 0x00, 0x00, 0x00}, {0xc3, 0x00, 0x00, 0x00},
		{0xc5, 0x00, 0x00, 0x00}, {0xc6, 0x00, 0x00, 0x00},
		{0xc9, 0x00, 0x00, 0x00}, {0xca, 0x00, 0x00, 0x00},
		{0xcc, 0x00, 0x00, 0x00}, {0xcd, 0x00, 0x00, 0x00},
		{0xd2, 0x00, 0x00, 0x00}, {0xd5, 0x00, 0x00, 0x00},
		{0xd9, 0x00, 0x00, 0x00}, {0xdc, 0x00, 0x00, 0x00},
		{0xe1, 0x00, 0x00, 0x00}, {0xe7, 0x00, 0x00, 0x00},
		{0xef, 0x00, 0x00, 0x00}, {0xf6, 0x00, 0x00, 0x01}
	],
	[
		{0x00, 0x01, 0xc0, 0x01}, {0x00, 0x01, 0xc1, 0x01},
		{0x00, 0x01, 0xc8, 0x01}, {0x00, 0x01, 0xc9, 0x01},
		{0x00, 0x01, 0xca, 0x01}, {0x00, 0x01, 0xcd, 0x01},
		{0x00, 0x01, 0xd2, 0x01}, {0x00, 0x01, 0xd5, 0x01},
		{0x00, 0x01, 0xda, 0x01}, {0x00, 0x01, 0xdb, 0x01},
		{0x00, 0x01, 0xee, 0x01}, {0x00, 0x01, 0xf0, 0x01},
		{0x00, 0x01, 0xf2, 0x01}, {0x00, 0x01, 0xf3, 0x01},
		{0x00, 0x01, 0xff, 0x01}, {0xce, 0x00, 0x00, 0x00}
	],
	[
		{0x01, 0x01, 0xc0, 0x00}, {0x16, 0x01, 0xc0, 0x01},
		{0x01, 0x01, 0xc1, 0x00}, {0x16, 0x01, 0xc1, 0x01},
		{0x01, 0x01, 0xc8, 0x00}, {0x16, 0x01, 0xc8, 0x01},
		{0x01, 0x01, 0xc9, 0x00}, {0x16, 0x01, 0xc9, 0x01},
		{0x01, 0x01, 0xca, 0x00}, {0x16, 0x01, 0xca, 0x01},
		{0x01, 0x01, 0xcd, 0x00}, {0x16, 0x01, 0xcd, 0x01},
		{0x01, 0x01, 0xd2, 0x00}, {0x16, 0x01, 0xd2, 0x01},
		{0x01, 0x01, 0xd5, 0x00}, {0x16, 0x01, 0xd5, 0x01}
	],
	[
		{0x02, 0x01, 0xc0, 0x00}, {0x09, 0x01, 0xc0, 0x00},
		{0x17, 0x01, 0xc0, 0x00}, {0x28, 0x01, 0xc0, 0x01},
		{0x02, 0x01, 0xc1, 0x00}, {0x09, 0x01, 0xc1, 0x00},
		{0x17, 0x01, 0xc1, 0x00}, {0x28, 0x01, 0xc1, 0x01},
		{0x02, 0x01, 0xc8, 0x00}, {0x09, 0x01, 0xc8, 0x00},
		{0x17, 0x01, 0xc8, 0x00}, {0x28, 0x01, 0xc8, 0x01},
		{0x02, 0x01, 0xc9, 0x00}, {0x09, 0x01, 0xc9, 0x00},
		{0x17, 0x01, 0xc9, 0x00}, {0x28, 0x01, 0xc9, 0x01}
	],
	[
		{0x03, 0x01, 0xc0, 0x00}, {0x06, 0x01, 0xc0, 0x00},
		{0x0a, 0x01, 0xc0, 0x00}, {0x0f, 0x01, 0xc0, 0x00},
		{0x18, 0x01, 0xc0, 0x00}, {0x1f, 0x01, 0xc0, 0x00},
		{0x29, 0x01, 0xc0, 0x00}, {0x38, 0x01, 0xc0, 0x01},
		{0x03, 0x01, 0xc1, 0x00}, {0x06, 0x01, 0xc1, 0x00},
		{0x0a, 0x01, 0xc1, 0x00}, {0x0f, 0x01, 0xc1, 0x00},
		{0x18, 0x01, 0xc1, 0x00}, {0x1f, 0x01, 0xc1, 0x00},
		{0x29, 0x01, 0xc1, 0x00}, {0x38, 0x01, 0xc1, 0x01}
	],
	/* 195 */
	[
		{0x03, 0x01, 0xc8, 0x00}, {0x06, 0x01, 0xc8, 0x00},
		{0x0a, 0x01, 0xc8, 0x00}, {0x0f, 0x01, 0xc8, 0x00},
		{0x18, 0x01, 0xc8, 0x00}, {0x1f, 0x01, 0xc8, 0x00},
		{0x29, 0x01, 0xc8, 0x00}, {0x38, 0x01, 0xc8, 0x01},
		{0x03, 0x01, 0xc9, 0x00}, {0x06, 0x01, 0xc9, 0x00},
		{0x0a, 0x01, 0xc9, 0x00}, {0x0f, 0x01, 0xc9, 0x00},
		{0x18, 0x01, 0xc9, 0x00}, {0x1f, 0x01, 0xc9, 0x00},
		{0x29, 0x01, 0xc9, 0x00}, {0x38, 0x01, 0xc9, 0x01}
	],
	[
		{0x02, 0x01, 0xca, 0x00}, {0x09, 0x01, 0xca, 0x00},
		{0x17, 0x01, 0xca, 0x00}, {0x28, 0x01, 0xca, 0x01},
		{0x02, 0x01, 0xcd, 0x00}, {0x09, 0x01, 0xcd, 0x00},
		{0x17, 0x01, 0xcd, 0x00}, {0x28, 0x01, 0xcd, 0x01},
		{0x02, 0x01, 0xd2, 0x00}, {0x09, 0x01, 0xd2, 0x00},
		{0x17, 0x01, 0xd2, 0x00}, {0x28, 0x01, 0xd2, 0x01},
		{0x02, 0x01, 0xd5, 0x00}, {0x09, 0x01, 0xd5, 0x00},
		{0x17, 0x01, 0xd5, 0x00}, {0x28, 0x01, 0xd5, 0x01}
	],
	[
		{0x03, 0x01, 0xca, 0x00}, {0x06, 0x01, 0xca, 0x00},
		{0x0a, 0x01, 0xca, 0x00}, {0x0f, 0x01, 0xca, 0x00},
		{0x18, 0x01, 0xca, 0x00}, {0x1f, 0x01, 0xca, 0x00},
		{0x29, 0x01, 0xca, 0x00}, {0x38, 0x01, 0xca, 0x01},
		{0x03, 0x01, 0xcd, 0x00}, {0x06, 0x01, 0xcd, 0x00},
		{0x0a, 0x01, 0xcd, 0x00}, {0x0f, 0x01, 0xcd, 0x00},
		{0x18, 0x01, 0xcd, 0x00}, {0x1f, 0x01, 0xcd, 0x00},
		{0x29, 0x01, 0xcd, 0x00}, {0x38, 0x01, 0xcd, 0x01}
	],
	[
		{0x03, 0x01, 0xd2, 0x00}, {0x06, 0x01, 0xd2, 0x00},
		{0x0a, 0x01, 0xd2, 0x00}, {0x0f, 0x01, 0xd2, 0x00},
		{0x18, 0x01, 0xd2, 0x00}, {0x1f, 0x01, 0xd2, 0x00},
		{0x29, 0x01, 0xd2, 0x00}, {0x38, 0x01, 0xd2, 0x01},
		{0x03, 0x01, 0xd5, 0x00}, {0x06, 0x01, 0xd5, 0x00},
		{0x0a, 0x01, 0xd5, 0x00}, {0x0f, 0x01, 0xd5, 0x00},
		{0x18, 0x01, 0xd5, 0x00}, {0x1f, 0x01, 0xd5, 0x00},
		{0x29, 0x01, 0xd5, 0x00}, {0x38, 0x01, 0xd5, 0x01}
	],
	[
		{0x01, 0x01, 0xda, 0x00}, {0x16, 0x01, 0xda, 0x01},
		{0x01, 0x01, 0xdb, 0x00}, {0x16, 0x01, 0xdb, 0x01},
		{0x01, 0x01, 0xee, 0x00}, {0x16, 0x01, 0xee, 0x01},
		{0x01, 0x01, 0xf0, 0x00}, {0x16, 0x01, 0xf0, 0x01},
		{0x01, 0x01, 0xf2, 0x00}, {0x16, 0x01, 0xf2, 0x01},
		{0x01, 0x01, 0xf3, 0x00}, {0x16, 0x01, 0xf3, 0x01},
		{0x01, 0x01, 0xff, 0x00}, {0x16, 0x01, 0xff, 0x01},
		{0x00, 0x01, 0xcb, 0x01}, {0x00, 0x01, 0xcc, 0x01}
	],
	/* 200 */
	[
		{0x02, 0x01, 0xda, 0x00}, {0x09, 0x01, 0xda, 0x00},
		{0x17, 0x01, 0xda, 0x00}, {0x28, 0x01, 0xda, 0x01},
		{0x02, 0x01, 0xdb, 0x00}, {0x09, 0x01, 0xdb, 0x00},
		{0x17, 0x01, 0xdb, 0x00}, {0x28, 0x01, 0xdb, 0x01},
		{0x02, 0x01, 0xee, 0x00}, {0x09, 0x01, 0xee, 0x00},
		{0x17, 0x01, 0xee, 0x00}, {0x28, 0x01, 0xee, 0x01},
		{0x02, 0x01, 0xf0, 0x00}, {0x09, 0x01, 0xf0, 0x00},
		{0x17, 0x01, 0xf0, 0x00}, {0x28, 0x01, 0xf0, 0x01}
	],
	[
		{0x03, 0x01, 0xda, 0x00}, {0x06, 0x01, 0xda, 0x00},
		{0x0a, 0x01, 0xda, 0x00}, {0x0f, 0x01, 0xda, 0x00},
		{0x18, 0x01, 0xda, 0x00}, {0x1f, 0x01, 0xda, 0x00},
		{0x29, 0x01, 0xda, 0x00}, {0x38, 0x01, 0xda, 0x01},
		{0x03, 0x01, 0xdb, 0x00}, {0x06, 0x01, 0xdb, 0x00},
		{0x0a, 0x01, 0xdb, 0x00}, {0x0f, 0x01, 0xdb, 0x00},
		{0x18, 0x01, 0xdb, 0x00}, {0x1f, 0x01, 0xdb, 0x00},
		{0x29, 0x01, 0xdb, 0x00}, {0x38, 0x01, 0xdb, 0x01}
	],
	[
		{0x03, 0x01, 0xee, 0x00}, {0x06, 0x01, 0xee, 0x00},
		{0x0a, 0x01, 0xee, 0x00}, {0x0f, 0x01, 0xee, 0x00},
		{0x18, 0x01, 0xee, 0x00}, {0x1f, 0x01, 0xee, 0x00},
		{0x29, 0x01, 0xee, 0x00}, {0x38, 0x01, 0xee, 0x01},
		{0x03, 0x01, 0xf0, 0x00}, {0x06, 0x01, 0xf0, 0x00},
		{0x0a, 0x01, 0xf0, 0x00}, {0x0f, 0x01, 0xf0, 0x00},
		{0x18, 0x01, 0xf0, 0x00}, {0x1f, 0x01, 0xf0, 0x00},
		{0x29, 0x01, 0xf0, 0x00}, {0x38, 0x01, 0xf0, 0x01}
	],
	[
		{0x02, 0x01, 0xf2, 0x00}, {0x09, 0x01, 0xf2, 0x00},
		{0x17, 0x01, 0xf2, 0x00}, {0x28, 0x01, 0xf2, 0x01},
		{0x02, 0x01, 0xf3, 0x00}, {0x09, 0x01, 0xf3, 0x00},
		{0x17, 0x01, 0xf3, 0x00}, {0x28, 0x01, 0xf3, 0x01},
		{0x02, 0x01, 0xff, 0x00}, {0x09, 0x01, 0xff, 0x00},
		{0x17, 0x01, 0xff, 0x00}, {0x28, 0x01, 0xff, 0x01},
		{0x01, 0x01, 0xcb, 0x00}, {0x16, 0x01, 0xcb, 0x01},
		{0x01, 0x01, 0xcc, 0x00}, {0x16, 0x01, 0xcc, 0x01}
	],
	[
		{0x03, 0x01, 0xf2, 0x00}, {0x06, 0x01, 0xf2, 0x00},
		{0x0a, 0x01, 0xf2, 0x00}, {0x0f, 0x01, 0xf2, 0x00},
		{0x18, 0x01, 0xf2, 0x00}, {0x1f, 0x01, 0xf2, 0x00},
		{0x29, 0x01, 0xf2, 0x00}, {0x38, 0x01, 0xf2, 0x01},
		{0x03, 0x01, 0xf3, 0x00}, {0x06, 0x01, 0xf3, 0x00},
		{0x0a, 0x01, 0xf3, 0x00}, {0x0f, 0x01, 0xf3, 0x00},
		{0x18, 0x01, 0xf3, 0x00}, {0x1f, 0x01, 0xf3, 0x00},
		{0x29, 0x01, 0xf3, 0x00}, {0x38, 0x01, 0xf3, 0x01}
	],
	/* 205 */
	[
		{0x03, 0x01, 0xff, 0x00}, {0x06, 0x01, 0xff, 0x00},
		{0x0a, 0x01, 0xff, 0x00}, {0x0f, 0x01, 0xff, 0x00},
		{0x18, 0x01, 0xff, 0x00}, {0x1f, 0x01, 0xff, 0x00},
		{0x29, 0x01, 0xff, 0x00}, {0x38, 0x01, 0xff, 0x01},
		{0x02, 0x01, 0xcb, 0x00}, {0x09, 0x01, 0xcb, 0x00},
		{0x17, 0x01, 0xcb, 0x00}, {0x28, 0x01, 0xcb, 0x01},
		{0x02, 0x01, 0xcc, 0x00}, {0x09, 0x01, 0xcc, 0x00},
		{0x17, 0x01, 0xcc, 0x00}, {0x28, 0x01, 0xcc, 0x01}
	],
	[
		{0x03, 0x01, 0xcb, 0x00}, {0x06, 0x01, 0xcb, 0x00},
		{0x0a, 0x01, 0xcb, 0x00}, {0x0f, 0x01, 0xcb, 0x00},
		{0x18, 0x01, 0xcb, 0x00}, {0x1f, 0x01, 0xcb, 0x00},
		{0x29, 0x01, 0xcb, 0x00}, {0x38, 0x01, 0xcb, 0x01},
		{0x03, 0x01, 0xcc, 0x00}, {0x06, 0x01, 0xcc, 0x00},
		{0x0a, 0x01, 0xcc, 0x00}, {0x0f, 0x01, 0xcc, 0x00},
		{0x18, 0x01, 0xcc, 0x00}, {0x1f, 0x01, 0xcc, 0x00},
		{0x29, 0x01, 0xcc, 0x00}, {0x38, 0x01, 0xcc, 0x01}
	],
	[
		{0xd3, 0x00, 0x00, 0x00}, {0xd4, 0x00, 0x00, 0x00},
		{0xd6, 0x00, 0x00, 0x00}, {0xd7, 0x00, 0x00, 0x00},
		{0xda, 0x00, 0x00, 0x00}, {0xdb, 0x00, 0x00, 0x00},
		{0xdd, 0x00, 0x00, 0x00}, {0xde, 0x00, 0x00, 0x00},
		{0xe2, 0x00, 0x00, 0x00}, {0xe4, 0x00, 0x00, 0x00},
		{0xe8, 0x00, 0x00, 0x00}, {0xeb, 0x00, 0x00, 0x00},
		{0xf0, 0x00, 0x00, 0x00}, {0xf3, 0x00, 0x00, 0x00},
		{0xf7, 0x00, 0x00, 0x00}, {0xfa, 0x00, 0x00, 0x01}
	],
	[
		{0x00, 0x01, 0xd3, 0x01}, {0x00, 0x01, 0xd4, 0x01},
		{0x00, 0x01, 0xd6, 0x01}, {0x00, 0x01, 0xdd, 0x01},
		{0x00, 0x01, 0xde, 0x01}, {0x00, 0x01, 0xdf, 0x01},
		{0x00, 0x01, 0xf1, 0x01}, {0x00, 0x01, 0xf4, 0x01},
		{0x00, 0x01, 0xf5, 0x01}, {0x00, 0x01, 0xf6, 0x01},
		{0x00, 0x01, 0xf7, 0x01}, {0x00, 0x01, 0xf8, 0x01},
		{0x00, 0x01, 0xfa, 0x01}, {0x00, 0x01, 0xfb, 0x01},
		{0x00, 0x01, 0xfc, 0x01}, {0x00, 0x01, 0xfd, 0x01}
	],
	[
		{0x01, 0x01, 0xd3, 0x00}, {0x16, 0x01, 0xd3, 0x01},
		{0x01, 0x01, 0xd4, 0x00}, {0x16, 0x01, 0xd4, 0x01},
		{0x01, 0x01, 0xd6, 0x00}, {0x16, 0x01, 0xd6, 0x01},
		{0x01, 0x01, 0xdd, 0x00}, {0x16, 0x01, 0xdd, 0x01},
		{0x01, 0x01, 0xde, 0x00}, {0x16, 0x01, 0xde, 0x01},
		{0x01, 0x01, 0xdf, 0x00}, {0x16, 0x01, 0xdf, 0x01},
		{0x01, 0x01, 0xf1, 0x00}, {0x16, 0x01, 0xf1, 0x01},
		{0x01, 0x01, 0xf4, 0x00}, {0x16, 0x01, 0xf4, 0x01}
	],
	/* 210 */
	[
		{0x02, 0x01, 0xd3, 0x00}, {0x09, 0x01, 0xd3, 0x00},
		{0x17, 0x01, 0xd3, 0x00}, {0x28, 0x01, 0xd3, 0x01},
		{0x02, 0x01, 0xd4, 0x00}, {0x09, 0x01, 0xd4, 0x00},
		{0x17, 0x01, 0xd4, 0x00}, {0x28, 0x01, 0xd4, 0x01},
		{0x02, 0x01, 0xd6, 0x00}, {0x09, 0x01, 0xd6, 0x00},
		{0x17, 0x01, 0xd6, 0x00}, {0x28, 0x01, 0xd6, 0x01},
		{0x02, 0x01, 0xdd, 0x00}, {0x09, 0x01, 0xdd, 0x00},
		{0x17, 0x01, 0xdd, 0x00}, {0x28, 0x01, 0xdd, 0x01}
	],
	[
		{0x03, 0x01, 0xd3, 0x00}, {0x06, 0x01, 0xd3, 0x00},
		{0x0a, 0x01, 0xd3, 0x00}, {0x0f, 0x01, 0xd3, 0x00},
		{0x18, 0x01, 0xd3, 0x00}, {0x1f, 0x01, 0xd3, 0x00},
		{0x29, 0x01, 0xd3, 0x00}, {0x38, 0x01, 0xd3, 0x01},
		{0x03, 0x01, 0xd4, 0x00}, {0x06, 0x01, 0xd4, 0x00},
		{0x0a, 0x01, 0xd4, 0x00}, {0x0f, 0x01, 0xd4, 0x00},
		{0x18, 0x01, 0xd4, 0x00}, {0x1f, 0x01, 0xd4, 0x00},
		{0x29, 0x01, 0xd4, 0x00}, {0x38, 0x01, 0xd4, 0x01}
	],
	[
		{0x03, 0x01, 0xd6, 0x00}, {0x06, 0x01, 0xd6, 0x00},
		{0x0a, 0x01, 0xd6, 0x00}, {0x0f, 0x01, 0xd6, 0x00},
		{0x18, 0x01, 0xd6, 0x00}, {0x1f, 0x01, 0xd6, 0x00},
		{0x29, 0x01, 0xd6, 0x00}, {0x38, 0x01, 0xd6, 0x01},
		{0x03, 0x01, 0xdd, 0x00}, {0x06, 0x01, 0xdd, 0x00},
		{0x0a, 0x01, 0xdd, 0x00}, {0x0f, 0x01, 0xdd, 0x00},
		{0x18, 0x01, 0xdd, 0x00}, {0x1f, 0x01, 0xdd, 0x00},
		{0x29, 0x01, 0xdd, 0x00}, {0x38, 0x01, 0xdd, 0x01}
	],
	[
		{0x02, 0x01, 0xde, 0x00}, {0x09, 0x01, 0xde, 0x00},
		{0x17, 0x01, 0xde, 0x00}, {0x28, 0x01, 0xde, 0x01},
		{0x02, 0x01, 0xdf, 0x00}, {0x09, 0x01, 0xdf, 0x00},
		{0x17, 0x01, 0xdf, 0x00}, {0x28, 0x01, 0xdf, 0x01},
		{0x02, 0x01, 0xf1, 0x00}, {0x09, 0x01, 0xf1, 0x00},
		{0x17, 0x01, 0xf1, 0x00}, {0x28, 0x01, 0xf1, 0x01},
		{0x02, 0x01, 0xf4, 0x00}, {0x09, 0x01, 0xf4, 0x00},
		{0x17, 0x01, 0xf4, 0x00}, {0x28, 0x01, 0xf4, 0x01}
	],
	[
		{0x03, 0x01, 0xde, 0x00}, {0x06, 0x01, 0xde, 0x00},
		{0x0a, 0x01, 0xde, 0x00}, {0x0f, 0x01, 0xde, 0x00},
		{0x18, 0x01, 0xde, 0x00}, {0x1f, 0x01, 0xde, 0x00},
		{0x29, 0x01, 0xde, 0x00}, {0x38, 0x01, 0xde, 0x01},
		{0x03, 0x01, 0xdf, 0x00}, {0x06, 0x01, 0xdf, 0x00},
		{0x0a, 0x01, 0xdf, 0x00}, {0x0f, 0x01, 0xdf, 0x00},
		{0x18, 0x01, 0xdf, 0x00}, {0x1f, 0x01, 0xdf, 0x00},
		{0x29, 0x01, 0xdf, 0x00}, {0x38, 0x01, 0xdf, 0x01}
	],
	/* 215 */
	[
		{0x03, 0x01, 0xf1, 0x00}, {0x06, 0x01, 0xf1, 0x00},
		{0x0a, 0x01, 0xf1, 0x00}, {0x0f, 0x01, 0xf1, 0x00},
		{0x18, 0x01, 0xf1, 0x00}, {0x1f, 0x01, 0xf1, 0x00},
		{0x29, 0x01, 0xf1, 0x00}, {0x38, 0x01, 0xf1, 0x01},
		{0x03, 0x01, 0xf4, 0x00}, {0x06, 0x01, 0xf4, 0x00},
		{0x0a, 0x01, 0xf4, 0x00}, {0x0f, 0x01, 0xf4, 0x00},
		{0x18, 0x01, 0xf4, 0x00}, {0x1f, 0x01, 0xf4, 0x00},
		{0x29, 0x01, 0xf4, 0x00}, {0x38, 0x01, 0xf4, 0x01}
	],
	[
		{0x01, 0x01, 0xf5, 0x00}, {0x16, 0x01, 0xf5, 0x01},
		{0x01, 0x01, 0xf6, 0x00}, {0x16, 0x01, 0xf6, 0x01},
		{0x01, 0x01, 0xf7, 0x00}, {0x16, 0x01, 0xf7, 0x01},
		{0x01, 0x01, 0xf8, 0x00}, {0x16, 0x01, 0xf8, 0x01},
		{0x01, 0x01, 0xfa, 0x00}, {0x16, 0x01, 0xfa, 0x01},
		{0x01, 0x01, 0xfb, 0x00}, {0x16, 0x01, 0xfb, 0x01},
		{0x01, 0x01, 0xfc, 0x00}, {0x16, 0x01, 0xfc, 0x01},
		{0x01, 0x01, 0xfd, 0x00}, {0x16, 0x01, 0xfd, 0x01}
	],
	[
		{0x02, 0x01, 0xf5, 0x00}, {0x09, 0x01, 0xf5, 0x00},
		{0x17, 0x01, 0xf5, 0x00}, {0x28, 0x01, 0xf5, 0x01},
		{0x02, 0x01, 0xf6, 0x00}, {0x09, 0x01, 0xf6, 0x00},
		{0x17, 0x01, 0xf6, 0x00}, {0x28, 0x01, 0xf6, 0x01},
		{0x02, 0x01, 0xf7, 0x00}, {0x09, 0x01, 0xf7, 0x00},
		{0x17, 0x01, 0xf7, 0x00}, {0x28, 0x01, 0xf7, 0x01},
		{0x02, 0x01, 0xf8, 0x00}, {0x09, 0x01, 0xf8, 0x00},
		{0x17, 0x01, 0xf8, 0x00}, {0x28, 0x01, 0xf8, 0x01}
	],
	[
		{0x03, 0x01, 0xf5, 0x00}, {0x06, 0x01, 0xf5, 0x00},
		{0x0a, 0x01, 0xf5, 0x00}, {0x0f, 0x01, 0xf5, 0x00},
		{0x18, 0x01, 0xf5, 0x00}, {0x1f, 0x01, 0xf5, 0x00},
		{0x29, 0x01, 0xf5, 0x00}, {0x38, 0x01, 0xf5, 0x01},
		{0x03, 0x01, 0xf6, 0x00}, {0x06, 0x01, 0xf6, 0x00},
		{0x0a, 0x01, 0xf6, 0x00}, {0x0f, 0x01, 0xf6, 0x00},
		{0x18, 0x01, 0xf6, 0x00}, {0x1f, 0x01, 0xf6, 0x00},
		{0x29, 0x01, 0xf6, 0x00}, {0x38, 0x01, 0xf6, 0x01}
	],
	[
		{0x03, 0x01, 0xf7, 0x00}, {0x06, 0x01, 0xf7, 0x00},
		{0x0a, 0x01, 0xf7, 0x00}, {0x0f, 0x01, 0xf7, 0x00},
		{0x18, 0x01, 0xf7, 0x00}, {0x1f, 0x01, 0xf7, 0x00},
		{0x29, 0x01, 0xf7, 0x00}, {0x38, 0x01, 0xf7, 0x01},
		{0x03, 0x01, 0xf8, 0x00}, {0x06, 0x01, 0xf8, 0x00},
		{0x0a, 0x01, 0xf8, 0x00}, {0x0f, 0x01, 0xf8, 0x00},
		{0x18, 0x01, 0xf8, 0x00}, {0x1f, 0x01, 0xf8, 0x00},
		{0x29, 0x01, 0xf8, 0x00}, {0x38, 0x01, 0xf8, 0x01}
	],
	/* 220 */
	[
		{0x02, 0x01, 0xfa, 0x00}, {0x09, 0x01, 0xfa, 0x00},
		{0x17, 0x01, 0xfa, 0x00}, {0x28, 0x01, 0xfa, 0x01},
		{0x02, 0x01, 0xfb, 0x00}, {0x09, 0x01, 0xfb, 0x00},
		{0x17, 0x01, 0xfb, 0x00}, {0x28, 0x01, 0xfb, 0x01},
		{0x02, 0x01, 0xfc, 0x00}, {0x09, 0x01, 0xfc, 0x00},
		{0x17, 0x01, 0xfc, 0x00}, {0x28, 0x01, 0xfc, 0x01},
		{0x02, 0x01, 0xfd, 0x00}, {0x09, 0x01, 0xfd, 0x00},
		{0x17, 0x01, 0xfd, 0x00}, {0x28, 0x01, 0xfd, 0x01}
	],
	[
		{0x03, 0x01, 0xfa, 0x00}, {0x06, 0x01, 0xfa, 0x00},
		{0x0a, 0x01, 0xfa, 0x00}, {0x0f, 0x01, 0xfa, 0x00},
		{0x18, 0x01, 0xfa, 0x00}, {0x1f, 0x01, 0xfa, 0x00},
		{0x29, 0x01, 0xfa, 0x00}, {0x38, 0x01, 0xfa, 0x01},
		{0x03, 0x01, 0xfb, 0x00}, {0x06, 0x01, 0xfb, 0x00},
		{0x0a, 0x01, 0xfb, 0x00}, {0x0f, 0x01, 0xfb, 0x00},
		{0x18, 0x01, 0xfb, 0x00}, {0x1f, 0x01, 0xfb, 0x00},
		{0x29, 0x01, 0xfb, 0x00}, {0x38, 0x01, 0xfb, 0x01}
	],
	[
		{0x03, 0x01, 0xfc, 0x00}, {0x06, 0x01, 0xfc, 0x00},
		{0x0a, 0x01, 0xfc, 0x00}, {0x0f, 0x01, 0xfc, 0x00},
		{0x18, 0x01, 0xfc, 0x00}, {0x1f, 0x01, 0xfc, 0x00},
		{0x29, 0x01, 0xfc, 0x00}, {0x38, 0x01, 0xfc, 0x01},
		{0x03, 0x01, 0xfd, 0x00}, {0x06, 0x01, 0xfd, 0x00},
		{0x0a, 0x01, 0xfd, 0x00}, {0x0f, 0x01, 0xfd, 0x00},
		{0x18, 0x01, 0xfd, 0x00}, {0x1f, 0x01, 0xfd, 0x00},
		{0x29, 0x01, 0xfd, 0x00}, {0x38, 0x01, 0xfd, 0x01}
	],
	[
		{0x00, 0x01, 0xfe, 0x01}, {0xe3, 0x00, 0x00, 0x00},
		{0xe5, 0x00, 0x00, 0x00}, {0xe6, 0x00, 0x00, 0x00},
		{0xe9, 0x00, 0x00, 0x00}, {0xea, 0x00, 0x00, 0x00},
		{0xec, 0x00, 0x00, 0x00}, {0xed, 0x00, 0x00, 0x00},
		{0xf1, 0x00, 0x00, 0x00}, {0xf2, 0x00, 0x00, 0x00},
		{0xf4, 0x00, 0x00, 0x00}, {0xf5, 0x00, 0x00, 0x00},
		{0xf8, 0x00, 0x00, 0x00}, {0xf9, 0x00, 0x00, 0x00},
		{0xfb, 0x00, 0x00, 0x00}, {0xfc, 0x00, 0x00, 0x01}
	],
	[
		{0x01, 0x01, 0xfe, 0x00}, {0x16, 0x01, 0xfe, 0x01},
		{0x00, 0x01, 0x02, 0x01}, {0x00, 0x01, 0x03, 0x01},
		{0x00, 0x01, 0x04, 0x01}, {0x00, 0x01, 0x05, 0x01},
		{0x00, 0x01, 0x06, 0x01}, {0x00, 0x01, 0x07, 0x01},
		{0x00, 0x01, 0x08, 0x01}, {0x00, 0x01, 0x0b, 0x01},
		{0x00, 0x01, 0x0c, 0x01}, {0x00, 0x01, 0x0e, 0x01},
		{0x00, 0x01, 0x0f, 0x01}, {0x00, 0x01, 0x10, 0x01},
		{0x00, 0x01, 0x11, 0x01}, {0x00, 0x01, 0x12, 0x01}
	],
	/* 225 */
	[
		{0x02, 0x01, 0xfe, 0x00}, {0x09, 0x01, 0xfe, 0x00},
		{0x17, 0x01, 0xfe, 0x00}, {0x28, 0x01, 0xfe, 0x01},
		{0x01, 0x01, 0x02, 0x00}, {0x16, 0x01, 0x02, 0x01},
		{0x01, 0x01, 0x03, 0x00}, {0x16, 0x01, 0x03, 0x01},
		{0x01, 0x01, 0x04, 0x00}, {0x16, 0x01, 0x04, 0x01},
		{0x01, 0x01, 0x05, 0x00}, {0x16, 0x01, 0x05, 0x01},
		{0x01, 0x01, 0x06, 0x00}, {0x16, 0x01, 0x06, 0x01},
		{0x01, 0x01, 0x07, 0x00}, {0x16, 0x01, 0x07, 0x01}
	],
	[
		{0x03, 0x01, 0xfe, 0x00}, {0x06, 0x01, 0xfe, 0x00},
		{0x0a, 0x01, 0xfe, 0x00}, {0x0f, 0x01, 0xfe, 0x00},
		{0x18, 0x01, 0xfe, 0x00}, {0x1f, 0x01, 0xfe, 0x00},
		{0x29, 0x01, 0xfe, 0x00}, {0x38, 0x01, 0xfe, 0x01},
		{0x02, 0x01, 0x02, 0x00}, {0x09, 0x01, 0x02, 0x00},
		{0x17, 0x01, 0x02, 0x00}, {0x28, 0x01, 0x02, 0x01},
		{0x02, 0x01, 0x03, 0x00}, {0x09, 0x01, 0x03, 0x00},
		{0x17, 0x01, 0x03, 0x00}, {0x28, 0x01, 0x03, 0x01}
	],
	[
		{0x03, 0x01, 0x02, 0x00}, {0x06, 0x01, 0x02, 0x00},
		{0x0a, 0x01, 0x02, 0x00}, {0x0f, 0x01, 0x02, 0x00},
		{0x18, 0x01, 0x02, 0x00}, {0x1f, 0x01, 0x02, 0x00},
		{0x29, 0x01, 0x02, 0x00}, {0x38, 0x01, 0x02, 0x01},
		{0x03, 0x01, 0x03, 0x00}, {0x06, 0x01, 0x03, 0x00},
		{0x0a, 0x01, 0x03, 0x00}, {0x0f, 0x01, 0x03, 0x00},
		{0x18, 0x01, 0x03, 0x00}, {0x1f, 0x01, 0x03, 0x00},
		{0x29, 0x01, 0x03, 0x00}, {0x38, 0x01, 0x03, 0x01}
	],
	[
		{0x02, 0x01, 0x04, 0x00}, {0x09, 0x01, 0x04, 0x00},
		{0x17, 0x01, 0x04, 0x00}, {0x28, 0x01, 0x04, 0x01},
		{0x02, 0x01, 0x05, 0x00}, {0x09, 0x01, 0x05, 0x00},
		{0x17, 0x01, 0x05, 0x00}, {0x28, 0x01, 0x05, 0x01},
		{0x02, 0x01, 0x06, 0x00}, {0x09, 0x01, 0x06, 0x00},
		{0x17, 0x01, 0x06, 0x00}, {0x28, 0x01, 0x06, 0x01},
		{0x02, 0x01, 0x07, 0x00}, {0x09, 0x01, 0x07, 0x00},
		{0x17, 0x01, 0x07, 0x00}, {0x28, 0x01, 0x07, 0x01}
	],
	[
		{0x03, 0x01, 0x04, 0x00}, {0x06, 0x01, 0x04, 0x00},
		{0x0a, 0x01, 0x04, 0x00}, {0x0f, 0x01, 0x04, 0x00},
		{0x18, 0x01, 0x04, 0x00}, {0x1f, 0x01, 0x04, 0x00},
		{0x29, 0x01, 0x04, 0x00}, {0x38, 0x01, 0x04, 0x01},
		{0x03, 0x01, 0x05, 0x00}, {0x06, 0x01, 0x05, 0x00},
		{0x0a, 0x01, 0x05, 0x00}, {0x0f, 0x01, 0x05, 0x00},
		{0x18, 0x01, 0x05, 0x00}, {0x1f, 0x01, 0x05, 0x00},
		{0x29, 0x01, 0x05, 0x00}, {0x38, 0x01, 0x05, 0x01}
	],
	/* 230 */
	[
		{0x03, 0x01, 0x06, 0x00}, {0x06, 0x01, 0x06, 0x00},
		{0x0a, 0x01, 0x06, 0x00}, {0x0f, 0x01, 0x06, 0x00},
		{0x18, 0x01, 0x06, 0x00}, {0x1f, 0x01, 0x06, 0x00},
		{0x29, 0x01, 0x06, 0x00}, {0x38, 0x01, 0x06, 0x01},
		{0x03, 0x01, 0x07, 0x00}, {0x06, 0x01, 0x07, 0x00},
		{0x0a, 0x01, 0x07, 0x00}, {0x0f, 0x01, 0x07, 0x00},
		{0x18, 0x01, 0x07, 0x00}, {0x1f, 0x01, 0x07, 0x00},
		{0x29, 0x01, 0x07, 0x00}, {0x38, 0x01, 0x07, 0x01}
	],
	[
		{0x01, 0x01, 0x08, 0x00}, {0x16, 0x01, 0x08, 0x01},
		{0x01, 0x01, 0x0b, 0x00}, {0x16, 0x01, 0x0b, 0x01},
		{0x01, 0x01, 0x0c, 0x00}, {0x16, 0x01, 0x0c, 0x01},
		{0x01, 0x01, 0x0e, 0x00}, {0x16, 0x01, 0x0e, 0x01},
		{0x01, 0x01, 0x0f, 0x00}, {0x16, 0x01, 0x0f, 0x01},
		{0x01, 0x01, 0x10, 0x00}, {0x16, 0x01, 0x10, 0x01},
		{0x01, 0x01, 0x11, 0x00}, {0x16, 0x01, 0x11, 0x01},
		{0x01, 0x01, 0x12, 0x00}, {0x16, 0x01, 0x12, 0x01}
	],
	[
		{0x02, 0x01, 0x08, 0x00}, {0x09, 0x01, 0x08, 0x00},
		{0x17, 0x01, 0x08, 0x00}, {0x28, 0x01, 0x08, 0x01},
		{0x02, 0x01, 0x0b, 0x00}, {0x09, 0x01, 0x0b, 0x00},
		{0x17, 0x01, 0x0b, 0x00}, {0x28, 0x01, 0x0b, 0x01},
		{0x02, 0x01, 0x0c, 0x00}, {0x09, 0x01, 0x0c, 0x00},
		{0x17, 0x01, 0x0c, 0x00}, {0x28, 0x01, 0x0c, 0x01},
		{0x02, 0x01, 0x0e, 0x00}, {0x09, 0x01, 0x0e, 0x00},
		{0x17, 0x01, 0x0e, 0x00}, {0x28, 0x01, 0x0e, 0x01}
	],
	[
		{0x03, 0x01, 0x08, 0x00}, {0x06, 0x01, 0x08, 0x00},
		{0x0a, 0x01, 0x08, 0x00}, {0x0f, 0x01, 0x08, 0x00},
		{0x18, 0x01, 0x08, 0x00}, {0x1f, 0x01, 0x08, 0x00},
		{0x29, 0x01, 0x08, 0x00}, {0x38, 0x01, 0x08, 0x01},
		{0x03, 0x01, 0x0b, 0x00}, {0x06, 0x01, 0x0b, 0x00},
		{0x0a, 0x01, 0x0b, 0x00}, {0x0f, 0x01, 0x0b, 0x00},
		{0x18, 0x01, 0x0b, 0x00}, {0x1f, 0x01, 0x0b, 0x00},
		{0x29, 0x01, 0x0b, 0x00}, {0x38, 0x01, 0x0b, 0x01}
	],
	[
		{0x03, 0x01, 0x0c, 0x00}, {0x06, 0x01, 0x0c, 0x00},
		{0x0a, 0x01, 0x0c, 0x00}, {0x0f, 0x01, 0x0c, 0x00},
		{0x18, 0x01, 0x0c, 0x00}, {0x1f, 0x01, 0x0c, 0x00},
		{0x29, 0x01, 0x0c, 0x00}, {0x38, 0x01, 0x0c, 0x01},
		{0x03, 0x01, 0x0e, 0x00}, {0x06, 0x01, 0x0e, 0x00},
		{0x0a, 0x01, 0x0e, 0x00}, {0x0f, 0x01, 0x0e, 0x00},
		{0x18, 0x01, 0x0e, 0x00}, {0x1f, 0x01, 0x0e, 0x00},
		{0x29, 0x01, 0x0e, 0x00}, {0x38, 0x01, 0x0e, 0x01}
	],
	/* 235 */
	[
		{0x02, 0x01, 0x0f, 0x00}, {0x09, 0x01, 0x0f, 0x00},
		{0x17, 0x01, 0x0f, 0x00}, {0x28, 0x01, 0x0f, 0x01},
		{0x02, 0x01, 0x10, 0x00}, {0x09, 0x01, 0x10, 0x00},
		{0x17, 0x01, 0x10, 0x00}, {0x28, 0x01, 0x10, 0x01},
		{0x02, 0x01, 0x11, 0x00}, {0x09, 0x01, 0x11, 0x00},
		{0x17, 0x01, 0x11, 0x00}, {0x28, 0x01, 0x11, 0x01},
		{0x02, 0x01, 0x12, 0x00}, {0x09, 0x01, 0x12, 0x00},
		{0x17, 0x01, 0x12, 0x00}, {0x28, 0x01, 0x12, 0x01}
	],
	[
		{0x03, 0x01, 0x0f, 0x00}, {0x06, 0x01, 0x0f, 0x00},
		{0x0a, 0x01, 0x0f, 0x00}, {0x0f, 0x01, 0x0f, 0x00},
		{0x18, 0x01, 0x0f, 0x00}, {0x1f, 0x01, 0x0f, 0x00},
		{0x29, 0x01, 0x0f, 0x00}, {0x38, 0x01, 0x0f, 0x01},
		{0x03, 0x01, 0x10, 0x00}, {0x06, 0x01, 0x10, 0x00},
		{0x0a, 0x01, 0x10, 0x00}, {0x0f, 0x01, 0x10, 0x00},
		{0x18, 0x01, 0x10, 0x00}, {0x1f, 0x01, 0x10, 0x00},
		{0x29, 0x01, 0x10, 0x00}, {0x38, 0x01, 0x10, 0x01}
	],
	[
		{0x03, 0x01, 0x11, 0x00}, {0x06, 0x01, 0x11, 0x00},
		{0x0a, 0x01, 0x11, 0x00}, {0x0f, 0x01, 0x11, 0x00},
		{0x18, 0x01, 0x11, 0x00}, {0x1f, 0x01, 0x11, 0x00},
		{0x29, 0x01, 0x11, 0x00}, {0x38, 0x01, 0x11, 0x01},
		{0x03, 0x01, 0x12, 0x00}, {0x06, 0x01, 0x12, 0x00},
		{0x0a, 0x01, 0x12, 0x00}, {0x0f, 0x01, 0x12, 0x00},
		{0x18, 0x01, 0x12, 0x00}, {0x1f, 0x01, 0x12, 0x00},
		{0x29, 0x01, 0x12, 0x00}, {0x38, 0x01, 0x12, 0x01}
	],
	[
		{0x00, 0x01, 0x13, 0x01}, {0x00, 0x01, 0x14, 0x01},
		{0x00, 0x01, 0x15, 0x01}, {0x00, 0x01, 0x17, 0x01},
		{0x00, 0x01, 0x18, 0x01}, {0x00, 0x01, 0x19, 0x01},
		{0x00, 0x01, 0x1a, 0x01}, {0x00, 0x01, 0x1b, 0x01},
		{0x00, 0x01, 0x1c, 0x01}, {0x00, 0x01, 0x1d, 0x01},
		{0x00, 0x01, 0x1e, 0x01}, {0x00, 0x01, 0x1f, 0x01},
		{0x00, 0x01, 0x7f, 0x01}, {0x00, 0x01, 0xdc, 0x01},
		{0x00, 0x01, 0xf9, 0x01}, {0xfd, 0x00, 0x00, 0x01}
	],
	[
		{0x01, 0x01, 0x13, 0x00}, {0x16, 0x01, 0x13, 0x01},
		{0x01, 0x01, 0x14, 0x00}, {0x16, 0x01, 0x14, 0x01},
		{0x01, 0x01, 0x15, 0x00}, {0x16, 0x01, 0x15, 0x01},
		{0x01, 0x01, 0x17, 0x00}, {0x16, 0x01, 0x17, 0x01},
		{0x01, 0x01, 0x18, 0x00}, {0x16, 0x01, 0x18, 0x01},
		{0x01, 0x01, 0x19, 0x00}, {0x16, 0x01, 0x19, 0x01},
		{0x01, 0x01, 0x1a, 0x00}, {0x16, 0x01, 0x1a, 0x01},
		{0x01, 0x01, 0x1b, 0x00}, {0x16, 0x01, 0x1b, 0x01}
	],
	/* 240 */
	[
		{0x02, 0x01, 0x13, 0x00}, {0x09, 0x01, 0x13, 0x00},
		{0x17, 0x01, 0x13, 0x00}, {0x28, 0x01, 0x13, 0x01},
		{0x02, 0x01, 0x14, 0x00}, {0x09, 0x01, 0x14, 0x00},
		{0x17, 0x01, 0x14, 0x00}, {0x28, 0x01, 0x14, 0x01},
		{0x02, 0x01, 0x15, 0x00}, {0x09, 0x01, 0x15, 0x00},
		{0x17, 0x01, 0x15, 0x00}, {0x28, 0x01, 0x15, 0x01},
		{0x02, 0x01, 0x17, 0x00}, {0x09, 0x01, 0x17, 0x00},
		{0x17, 0x01, 0x17, 0x00}, {0x28, 0x01, 0x17, 0x01}
	],
	[
		{0x03, 0x01, 0x13, 0x00}, {0x06, 0x01, 0x13, 0x00},
		{0x0a, 0x01, 0x13, 0x00}, {0x0f, 0x01, 0x13, 0x00},
		{0x18, 0x01, 0x13, 0x00}, {0x1f, 0x01, 0x13, 0x00},
		{0x29, 0x01, 0x13, 0x00}, {0x38, 0x01, 0x13, 0x01},
		{0x03, 0x01, 0x14, 0x00}, {0x06, 0x01, 0x14, 0x00},
		{0x0a, 0x01, 0x14, 0x00}, {0x0f, 0x01, 0x14, 0x00},
		{0x18, 0x01, 0x14, 0x00}, {0x1f, 0x01, 0x14, 0x00},
		{0x29, 0x01, 0x14, 0x00}, {0x38, 0x01, 0x14, 0x01}
	],
	[
		{0x03, 0x01, 0x15, 0x00}, {0x06, 0x01, 0x15, 0x00},
		{0x0a, 0x01, 0x15, 0x00}, {0x0f, 0x01, 0x15, 0x00},
		{0x18, 0x01, 0x15, 0x00}, {0x1f, 0x01, 0x15, 0x00},
		{0x29, 0x01, 0x15, 0x00}, {0x38, 0x01, 0x15, 0x01},
		{0x03, 0x01, 0x17, 0x00}, {0x06, 0x01, 0x17, 0x00},
		{0x0a, 0x01, 0x17, 0x00}, {0x0f, 0x01, 0x17, 0x00},
		{0x18, 0x01, 0x17, 0x00}, {0x1f, 0x01, 0x17, 0x00},
		{0x29, 0x01, 0x17, 0x00}, {0x38, 0x01, 0x17, 0x01}
	],
	[
		{0x02, 0x01, 0x18, 0x00}, {0x09, 0x01, 0x18, 0x00},
		{0x17, 0x01, 0x18, 0x00}, {0x28, 0x01, 0x18, 0x01},
		{0x02, 0x01, 0x19, 0x00}, {0x09, 0x01, 0x19, 0x00},
		{0x17, 0x01, 0x19, 0x00}, {0x28, 0x01, 0x19, 0x01},
		{0x02, 0x01, 0x1a, 0x00}, {0x09, 0x01, 0x1a, 0x00},
		{0x17, 0x01, 0x1a, 0x00}, {0x28, 0x01, 0x1a, 0x01},
		{0x02, 0x01, 0x1b, 0x00}, {0x09, 0x01, 0x1b, 0x00},
		{0x17, 0x01, 0x1b, 0x00}, {0x28, 0x01, 0x1b, 0x01}
	],
	[
		{0x03, 0x01, 0x18, 0x00}, {0x06, 0x01, 0x18, 0x00},
		{0x0a, 0x01, 0x18, 0x00}, {0x0f, 0x01, 0x18, 0x00},
		{0x18, 0x01, 0x18, 0x00}, {0x1f, 0x01, 0x18, 0x00},
		{0x29, 0x01, 0x18, 0x00}, {0x38, 0x01, 0x18, 0x01},
		{0x03, 0x01, 0x19, 0x00}, {0x06, 0x01, 0x19, 0x00},
		{0x0a, 0x01, 0x19, 0x00}, {0x0f, 0x01, 0x19, 0x00},
		{0x18, 0x01, 0x19, 0x00}, {0x1f, 0x01, 0x19, 0x00},
		{0x29, 0x01, 0x19, 0x00}, {0x38, 0x01, 0x19, 0x01}
	],
	/* 245 */
	[
		{0x03, 0x01, 0x1a, 0x00}, {0x06, 0x01, 0x1a, 0x00},
		{0x0a, 0x01, 0x1a, 0x00}, {0x0f, 0x01, 0x1a, 0x00},
		{0x18, 0x01, 0x1a, 0x00}, {0x1f, 0x01, 0x1a, 0x00},
		{0x29, 0x01, 0x1a, 0x00}, {0x38, 0x01, 0x1a, 0x01},
		{0x03, 0x01, 0x1b, 0x00}, {0x06, 0x01, 0x1b, 0x00},
		{0x0a, 0x01, 0x1b, 0x00}, {0x0f, 0x01, 0x1b, 0x00},
		{0x18, 0x01, 0x1b, 0x00}, {0x1f, 0x01, 0x1b, 0x00},
		{0x29, 0x01, 0x1b, 0x00}, {0x38, 0x01, 0x1b, 0x01}
	],
	[
		{0x01, 0x01, 0x1c, 0x00}, {0x16, 0x01, 0x1c, 0x01},
		{0x01, 0x01, 0x1d, 0x00}, {0x16, 0x01, 0x1d, 0x01},
		{0x01, 0x01, 0x1e, 0x00}, {0x16, 0x01, 0x1e, 0x01},
		{0x01, 0x01, 0x1f, 0x00}, {0x16, 0x01, 0x1f, 0x01},
		{0x01, 0x01, 0x7f, 0x00}, {0x16, 0x01, 0x7f, 0x01},
		{0x01, 0x01, 0xdc, 0x00}, {0x16, 0x01, 0xdc, 0x01},
		{0x01, 0x01, 0xf9, 0x00}, {0x16, 0x01, 0xf9, 0x01},
		{0xfe, 0x00, 0x00, 0x00}, {0xff, 0x00, 0x00, 0x01}
	],
	[
		{0x02, 0x01, 0x1c, 0x00}, {0x09, 0x01, 0x1c, 0x00},
		{0x17, 0x01, 0x1c, 0x00}, {0x28, 0x01, 0x1c, 0x01},
		{0x02, 0x01, 0x1d, 0x00}, {0x09, 0x01, 0x1d, 0x00},
		{0x17, 0x01, 0x1d, 0x00}, {0x28, 0x01, 0x1d, 0x01},
		{0x02, 0x01, 0x1e, 0x00}, {0x09, 0x01, 0x1e, 0x00},
		{0x17, 0x01, 0x1e, 0x00}, {0x28, 0x01, 0x1e, 0x01},
		{0x02, 0x01, 0x1f, 0x00}, {0x09, 0x01, 0x1f, 0x00},
		{0x17, 0x01, 0x1f, 0x00}, {0x28, 0x01, 0x1f, 0x01}
	],
	[
		{0x03, 0x01, 0x1c, 0x00}, {0x06, 0x01, 0x1c, 0x00},
		{0x0a, 0x01, 0x1c, 0x00}, {0x0f, 0x01, 0x1c, 0x00},
		{0x18, 0x01, 0x1c, 0x00}, {0x1f, 0x01, 0x1c, 0x00},
		{0x29, 0x01, 0x1c, 0x00}, {0x38, 0x01, 0x1c, 0x01},
		{0x03, 0x01, 0x1d, 0x00}, {0x06, 0x01, 0x1d, 0x00},
		{0x0a, 0x01, 0x1d, 0x00}, {0x0f, 0x01, 0x1d, 0x00},
		{0x18, 0x01, 0x1d, 0x00}, {0x1f, 0x01, 0x1d, 0x00},
		{0x29, 0x01, 0x1d, 0x00}, {0x38, 0x01, 0x1d, 0x01}
	],
	[
		{0x03, 0x01, 0x1e, 0x00}, {0x06, 0x01, 0x1e, 0x00},
		{0x0a, 0x01, 0x1e, 0x00}, {0x0f, 0x01, 0x1e, 0x00},
		{0x18, 0x01, 0x1e, 0x00}, {0x1f, 0x01, 0x1e, 0x00},
		{0x29, 0x01, 0x1e, 0x00}, {0x38, 0x01, 0x1e, 0x01},
		{0x03, 0x01, 0x1f, 0x00}, {0x06, 0x01, 0x1f, 0x00},
		{0x0a, 0x01, 0x1f, 0x00}, {0x0f, 0x01, 0x1f, 0x00},
		{0x18, 0x01, 0x1f, 0x00}, {0x1f, 0x01, 0x1f, 0x00},
		{0x29, 0x01, 0x1f, 0x00}, {0x38, 0x01, 0x1f, 0x01}
	],
	/* 250 */
	[
		{0x02, 0x01, 0x7f, 0x00}, {0x09, 0x01, 0x7f, 0x00},
		{0x17, 0x01, 0x7f, 0x00}, {0x28, 0x01, 0x7f, 0x01},
		{0x02, 0x01, 0xdc, 0x00}, {0x09, 0x01, 0xdc, 0x00},
		{0x17, 0x01, 0xdc, 0x00}, {0x28, 0x01, 0xdc, 0x01},
		{0x02, 0x01, 0xf9, 0x00}, {0x09, 0x01, 0xf9, 0x00},
		{0x17, 0x01, 0xf9, 0x00}, {0x28, 0x01, 0xf9, 0x01},
		{0x00, 0x01, 0x0a, 0x01}, {0x00, 0x01, 0x0d, 0x01},
		{0x00, 0x01, 0x16, 0x01}, {0xfa, 0x00, 0x00, 0x00}
	],
	[
		{0x03, 0x01, 0x7f, 0x00}, {0x06, 0x01, 0x7f, 0x00},
		{0x0a, 0x01, 0x7f, 0x00}, {0x0f, 0x01, 0x7f, 0x00},
		{0x18, 0x01, 0x7f, 0x00}, {0x1f, 0x01, 0x7f, 0x00},
		{0x29, 0x01, 0x7f, 0x00}, {0x38, 0x01, 0x7f, 0x01},
		{0x03, 0x01, 0xdc, 0x00}, {0x06, 0x01, 0xdc, 0x00},
		{0x0a, 0x01, 0xdc, 0x00}, {0x0f, 0x01, 0xdc, 0x00},
		{0x18, 0x01, 0xdc, 0x00}, {0x1f, 0x01, 0xdc, 0x00},
		{0x29, 0x01, 0xdc, 0x00}, {0x38, 0x01, 0xdc, 0x01}
	],
	[
		{0x03, 0x01, 0xf9, 0x00}, {0x06, 0x01, 0xf9, 0x00},
		{0x0a, 0x01, 0xf9, 0x00}, {0x0f, 0x01, 0xf9, 0x00},
		{0x18, 0x01, 0xf9, 0x00}, {0x1f, 0x01, 0xf9, 0x00},
		{0x29, 0x01, 0xf9, 0x00}, {0x38, 0x01, 0xf9, 0x01},
		{0x01, 0x01, 0x0a, 0x00}, {0x16, 0x01, 0x0a, 0x01},
		{0x01, 0x01, 0x0d, 0x00}, {0x16, 0x01, 0x0d, 0x01},
		{0x01, 0x01, 0x16, 0x00}, {0x16, 0x01, 0x16, 0x01},
		{0xfc, 0x00, 0x00, 0x00}, {0xfc, 0x00, 0x00, 0x00}
	],
	[
		{0x02, 0x01, 0x0a, 0x00}, {0x09, 0x01, 0x0a, 0x00},
		{0x17, 0x01, 0x0a, 0x00}, {0x28, 0x01, 0x0a, 0x01},
		{0x02, 0x01, 0x0d, 0x00}, {0x09, 0x01, 0x0d, 0x00},
		{0x17, 0x01, 0x0d, 0x00}, {0x28, 0x01, 0x0d, 0x01},
		{0x02, 0x01, 0x16, 0x00}, {0x09, 0x01, 0x16, 0x00},
		{0x17, 0x01, 0x16, 0x00}, {0x28, 0x01, 0x16, 0x01},
		{0xfd, 0x00, 0x00, 0x00}, {0xfd, 0x00, 0x00, 0x00},
		{0xfd, 0x00, 0x00, 0x00}, {0xfd, 0x00, 0x00, 0x00}
	],
	[
		{0x03, 0x01, 0x0a, 0x00}, {0x06, 0x01, 0x0a, 0x00},
		{0x0a, 0x01, 0x0a, 0x00}, {0x0f, 0x01, 0x0a, 0x00},
		{0x18, 0x01, 0x0a, 0x00}, {0x1f, 0x01, 0x0a, 0x00},
		{0x29, 0x01, 0x0a, 0x00}, {0x38, 0x01, 0x0a, 0x01},
		{0x03, 0x01, 0x0d, 0x00}, {0x06, 0x01, 0x0d, 0x00},
		{0x0a, 0x01, 0x0d, 0x00}, {0x0f, 0x01, 0x0d, 0x00},
		{0x18, 0x01, 0x0d, 0x00}, {0x1f, 0x01, 0x0d, 0x00},
		{0x29, 0x01, 0x0d, 0x00}, {0x38, 0x01, 0x0d, 0x01}
	],
	/* 255 */
	[
		{0x03, 0x01, 0x16, 0x00}, {0x06, 0x01, 0x16, 0x00},
		{0x0a, 0x01, 0x16, 0x00}, {0x0f, 0x01, 0x16, 0x00},
		{0x18, 0x01, 0x16, 0x00}, {0x1f, 0x01, 0x16, 0x00},
		{0x29, 0x01, 0x16, 0x00}, {0x38, 0x01, 0x16, 0x01},
		{0xff, 0x00, 0x00, 0x00}, {0xff, 0x00, 0x00, 0x00},
		{0xff, 0x00, 0x00, 0x00}, {0xff, 0x00, 0x00, 0x00},
		{0xff, 0x00, 0x00, 0x00}, {0xff, 0x00, 0x00, 0x00},
		{0xff, 0x00, 0x00, 0x00}, {0xff, 0x00, 0x00, 0x00}
	]
];
